/*
    SDL - Simple DirectMedia Layer

    Copyright (C) SEIKO EPSON CORPORATION 2002-2004. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifdef SAVE_RCSID
static char rcsid =
 "@(#) $Id: SDL_aclvideo.c,v 1.8 2004/06/15 09:20:11 acyr Exp $";
#endif

/* 
 * This is a video driver for for SDL using the ACL library from ATI.
 */

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "imconfig.h"

#include "SDL.h"
#include "SDL_error.h"
#include "SDL_video.h"
#include "SDL_mouse.h"
#include "SDL_sysvideo.h"
#include "SDL_pixels_c.h"
#include "SDL_events_c.h"
#include "SDL_aclvideo.h"
#include "SDL_aclevents_c.h"
#include "SDL_aclmouse_c.h"
#include "SDL_aclyuv_c.h"
#include "SDL_aclevents_c.h"

#define ACLVID_DRIVER_NAME "acl"

#define ACL_CHIP_INDEX 0


#if !defined(LCD_WIDTH) || !defined(LCD_HEIGHT)
#error "LCD_WIDTH/LCD_HEIGHT is not defined"
#endif

#define ACL_OUT_WIDTH  1024
#define ACL_OUT_HEIGHT 768

#if (LCD_WIDTH == 1024) && (LCD_HEIGHT == 768)
#define ACL_OUT_FORMAT ACL_VIDEO_FMT_DVI_1024x768P
#elif (LCD_WIDTH == 640) && (LCD_HEIGHT == 480)
#define ACL_OUT_FORMAT ACL_VIDEO_FMT_DVI_640x480P
#else
#error "this resolution is invalid."
#endif


enum {
    ACL_MODELIST_16 = 0,
    ACL_MODELIST_32 = 1
};

static SDL_Rect ACL_modelist16[] = {

    {  0, 0, 1600, 1200 },
    {  0, 0, 1408, 1056 },
    {  0, 0, 1280, 1024 },
    {  0, 0, 1280,  960 },
    {  0, 0, 1152,  864 },
    {  0, 0, 1024,  768 },
    {  0, 0,  960,  720 },
    {  0, 0,  800,  600 },
    {  0, 0,  768,  576 },
    {  0, 0,  640,  480 }
};

static SDL_Rect ACL_modelist32[] = {
    {  0, 0, 1024,  768 },
    {  0, 0,  960,  720 },
    {  0, 0,  800,  600 },
    {  0, 0,  768,  576 },
    {  0, 0,  640,  480 }
};

/* Initialization/Query functions */
static int ACL_VideoInit(_THIS, SDL_PixelFormat *vformat);
static SDL_Rect **ACL_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags);
static SDL_Surface *ACL_SetVideoMode(_THIS, SDL_Surface *current, int width, 
				     int height, int bpp, Uint32 flags);
static int ACL_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors);
static void ACL_VideoQuit(_THIS);

/* Hardware surface functions */
static int ACL_AllocHWSurface(_THIS, SDL_Surface *surface);
static int ACL_LockHWSurface(_THIS, SDL_Surface *surface);
static void ACL_UnlockHWSurface(_THIS, SDL_Surface *surface);
static void ACL_FreeHWSurface(_THIS, SDL_Surface *surface);

/* etc. */
static void ACL_UpdateRects(_THIS, int numrects, SDL_Rect *rects);
static int  ACL_InitHW(_THIS);
static int  ACL_VideoModeInit (_THIS, SDL_Surface* current, 
			       int width, int height, int bpp);
static int ACL_CheckHWBlit(_THIS, SDL_Surface *src, SDL_Surface *dst); 
static int ACL_SetAlpha(_THIS, SDL_Surface *surface, Uint8 value); 
static void ACL_PreventPink();
static void ACL_ZoomWindow(_THIS, SDL_Rect* win);

static void ACL_PreventPink() {
	AtiCore_WriteReg32 (0, 0x12C0, 0);
	AtiCore_WriteReg32 (0, 0x12C4, 0);
	AtiCore_WriteReg32 (0, 0x12C8, 0);
	AtiCore_WriteReg32 (0, 0x12CC, 0);
	AtiCore_WriteReg32 (0, 0x12D0, 0);
	AtiCore_WriteReg32 (0, 0x12D4, 0);
}


/**
 * Disable the overlay according to the scaler parameter
 *
 * \param scaler ACL_OVL_G0SCALE, ACL_OVL_V0SCALE or ACL_OVL_V1SCALE
 */
static void ACL_DisableOverlay(_THIS, ACL_NAME scaler)
{
    ACL_HANDLE ovlhnd;
    ACL_RESULT rc;
    BOOL exists;

    assert( scaler == ACL_OVL_G0SCALE ||
	    scaler == ACL_OVL_V0SCALE ||
	    scaler == ACL_OVL_V1SCALE );

    rc = AclOvl_Open (&ovlhnd, this->hidden->chip, scaler, 
		      ACL_MODE_SHARED, &exists);
    if (ACL_ERROR (rc)) {
	return;
    }

    AclOvl_Enable (ovlhnd, FALSE);
    AclOvl_Close(ovlhnd);
}

/* ACL driver bootstrap functions */
static int ACL_Available(void)
{
	return(1);
}

static void ACL_DeleteDevice(SDL_VideoDevice *device)
{
    if ( device->hidden ) {
	free(device->hidden);
    }

    if ( device ) {
	free(device);
    }
}

static SDL_VideoDevice *ACL_CreateDevice(int devindex)
{
	SDL_VideoDevice *device;
	int rgbYuvScalers, rgbScaler, yuvScaler; 

	rgbYuvScalers = ACL_CheckEnvForScalars(&rgbScaler,
					       &yuvScaler) ;

	if (rgbYuvScalers == SDL_ACL_OVL_INVALID)
	    return NULL ;				      

	/* Initialize all variables that we clean on shutdown */
	device = (SDL_VideoDevice *)malloc(sizeof(SDL_VideoDevice));
	if ( device ) {
		memset(device, 0, (sizeof *device));
		device->hidden = (struct SDL_PrivateVideoData *)
				malloc(sizeof(struct SDL_PrivateVideoData));
	}
	if ( (device == NULL) || (device->hidden == NULL) ) {
		SDL_OutOfMemory();
	if( device->hidden ) {
	    free(device->hidden);
	}
		if ( device ) {
			free(device);
		}
		return(0);
	}
	memset(device->hidden, 0, sizeof(struct SDL_PrivateVideoData));
	device->hidden->rgbYuvScalers = rgbYuvScalers ;
	device->hidden->rgbScaler = rgbScaler ;
	device->hidden->yuvScaler = yuvScaler ;
	
	/* Set the function pointers */
	device->VideoInit = ACL_VideoInit;
	device->ListModes = ACL_ListModes;
	device->SetVideoMode = ACL_SetVideoMode;
	device->CreateYUVOverlay = ACL_CreateYUV; 
	device->SetColors = ACL_SetColors;
	device->UpdateRects = ACL_UpdateRects;
	device->VideoQuit = ACL_VideoQuit;
	device->AllocHWSurface = ACL_AllocHWSurface;
	device->CheckHWBlit = ACL_CheckHWBlit;
	device->FillHWRect = NULL;
	device->SetHWColorKey = NULL;
	device->SetHWAlpha = ACL_SetAlpha;
	device->LockHWSurface = ACL_LockHWSurface;
	device->UnlockHWSurface = ACL_UnlockHWSurface;
	device->FlipHWSurface = NULL;
	device->FreeHWSurface = ACL_FreeHWSurface;
	device->SetCaption = NULL;
	device->SetIcon = NULL;
	device->IconifyWindow = NULL;
	device->GrabInput = NULL;
	device->GetWMInfo = NULL;
	device->InitOSKeymap = ACL_InitOSKeymap;
	device->PumpEvents = ACL_PumpEvents;
    device->ZoomWindow = ACL_ZoomWindow;

	device->free = ACL_DeleteDevice;

    /* Set the driver flags */
    device->handles_any_size = 0;

	return device;
}

VideoBootStrap ACL_bootstrap = {
	ACLVID_DRIVER_NAME, "SDL ACL video driver",
	ACL_Available, ACL_CreateDevice
};

/*
 * bpp is Bytes per pixel 
 */
static void clear_buffer(_THIS, SDL_Surface *surface)
{
	struct SDL_PrivateVideoData *priv = this->hidden;
	int x, y;
	unsigned char *buf = priv->buffer;

	if (priv->rgbScaler == SDL_ACL_OVL_SCALAR_G0) {
		if (surface->format->BytesPerPixel == 4 &&
			surface->format->Amask != 0) {
			unsigned int *tmp;
			for (y = 0; y < surface->h; y++) {
				for (tmp = (unsigned int *)buf, x = 0; x < surface->w; x++, 
tmp++) {
					*tmp = surface->format->Amask;
				}
				buf += surface->pitch;
			}
		} else if (surface->format->BytesPerPixel == 2) {
			unsigned short *tmp;
			for (y = 0; y < surface->h; y++) {
				for (tmp = (unsigned short *)buf, x = 0; x < surface->w; x++
, tmp++) {
					*tmp = surface->format->Amask;
				}
				buf += surface->pitch;
			}
		}
	} else {
		memset(buf, 0, surface->pitch * surface->h);
	}
}

int ACL_VideoInit(_THIS, SDL_PixelFormat *vformat)
{
    int resCount;
    int i;
    SDL_Rect* mode;

	/* Determine the screen depth (use default 16-bit depth) */
	/* we change this during the SDL_SetVideoMode implementation... */
	vformat->BitsPerPixel = 16;
	vformat->BytesPerPixel = 2;

    vformat->Amask = 0x00000000;
    vformat->Rmask = 0x0000F800;
    vformat->Gmask = 0x000007E0;
    vformat->Bmask = 0x0000001F;

    /* generate 16bpp mode list */
    resCount = sizeof(ACL_modelist16) / sizeof(ACL_modelist16[0]);
    SDL_modelist[ACL_MODELIST_16] = malloc((1+resCount) * sizeof(SDL_Rect*));
    if ( !SDL_modelist[ACL_MODELIST_16] ) {
	SDL_OutOfMemory();
	return -1;
    }
    for( i = 0; i < resCount; i++ ) {
	mode = malloc(sizeof(SDL_Rect));
	if (!mode) {
	    SDL_OutOfMemory();
	    return -1;
	}
	mode->x = ACL_modelist16[i].x;
	mode->y = ACL_modelist16[i].y;
	mode->w = ACL_modelist16[i].w;
	mode->h = ACL_modelist16[i].h;

	DEBUG_PRINT("Adding 16bpp resolution: %d x %d\n", mode->w, mode->h);
	SDL_modelist[ACL_MODELIST_16][i] = mode;
    }
    SDL_modelist[ACL_MODELIST_16][resCount] = 0;

    /* generate 32bpp mode list */
    resCount = sizeof(ACL_modelist32) / sizeof(ACL_modelist32[0]);
    SDL_modelist[ACL_MODELIST_32] = malloc((1+resCount) * sizeof(SDL_Rect*));
    if ( !SDL_modelist[ACL_MODELIST_32] ) {
	SDL_OutOfMemory();

	return -1;
    }
    for( i = 0; i < resCount; i++ ) {
	mode = malloc(sizeof(SDL_Rect));
	if (!mode) {
	    SDL_OutOfMemory();
	    return -1;
	}
	mode->x = ACL_modelist32[i].x;
	mode->y = ACL_modelist32[i].y;
	mode->w = ACL_modelist32[i].w;
	mode->h = ACL_modelist32[i].h;

	DEBUG_PRINT("Adding 32bpp resolution: %d x %d\n", mode->w, mode->h);
	SDL_modelist[ACL_MODELIST_32][i] = mode;
    }
    SDL_modelist[ACL_MODELIST_32][resCount] = 0;
    mouse_fd = -1 ; 
    keyboard_fd = -1 ;  
    /* init the keyboard...NOT the mouse!! */ 

#if 0
    if ( ACL_OpenKeyboard(this) < 0 ) {
	ACL_VideoQuit(this);
	return(-1);
    }
#else
    ACL_OpenKeyboard(this);
#endif

	/* We're done! */
	return(0);
}

SDL_Rect **ACL_ListModes(_THIS, SDL_PixelFormat *format, Uint32 flags)
{
    if ( format->BitsPerPixel == 16 ) {
	return SDL_modelist[ACL_MODELIST_16];
    } else if (format->BitsPerPixel == 32 ) {
	return SDL_modelist[ACL_MODELIST_32];
    }
   	 
    return ((SDL_Rect **)0);
}


static void UndoInit(_THIS) {
	if (this->hidden->ovlhnd) {
		if (this->hidden->disphnd && this->hidden->connected) {
			AclOvl_Disconnect(this->hidden->ovlhnd, ACL_DISP,
					this->hidden->disphnd);
		}
		AclOvl_Close(this->hidden->ovlhnd);
	}

	if (this->hidden->disphnd) {
		AclDisp_Close(this->hidden->disphnd);
	}

	Acl_Close(this->hidden->chip);
	SDL_SetError("Couldn't initialize hardware");

	this->hidden->chip = 0 ;
	this->hidden->disphnd = 0 ;
	this->hidden->ovlhnd = 0 ;
}



SDL_Surface *ACL_SetVideoMode(_THIS, SDL_Surface *current,
				int width, int height, int bpp, Uint32 flags)
{
    UINT32 Amask, Rmask, Gmask, Bmask;
    ACL_EnterGraphicsMode(this) ; 

	if ( this->hidden->buffer != NULL ) {
#if 0
		memset(this->hidden->buffer, 0, current->pitch * current->h);
#else
		clear_buffer(this, current);
#endif
		AtiCore_VideoMemFree(ACL_CHIP_INDEX, this->hidden->buffer);
		this->hidden->buffer = NULL;
	}

    /* We only support 16 and 32 BPP modes */
    if ( bpp == 32 ) {
		if ( this->hidden->rgbScaler == SDL_ACL_OVL_SCALAR_G0) {
			Amask = 0xFF000000;
		} else {
			Amask = 0x0;
		}
		Rmask = 0x00FF0000;
		Gmask = 0x0000FF00;
		Bmask = 0x000000FF;
    } else {

		if ( this->hidden->rgbScaler == SDL_ACL_OVL_SCALAR_G0) {
			//for 1555 in G0
			Amask = 0x00008000;
			Rmask = 0x00007C00;
			Gmask = 0x000003E0;
			Bmask = 0x0000001F;
		} else {
			Amask = 0x00000000;
			Rmask = 0x0000F800;
			Gmask = 0x000007E0;
			Bmask = 0x0000001F;
		}
    }

	/* Allocate the new pixel format for the screen */
	if ( ! SDL_ReallocFormat(current, bpp, Rmask, Gmask, Bmask, Amask) ) {
		SDL_SetError("Couldn't allocate new pixel format for requested mode");
		return(NULL);
	}

	/* Set up the new mode framebuffer */
	this->hidden->w = current->w = width;
	this->hidden->h = current->h = height;
	current->flags = (SDL_FULLSCREEN|SDL_HWSURFACE);
    current->format->Rmask = Rmask;
    current->format->Gmask = Gmask;
    current->format->Bmask = Bmask;
    current->format->Amask = Amask;

    if (ACL_InitHW(this) != 0) {
	UndoInit(this) ;
	return(NULL);
    }

    if (ACL_VideoModeInit (this, current, width, height, bpp)) {
	if ( this->hidden->buffer != NULL ) {
	    AtiCore_VideoMemFree(ACL_CHIP_INDEX, this->hidden->buffer);
	    this->hidden->buffer = NULL;
	}
		UndoInit(this);  
		SDL_SetError("Couldn't initialize mode");
		return(NULL);	
    }

    DEBUG_PRINT("Created surface:\n");
    DEBUG_PRINT("   width: %d\n", current->w);
    DEBUG_PRINT("   height: %d\n", current->h);
    DEBUG_PRINT("   bpp: %d\n", current->format->BitsPerPixel);
    DEBUG_PRINT("   Bpp: %d\n", current->format->BytesPerPixel);
    DEBUG_PRINT("   Amask: %d\n", current->format->Amask);
    DEBUG_PRINT("   offset: %d\n", current->offset);
    DEBUG_PRINT("   pitch: %d\n", current->pitch);
    DEBUG_PRINT("   pixels: 0x%X\n", (unsigned int)current->pixels);
    DEBUG_PRINT("   flags: 0x%X\n", current->flags);
    DEBUG_PRINT("\n   chip hnd: 0x%X\n", (unsigned int)this->hidden->chip);
    DEBUG_PRINT("   disphnd: 0x%X\n", (unsigned int)this->hidden->disphnd);
    DEBUG_PRINT("   ovlhnd: 0x%X\n", (unsigned int)this->hidden->ovlhnd);

	/* we do this so that external functions can see us...
	 * we dont really  want to hide. 
	 */ 

	this->visible = current ; 

	this->visible->hwdata = &(this->hidden->hwdata) ; 
	this->visible->hwdata->info = this->hidden ; 
	/* We're done */
	return(current);
}

/* We don't actually allow hardware surfaces other than the main one */
static int ACL_AllocHWSurface(_THIS, SDL_Surface *surface)
{
	return(-1);
}
static void ACL_FreeHWSurface(_THIS, SDL_Surface *surface)
{
	return;
}

/* We need to wait for vertical retrace on page flipped displays */
static int ACL_LockHWSurface(_THIS, SDL_Surface *surface)
{
	return(0);
}

static void ACL_UnlockHWSurface(_THIS, SDL_Surface *surface)
{
	return;
}

static void ACL_UpdateRects(_THIS, int numrects, SDL_Rect *rects)
{
	/* do nothing. */
}

int ACL_SetColors(_THIS, int firstcolor, int ncolors, SDL_Color *colors)
{
	/* do nothing of note. */
	return(1);
}

/* Note:  If we are terminated, this could be called in the middle of
   another SDL video routine -- notably UpdateRects.
*/
void ACL_VideoQuit(_THIS)
{
    int resCount;
    int i;
  
//    fprintf(stderr, "ACL_VideoQuit\n"); fsync(STDERR_FILENO); 
    /* Create the supported resolution mode tables */
    resCount = sizeof(ACL_modelist16) / sizeof(ACL_modelist16[0]);
    for( i = 0; i < resCount; i++ ) {
	free(SDL_modelist[ACL_MODELIST_16][i]);
    }
    free(SDL_modelist[ACL_MODELIST_16]);

    resCount = sizeof(ACL_modelist32) / sizeof(ACL_modelist32[0]);
    for( i = 0; i < resCount; i++ ) {
	free(SDL_modelist[ACL_MODELIST_32][i]);
    }
    free(SDL_modelist[ACL_MODELIST_32]);
    
	if (this->hidden->buffer != NULL)
	{
		AtiCore_VideoMemFree(ACL_CHIP_INDEX, this->hidden->buffer);
		this->hidden->buffer = NULL;
		this->screen->pixels = NULL;
	}


    ACL_CloseKeyboard(this);

    // We don't worry about errors from the following shutdown
    // routines, they are siliently ignored.
    if (this->hidden->ovlhnd) {
	// DONT disable overlay. Because the target board has
        // X225 and disable overlay on exit, you will get to see magenta
        // raster image for a while even if you are using dispd.
    	// AclOvl_Enable(this->hidden->ovlhnd, FALSE);

	if (this->hidden->disphnd && this->hidden->connected) {
	    AclOvl_Disconnect(this->hidden->ovlhnd, ACL_DISP, 
			      this->hidden->disphnd);
	}

//  	fprintf(stderr, "ACL_VideoQuit:: Closing Overlay\n"); 
	AclOvl_Close(this->hidden->ovlhnd); fsync(STDERR_FILENO);
    }

    if (this->hidden->disphnd) {
//  	fprintf(stderr, "ACL_VideoQuit:: Closing Display\n"); 
	AclDisp_Close(this->hidden->disphnd); fsync(STDERR_FILENO);
    }

//    fprintf(stderr, "ACL_VideoQuit:: Calling Acl_Close\n"); fsync(STDERR_FILENO); 
    if (this->hidden->chip) {
        Acl_Close(this->hidden->chip);    
    }
}


static int
ACL_VideoModeInit (_THIS, SDL_Surface* current, int width, int height, int bpp)
{
    ACL_RESULT rc;
    ACL_HANDLE ovlhnd = this->hidden->ovlhnd;
    UINT32 channel = MEM_CHAN_0_1;
    UINT32 pitch = 0;
    UINT32 alignment = 0;
    UINT32 datafmt;
    UINT32 size;
    SDL_Rect zoomRect;

    /* We only support 16 and 32 BPP modes */
    if ( bpp == 32 ) {
	if (this->hidden->rgbScaler == SDL_ACL_OVL_SCALAR_G0)  
	    datafmt = ACL_DATA_FORMAT_ARGB8888; 
	else if ((this->hidden->rgbScaler == SDL_ACL_OVL_SCALAR_V0) ||
		 (this->hidden->rgbScaler == SDL_ACL_OVL_SCALAR_V1)
		)
	    datafmt = ACL_DATA_FORMAT_ARGB8888_32BPP;
	else 
	    return -1 ; 
	current->format->BitsPerPixel = 32;
	current->format->BytesPerPixel = 4;
    } else  if ( bpp == 16 ) {
	if (this->hidden->rgbScaler == SDL_ACL_OVL_SCALAR_G0)
	    datafmt = ACL_DATA_FORMAT_ARGB1555;
	else if ((this->hidden->rgbScaler == SDL_ACL_OVL_SCALAR_V0) ||
		 (this->hidden->rgbScaler == SDL_ACL_OVL_SCALAR_V1) 
		)
	    datafmt = ACL_DATA_FORMAT_RGB565;
	else
	    return -1 ;  
	current->format->BitsPerPixel = 16;
	current->format->BytesPerPixel = 2;
    } else {	
	return -1;
    }

    /* Set source format */
    rc = AclOvl_SetSrcFormat (ovlhnd, datafmt, &alignment);
    if (ACL_ERROR (rc))
    {
	DEBUG_PRINT ("\nError: Unable to set graphics overlay source format\n");
	return (-1);
    }
    DEBUG_PRINT ("*** Video Mem Align: 0x%X\n", alignment);

    /* Set source window size */
    rc = AclOvl_SetSrcWindowSize (ovlhnd, width, height, &pitch);
    if (ACL_ERROR (rc))
    {
	DEBUG_PRINT ("\nError: Unable to set graphics overlay source window\n");
	return (-1);
    }
    DEBUG_PRINT ("\n*** Video Mem Pitch: 0x%X\n", pitch);

    /* Allocate video memory */
    current->pitch = pitch * current->format->BytesPerPixel;
    size = current->pitch * height;
    this->hidden->pitch = current->pitch;

    this->hidden->buffer = 
	AtiCore_VideoMemAlloc(ACL_CHIP_INDEX, 
			      &channel,
			      size, 
			      alignment * current->format->BytesPerPixel, 
			      &this->hidden->offset, 
			      ALLOC_TYPE_NORM_GUI);

	if ( ! this->hidden->buffer ) {
		SDL_SetError("Couldn't allocate buffer for requested mode");
		return(-1);
	}
	current->pixels = this->hidden->buffer;
    DEBUG_PRINT ("\n*** Video Mem Offset: 0x%X\n", this->hidden->offset);

    /* Clear the allocated memory */
#if 0
    memset(this->hidden->buffer, 0, size);
#else
	clear_buffer(this, current);
#endif

    /* Set the source offset */
    rc = AclOvl_SetSrcOffset (ovlhnd, this->hidden->offset, FALSE);
    if (ACL_ERROR (rc))
    {
	DEBUG_PRINT("\nError: Unable to set graphics overlay source offset\n");
	return (-1);
    }

    /* set the alpha mode of the graphics overlay (G0)*/ 
    if ( this->hidden->rgbScaler == SDL_ACL_OVL_SCALAR_G0) {
	    rc = AclOvl_SetAlpha(ovlhnd, ACL_OVL_ALPHA_MULTIPLY,  ACL_OVL_ALPHA_PIXEL, 255) ;
	    if (ACL_ERROR (rc))
	    {
		DEBUG_PRINT("\nError: Unable to do alpha in the G0\n");
		return (-1);
	    }
    }

	ACL_PreventPink();

    /* Make sure the scaler crop is set to full screen */
    zoomRect.x = 0;
    zoomRect.y = 0;
    zoomRect.w = ACL_OUT_WIDTH;
    zoomRect.h = ACL_OUT_HEIGHT;
    ACL_ZoomWindow(this, &zoomRect);

    /* Enable the graphics overlay */    
    rc = AclOvl_Enable (ovlhnd, TRUE);
    if (ACL_ERROR (rc))
    {
	DEBUG_PRINT("\nError: Unable to enable graphics overlay\n");
	return (-1);
    }

    return 0;
}

static int
ACL_InitHW(_THIS)
{
    BOOL dispExist, vidExist;
    ACL_HANDLE chip = 0;
    ACL_HANDLE disphnd = 0;
    ACL_HANDLE ovlhnd = 0;
    ACL_RESULT rc;
    ACL_RECT pWin;

    /* Open first HDTV2 chip only for this example */
    rc = Acl_Open (ACL_CHIP_INDEX, &chip);
    if (ACL_ERROR (rc))
    {
	DEBUG_PRINT("\nError: Unable to open ACL\n");
	return (-1);
    }
    this->hidden->chip = chip;

    // Turn on ACL debugging messages when in debug mode
#ifndef NDEBUG
    rc = Acl_SetDebugLevel(chip, ACL_COMPONENT_MAX,
			   ACL_DEBUGLEVEL_Verbose, NULL);
#else
    rc = Acl_SetDebugLevel(chip, ACL_COMPONENT_MAX,
			   ACL_DEBUGLEVEL_None, NULL);
#endif

    /* Retrieve primary display handle associated with this chip */
    rc = AclDisp_Open (&disphnd, chip, ACL_DISPLAY_0, ACL_MODE_SHARED, &dispExist);
    if (ACL_ERROR (rc))
    {
	DEBUG_PRINT("\nError: Unable to open Display object\n");
	return (-1);
    }

    /* Save the new display handle */
    this->hidden->disphnd = disphnd;

    /* Set a display mode */
    if (!dispExist) 
    {
	rc = AclDisp_SetMode (disphnd, ACL_OUT_FORMAT);
	if (ACL_ERROR (rc))
	{
	    DEBUG_PRINT("\nError: Unable to set mode\n");
	    return (-1);
	}
    }

    /* Retrieve main graphics overlay handle associated with this chip */
    if ( this->hidden->rgbScaler == SDL_ACL_OVL_SCALAR_G0) {
	rc = AclOvl_Open (&ovlhnd, chip, ACL_OVL_G0SCALE, 
			  ACL_MODE_SHARED, &vidExist);
	DEBUG_PRINT("main video surface is G0\n") ; 
	ACL_DisableOverlay(this, ACL_OVL_V0SCALE);
	ACL_DisableOverlay(this, ACL_OVL_V1SCALE);

    } else if ( this->hidden->rgbScaler == SDL_ACL_OVL_SCALAR_V0) {
	rc = AclOvl_Open (&ovlhnd, chip, ACL_OVL_V0SCALE, 
			  ACL_MODE_SHARED, &vidExist);
	DEBUG_PRINT("main video surface is V0\n") ; 
	ACL_DisableOverlay(this, ACL_OVL_G0SCALE);
	ACL_DisableOverlay(this, ACL_OVL_V1SCALE);

    } else if ( this->hidden->rgbScaler == SDL_ACL_OVL_SCALAR_V1) {
	rc = AclOvl_Open (&ovlhnd, chip, ACL_OVL_V1SCALE, 
			  ACL_MODE_SHARED, &vidExist); 
	DEBUG_PRINT("main video surface is V1\n") ; 
	ACL_DisableOverlay(this, ACL_OVL_G0SCALE);
	ACL_DisableOverlay(this, ACL_OVL_V0SCALE);

    } else {
	DEBUG_PRINT("%s(%s): Unknown scaler requested\n",
		    __FILE__,__LINE__); 
	return (-1);
    }

    if (ACL_ERROR (rc))
    {
	DEBUG_PRINT("\nError: Unable to open graphics overlay object\n");
	return (-1);
    }

    /* Save the new overlay handle */
    this->hidden->ovlhnd = ovlhnd;

    /* Connect to the display instance */
    this->hidden->connected = FALSE;

    if (!vidExist /* && !dispExist */ ) {
	DEBUG_PRINT("Warning:  going to do overlay connection with display!!\n");
	rc = AclOvl_Connect (ovlhnd, ACL_DISP, disphnd);
	if (ACL_ERROR (rc))
	{
	    DEBUG_PRINT("\nError: Unable to connect graphics overlay\n");
	    return (-1);
	}

	this->hidden->connected = TRUE;

	/* Set a graphics overlay destination window */
	pWin.Left = 0;
	pWin.Top = 0;
	pWin.Width = ACL_OUT_WIDTH;
	pWin.Height = ACL_OUT_HEIGHT;
	rc = AclOvl_SetDstWindow (ovlhnd, &pWin, FALSE);
	if (ACL_ERROR (rc))
	{
	  DEBUG_PRINT("\nError: Unable to set graphics overlay window\n");
	  return (-1);
	}

    }
    return 0;
}

static int ACL_CheckHWBlit(_THIS, SDL_Surface *src, SDL_Surface *dst) 
{
    /* Set initial acceleration on */

    src->flags |= SDL_HWACCEL;
    DEBUG_PRINT("made it into ACL_CheckHWBlit!!\n"); 
	return 0 ; 
}

static int ACL_SetAlpha(_THIS, SDL_Surface *surface, Uint8 value)
{
	fprintf(stderr, "calling SetAlpha\n");
	if (value == SDL_ALPHA_OPAQUE) {
		AclOvl_SetAlpha(this->hidden->ovlhnd, 
			ACL_OVL_ALPHA_MULTIPLY, 
			ACL_OVL_ALPHA_PIXEL, 255);
	}  else {
		AclOvl_SetAlpha(this->hidden->ovlhnd, 
			ACL_OVL_ALPHA_MULTIPLY, 
			ACL_OVL_ALPHA_GLOBAL, value) ;
	}
	return 0 ; 
}


/* If win is NULL, reset zoom to full screen, otherwise zoom in on the
 * rectangle of the display window
 */
static void ACL_ZoomWindow(_THIS, SDL_Rect* win)
{
    ACL_RECT srcWin;

    if ( win ) {
        srcWin.Left   = win->x;
        srcWin.Top    = win->y;
        srcWin.Width  = win->w;
        srcWin.Height = win->h;
    } else {
        srcWin.Left   = 0;
        srcWin.Top    = 0;
        srcWin.Width  = ACL_OUT_WIDTH;
        srcWin.Height = ACL_OUT_HEIGHT;
    }

    AclOvl_SetCropWindow(this->hidden->ovlhnd,
                         ACL_OVL_CROP_SRC,
                         &srcWin);
}
