/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2004. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include "lki_key.h"

#include "SDL_error.h"
#include "SDL_events.h"
#include "SDL_thread.h"
#include "SDL_timer.h"
#include "SDL_e21.h"


static SDL_Thread* s_pThread = NULL;

static int s_iActiveEventLoop = 0;

#if !defined(NDEBUG)
#define ___DEBUG___
#endif

static int g_keypressed;
static int g_count;
static SDL_Event g_pushedevent;

#define REPEAT_DELAY	2
#define EVENT_TIMEOUT	200000		


static SDLKey ConvertKeyPW2SDL(int iScanCode)
{
	switch (iScanCode) {
	case KEY_R_ENTER:	return SDLK_RETURN;
	case KEY_R_UP:		return SDLK_UP;
	case KEY_R_DOWN:	return SDLK_DOWN;
	case KEY_R_LEFT:	return SDLK_LEFT;
	case KEY_R_RIGHT:	return SDLK_RIGHT;
	case KEY_R_ESC:		return SDLK_ESCAPE;
	case KEY_R_PAGEUP:	return SDLK_PAGEUP;
	case KEY_R_PAGEDOWN:return SDLK_PAGEDOWN;
	default:			return SDLK_UNKNOWN;
	}
}

static int IsRepeatableKey(SDLKey key)
{
	switch (key) {
	case SDLK_UP:
	case SDLK_DOWN:
	case SDLK_LEFT:
	case SDLK_RIGHT:
		return 1;
	default:
		break;
	}
		
	return 0;
}

inline static void
___get_parms(int parms[4], const char *buf)
{
     int i;
	 char *cur, *end;
	 unsigned long tmp;

    cur = (char *)buf;
    for (i = 0; i < 4 && cur && cur[0]; i++) {
		tmp = strtoul(cur, &end, 16);
        if (tmp == ULONG_MAX)
            continue;

        parms[i] = tmp;

		if (end && !end[0])
			return;
        cur = end + 1;
    }

	cur = (char *)buf;
}

static inline int
get_params(int parms[4], const char *buf, size_t size)
{
	 size_t len;
	for (len = 0; len < size; len++) {	
		if (buf[len] != '=')
			continue;

		len++;
		___get_parms(parms, &buf[len]);
		return 0;
	}
	return 0;
}


static int
SetEvent(SDL_Event *ev, const int parms[4])
{
	 SDL_E21Event *imev;
	switch(parms[0]){
	case EV_KEYMAKE: 
		ev->type      = SDL_KEYDOWN;
		ev->key.state = SDL_PRESSED;
		ev->key.keysym.sym = ConvertKeyPW2SDL(parms[1]);
		g_keypressed = 1;
		g_pushedevent = *ev;
		break;

	case EV_KEYBREAK: 
		ev->type      = SDL_KEYUP;
		ev->key.state = SDL_RELEASED;
		ev->key.keysym.sym = ConvertKeyPW2SDL(parms[1]);
		g_keypressed = 0;
		g_count = 0;
		break;

	default: 
		imev = (SDL_E21Event *)calloc(sizeof(SDL_E21Event), 1);
		if (imev == NULL)
			return -1;

		ev->type		= SDL_USEREVENT;
		ev->user.code 	= SDL_E21_MAGIC;
		ev->user.data1	= (void *)imev;
		ev->user.data2	= free;
		imev->code		= parms[0];
		imev->param[0]	= parms[1];
		imev->param[1]	= parms[2];
		imev->param[2]	= parms[3];
		break;
	}
	return 0;
}

static int EventLoop(void *unused)
{
	char *buf;
	SDL_Event stEvent;
	SDL_E21Event *pEvent;
	int len, parms[4];

	buf = NULL;
	pEvent = NULL;

	while (s_iActiveEventLoop) {
		if (buf) free(buf);
		buf = NULL;

		
		len = LKI_ReceiveEventWait(&buf, 0, EVENT_TIMEOUT);
		if(len <= 2){
			
			if (len > 0) {
				 int i;
				fprintf(stderr, "SDLe21: garbage:");
				for (i = 0; i < len; i++) {
					fprintf(stderr, "%02X ", buf[i] & 0xff);
				}
				fprintf(stderr, "\n");
			}

			if (g_keypressed && (g_count++ > REPEAT_DELAY) && 
				IsRepeatableKey(g_pushedevent.key.keysym.sym)) {

				 SDL_Event breakEvent = g_pushedevent;
				breakEvent.type = SDL_KEYUP;
				breakEvent.key.state = SDL_RELEASED;
				
				SDL_PushEvent(&breakEvent);
				SDL_PushEvent(&g_pushedevent);
			}
			continue;
		}

		if (buf[0] == ESCVP21_STATUS_SUCCESS)
			continue;

		if (strncmp(buf, ESCVP21_STATUS_ERR, strlen(ESCVP21_STATUS_ERR)) == 0)
			continue;

		len -= 2;
		buf[len] = '\0';

		memset(parms, 0, sizeof(parms));
		get_params(parms, buf, len); 

		memset(&stEvent,  0, sizeof(SDL_Event));
		if (SetEvent(&stEvent, parms) == 0)
			SDL_PushEvent(&stEvent);

	}

	return 0;
}


static int StartEventThread(void)
{
	s_iActiveEventLoop = 1;

	s_pThread = SDL_CreateThread(EventLoop, NULL);
	if(s_pThread == NULL){
		return -1;
	}

	return 0;
}



static int StopEventThread(void)
{
	int iStatus;

	s_iActiveEventLoop = 0;

	if (s_pThread) {
		SDL_WaitThread(s_pThread, &iStatus);
		s_pThread = NULL;
	}

	return iStatus;
}


static int E21Open(char* pAddr, Uint32 uiEvent)
{
	int iRv = 0;
	int i;

	
	for (i = 0; (i < 10) && (iRv = LKI_OpenSocket(pAddr)) != LKI_SUCCESS; i++) {
		usleep(500000);
	}
	if (i == 10) {
		goto ERRRET_OPEN;
	}

	
	iRv = LKI_SetAcceptableEvent(uiEvent);
	if(iRv < 0){
		goto ERRRET_OPEN;
	}

	
	iRv = StartEventThread();
	if(iRv < 0){
		goto ERRRET_OPEN;
	}

	return 0;

ERRRET_OPEN:
	
	LKI_CloseSocket();
	return iRv;
}


static void E21Close(void)
{
	
	SDL_Lock_EventThread();

	
	StopEventThread();

	
	SDL_Unlock_EventThread();

	
	LKI_CloseSocket();
}






int SDL_E21Init(void)
{
	return E21Open(NULL, RECEIVE_EVENT_ID);
}


void SDL_E21Quit(void)
{
	E21Close();
}

