/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2004. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <assert.h>
#include <setjmp.h>
#include <jpeglib.h>

#include "SDL_jpeg.h"



struct my_error_mgr {
	struct jpeg_error_mgr pub;	
	jmp_buf setjmp_buffer;		
};
typedef struct my_error_mgr *my_error_ptr;


METHODDEF(void) my_error_exit(j_common_ptr cinfo)
{
	
	my_error_ptr myerr = (my_error_ptr) cinfo->err;

	
	longjmp(myerr->setjmp_buffer, 1);
}



static SDL_bool SDL_IsJPG(FILE * fp)
{
	const unsigned char SOI[] = { 0xFF, 0xD8 };

	size_t ret;
	unsigned char header[2];

	assert(fp);

	
	ret = fread(header, 1, 2, fp);
	if (ret != 2 || header[0] != SOI[0] || header[1] != SOI[1]) {
		return SDL_FALSE;
	}

	rewind(fp);

	return SDL_TRUE;
}


int SDL_JPGDimension(const char *file, int *width, int *height)
{
	struct jpeg_decompress_struct cinfo;
	struct my_error_mgr jerr;
	FILE *fp;

	assert(file);
	assert(width);
	assert(height);

	fp = fopen(file, "rb");
	if (fp == NULL) {
		return -1;
	}
	
	if (!SDL_IsJPG(fp)) {
		fclose(fp);
		return -1;
	}

	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = my_error_exit;

	if (setjmp(jerr.setjmp_buffer)) {
		
		jpeg_destroy_decompress(&cinfo);
		fclose(fp);
		return -1;
	}

	
	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, fp);

	jpeg_read_header(&cinfo, TRUE);

	jpeg_calc_output_dimensions(&cinfo);

	*width = cinfo.output_width;
	*height = cinfo.output_height;

	
	jpeg_destroy_decompress(&cinfo);

	fclose(fp);

	return 0;
}





SDL_Surface *SDL_LoadJPG(const char *file, SDL_JpegScale scale)
{
	struct jpeg_decompress_struct cinfo;
	struct my_error_mgr jerr;
	JSAMPROW srow;
	SDL_Surface *surface = NULL;
	Uint32 rmask, gmask, bmask, amask;
	FILE *fp;
	int i;

	assert(file);

	fp = fopen(file, "rb");
	if (fp == NULL) {
		return NULL;
	}
	
	if (!SDL_IsJPG(fp)) {
		fclose(fp);
		return NULL;
	}
	

	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = my_error_exit;

	if (setjmp(jerr.setjmp_buffer)) {
		
		jpeg_destroy_decompress(&cinfo);
		fclose(fp);

		
		if (surface) {
			if (surface->locked) {
				SDL_UnlockSurface(surface);
			}
			SDL_FreeSurface(surface);
		}
		return NULL;
	}

	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, fp);
	jpeg_read_header(&cinfo, TRUE);

	
	
	if (cinfo.progressive_mode) {
		jpeg_destroy_decompress(&cinfo);
		fclose(fp);
		return NULL;
	}
	
	cinfo.out_color_space = JCS_RGB;
#if 0
	cinfo.dct_method = JDCT_ISLOW;
#else
	cinfo.dct_method = JDCT_FASTEST;
#endif

	
	if (scale >= SDL_JPEG_SCALE_HALF && scale <= SDL_JPEG_SCALE_EIGHTH) {
		cinfo.scale_num = 1;
		cinfo.scale_denom = (unsigned int) scale;
	}

	jpeg_calc_output_dimensions(&cinfo);

#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	rmask = 0xff000000;
	gmask = 0x00ff0000;
	bmask = 0x0000ff00;
	amask = 0x00000000;
#else
	rmask = 0x000000ff;
	gmask = 0x0000ff00;
	bmask = 0x00ff0000;
	amask = 0x00000000;
#endif

	
	surface = SDL_CreateRGBSurface(SDL_SWSURFACE, cinfo.output_width, cinfo.output_height, 24,	
								   rmask, gmask, bmask, amask);
	if (surface == NULL) {
		jpeg_destroy_decompress(&cinfo);
		fclose(fp);
		return NULL;
	}
	
	jpeg_start_decompress(&cinfo);

	
	if (SDL_MUSTLOCK(surface)) {
		SDL_LockSurface(surface);
	}
	
	srow = (JSAMPLE *) surface->pixels;
	for (i = 0; i < cinfo.output_height; i++) {
		jpeg_read_scanlines(&cinfo, &srow, 1);
#ifdef WEIGHT_FOR_LOOP 
		if (!(i & 0xf)) usleep(0);
#endif
		srow += surface->pitch;
	}

	
	if (SDL_MUSTLOCK(surface)) {
		SDL_UnlockSurface(surface);
	}
	
	

	
	jpeg_destroy_decompress(&cinfo);
	fclose(fp);

	return surface;
}


#define NEAR_PRECISION  1024
#define NEAR_SHIFT      10
#define NEAR_MASK       0x3FF



static inline Uint32 roundFixedPoint(Uint32 in)
{
	Uint32 rem = in % NEAR_PRECISION;

	if (rem < (NEAR_PRECISION >> 1))
		return (in - rem);
	else
		return (in - rem + NEAR_PRECISION);
}



static void
doCalcNearestJumpVector(Uint32 * samples, Uint32 srcSpan, Uint32 dstSpan)
{
	int i;
	Uint32 ratio;
	Uint32 posOld, posNew;

	assert(samples != NULL);
	assert(srcSpan > 0);
	assert(dstSpan > 0);

	if (dstSpan < 2 || srcSpan < 2) {
		ratio = 0;
	} else {
		ratio = (srcSpan - 1) << NEAR_SHIFT;
		ratio /= (dstSpan - 1);

	}

	
	
	samples[0] = 0;
	samples[dstSpan - 1] = srcSpan - 1;

	
	for (i = 1; i < (dstSpan - 1); i++) {
		samples[i] = roundFixedPoint(ratio * i) >> NEAR_SHIFT;
	}

	posOld = samples[0];
	for (i = 1; i < dstSpan; i++) {
		posNew = samples[i] - posOld;
		posOld = samples[i];
		samples[i] = posNew;
	}



}


#define CRACK_PRECISION  256
#define CRACK_SHIFT      8
#define CRACK_MASK       0xFF


static void
doCalcBiLinJumpVector(Uint32 * samples, Uint32 * cracks,
					  Uint32 srcSpan, Uint32 dstSpan)
{
	int i;
	Uint32 ratio;
	Uint32 posOld, posNew;
	assert(samples != NULL);
	assert(cracks != NULL);
	assert(srcSpan > 0);
	assert(dstSpan > 0);


	if (dstSpan < 2 || srcSpan < 2) {
		ratio = 0;
	} else {
		ratio = ((srcSpan - 1) << CRACK_SHIFT) / (dstSpan);
	}

	
	samples[0] = 0;
	cracks[0] = 0;

	
	for (i = 1; i < dstSpan; i++) {
		samples[i] = (ratio * i) >> CRACK_SHIFT;
		cracks[i] = (ratio * i) & CRACK_MASK;
	}
	
	

	i--;
	assert(samples[i] <= srcSpan - 1);

	while (samples[i] == srcSpan - 1) {
		assert(i > 0);
		samples[i]--;
		cracks[i] = CRACK_PRECISION - 1;
		i--;
	}



	posOld = samples[0];
	for (i = 1; i < dstSpan; i++) {
		posNew = samples[i] - posOld;
		posOld = samples[i];
		samples[i] = posNew;
	}

	
}







SDL_Surface *SDL_LoadJPGLowMem(const char *file,
							   SDL_Rect * dstrect,
							   SDL_RotateAngle angle,
							   SDL_StretchMode mode, SDL_bool keepRatio)
{
	struct jpeg_decompress_struct cinfo;
	struct my_error_mgr jerr;
	JSAMPROW srow;
	SDL_Surface *surface = NULL;
	Uint32 rmask, gmask, bmask, amask;
	FILE *fp;
	int i;
	SDL_Rect thinking, rotated;

	int row, col;
	int maxWidth, maxHeight;
	JSAMPROW srcPixelRowZero;
	JSAMPROW srcPixelRowOne;
	Uint8 *srcPixel00;
	Uint8 *srcPixel01;
	Uint8 *srcPixel10;
	Uint8 *srcPixel11;
	Uint8 *dstPixel;
	Uint32 dstVal;

	assert(file);

	fp = fopen(file, "rb");
	if (fp == NULL) {
		return NULL;
	}
	
	if (!SDL_IsJPG(fp)) {
		fclose(fp);
		return NULL;
	}
	

	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = my_error_exit;

	if (setjmp(jerr.setjmp_buffer)) {
		
		jpeg_destroy_decompress(&cinfo);
		fclose(fp);
		
		if (surface) {
			if (surface->locked) {
				SDL_UnlockSurface(surface);
			}
			SDL_FreeSurface(surface);
		}
		return NULL;
	}

	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, fp);
	jpeg_read_header(&cinfo, TRUE);

	
	
	if (cinfo.progressive_mode) {
		jpeg_destroy_decompress(&cinfo);
		fclose(fp);
		return NULL;
	}
	
	cinfo.out_color_space = JCS_RGB;
	cinfo.dct_method = JDCT_ISLOW;

	jpeg_calc_output_dimensions(&cinfo);


#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	rmask = 0xff000000;
	gmask = 0x00ff0000;
	bmask = 0x0000ff00;
	amask = 0x00000000;
#else
	rmask = 0x000000ff;
	gmask = 0x0000ff00;
	bmask = 0x00ff0000;
	amask = 0x00000000;
#endif

	thinking.x = 0;
	thinking.y = 0;
	thinking.w = cinfo.output_width;
	thinking.h = cinfo.output_height;


	if (angle == SDL_ROTATE_90 || angle == SDL_ROTATE_270) {
		rotated.w = thinking.h;
		rotated.h = thinking.w;
	} else {
		rotated.w = thinking.w;
		rotated.h = thinking.h;
	}

	if (dstrect == NULL) {
		
		
	} else if (((dstrect->w == rotated.w) &&
				(dstrect->h == rotated.h)) ||
			   (dstrect->w == 0 || dstrect->h == 0)
		) {
		
		
	} else if (keepRatio == SDL_TRUE) {
		if (((float) (rotated.w) / (float) (rotated.h)) >
			((float) (dstrect->w) / (float) (dstrect->h))) {
			thinking.h = (dstrect->w * rotated.h) / rotated.w;
			thinking.w = dstrect->w;
		} else {
			thinking.w = (dstrect->h * rotated.w) / rotated.h;
			thinking.h = dstrect->h;
		}

		if (angle == SDL_ROTATE_90 || angle == SDL_ROTATE_270) {
			Uint16 swp = thinking.w;
			thinking.w = thinking.h;
			thinking.h = swp;
		}
	} else {
		if (angle == SDL_ROTATE_90 || angle == SDL_ROTATE_270) {
			thinking.w = dstrect->h;
			thinking.h = dstrect->w;
		} else {
			thinking.w = dstrect->w;
			thinking.h = dstrect->h;
		}
	}


	if (dstrect != NULL) {		

		unsigned int scale = 1;
		float wRatio = 1.0;
		unsigned int hscale = 1;
		float hRatio = 1.0;

		wRatio = ((float) cinfo.output_width / (float) thinking.w);
		if (wRatio < 2.0)
			scale = 1;
		else if (wRatio < 4.0)
			scale = 2;
		else if (wRatio < 8.0)
			scale = 4;
		else
			scale = 8;

		hRatio = ((float) cinfo.output_height / (float) thinking.h);
		if (hRatio < 2.0)
			hscale = 1;
		else if (hRatio < 4.0)
			hscale = 2;
		else if (hRatio < 8.0)
			hscale = 4;
		else
			hscale = 8;
		
		cinfo.scale_num = 1;

		if (scale < hscale) {
			cinfo.scale_denom = (unsigned int) scale;
		} else {
			cinfo.scale_denom = (unsigned int) hscale;
		}
		jpeg_calc_output_dimensions(&cinfo);
	}




	surface = SDL_CreateRGBSurface(SDL_SWSURFACE, thinking.w, thinking.h, 24,	
								   rmask, gmask, bmask, amask);


	if (surface == NULL) {
		jpeg_destroy_decompress(&cinfo);
		fclose(fp);
		return NULL;
	}
	
	jpeg_start_decompress(&cinfo);

	
	if (SDL_MUSTLOCK(surface)) {
		SDL_LockSurface(surface);
	}
	
	if ((thinking.w == cinfo.output_width)
		&& (thinking.h == cinfo.output_height)
		) {
		
		srow = (JSAMPLE *) surface->pixels;
		for (i = 0; i < cinfo.output_height; i++) {
			jpeg_read_scanlines(&cinfo, &srow, 1);
#ifdef WEIGHT_FOR_LOOP
			if (!(i & 0xf)) usleep(0);
#endif
			srow += surface->pitch;
		}
	}
	
	else if ((mode == SDL_STRETCH_NEAREST) ||
			 (cinfo.output_width * cinfo.output_height < 100)

		) {

		
		Uint32 *widthSamples =
			(Uint32 *) malloc(thinking.w * sizeof(Uint32));
		Uint32 *heightSamples =
			(Uint32 *) malloc(thinking.h * sizeof(Uint32));
		srcPixelRowZero =
			malloc(cinfo.output_width * cinfo.output_components);

		if ((widthSamples == NULL) || (heightSamples == NULL) ||
			(srcPixelRowZero == NULL)) {
			if (widthSamples != NULL)
				free(widthSamples);
			if (heightSamples != NULL)
				free(heightSamples);
			if (srcPixelRowZero != NULL)
				free(srcPixelRowZero);
			
			if (SDL_MUSTLOCK(surface)) {
				SDL_UnlockSurface(surface);
			}
			
			jpeg_destroy_decompress(&cinfo);
			fclose(fp);
			return NULL;
		}
		
		doCalcNearestJumpVector(widthSamples, cinfo.output_width,
								thinking.w);
		doCalcNearestJumpVector(heightSamples, cinfo.output_height,
								thinking.h);

		maxWidth = thinking.w;
		maxHeight = thinking.h;
		
		jpeg_read_scanlines(&cinfo, &srcPixelRowZero, 1);

		for (row = 0; row < maxHeight; row++) {
			Uint32 dump;
			for (dump = 0; dump < heightSamples[row]; dump++) {
				jpeg_read_scanlines(&cinfo, &srcPixelRowZero, 1);
			}
			dstPixel = (Uint8 *) surface->pixels + row * surface->pitch;
			srcPixel00 = srcPixelRowZero;

			for (col = 0; col < maxWidth; col++) {
				srcPixel00 += widthSamples[col] * cinfo.output_components;
				memcpy(dstPixel, srcPixel00, cinfo.output_components);
				dstPixel += cinfo.output_components;
			}					
		}						

		if (widthSamples != NULL)
			free(widthSamples);
		if (heightSamples != NULL)
			free(heightSamples);
		if (srcPixelRowZero != NULL)
			free(srcPixelRowZero);
		
		if (SDL_MUSTLOCK(surface)) {
			SDL_UnlockSurface(surface);
		}
	}


	else if (mode == SDL_STRETCH_BILINEAR) {
		
		Uint32 *widthSamples =
			(Uint32 *) malloc(thinking.w * sizeof(Uint32));
		Uint32 *heightSamples =
			(Uint32 *) malloc(thinking.h * sizeof(Uint32));
		Uint32 *widthCracks =
			(Uint32 *) malloc(thinking.w * sizeof(Uint32));
		Uint32 *heightCracks =
			(Uint32 *) malloc(thinking.h * sizeof(Uint32));
		srcPixelRowZero =
			malloc(cinfo.output_width * cinfo.output_components);
		srcPixelRowOne =
			malloc(cinfo.output_width * cinfo.output_components);

		if ((widthSamples == NULL) || (heightSamples == NULL) ||
			(widthCracks == NULL) || (heightCracks == NULL) ||
			(srcPixelRowZero == NULL)
			|| (srcPixelRowOne == NULL)) {
			if (widthSamples != NULL)
				free(widthSamples);
			if (heightSamples != NULL)
				free(heightSamples);
			if (widthCracks != NULL)
				free(widthCracks);
			if (heightCracks != NULL)
				free(heightCracks);
			if (srcPixelRowZero != NULL)
				free(srcPixelRowZero);
			if (srcPixelRowOne != NULL)
				free(srcPixelRowOne);
			
			if (SDL_MUSTLOCK(surface)) {
				SDL_UnlockSurface(surface);
			}
			
			jpeg_destroy_decompress(&cinfo);
			fclose(fp);
			return NULL;
		}
		
		doCalcBiLinJumpVector(widthSamples, widthCracks,
							  cinfo.output_width, thinking.w);
		doCalcBiLinJumpVector(heightSamples, heightCracks,
							  cinfo.output_height, thinking.h);

		maxWidth = thinking.w;
		maxHeight = thinking.h;

		
		jpeg_read_scanlines(&cinfo, &srcPixelRowZero, 1);
		jpeg_read_scanlines(&cinfo, &srcPixelRowOne, 1);
		
		
		for (row = 0; row < maxHeight; row++) {
			
			if (heightSamples[row] == 0) {
				
			} else if (heightSamples[row] == 1) {
				
				
				JSAMPROW swap = srcPixelRowZero;
				srcPixelRowZero = srcPixelRowOne;
				srcPixelRowOne = swap;
				jpeg_read_scanlines(&cinfo, &srcPixelRowOne, 1);
			} else {
				
				
				int dump = 0;
				for (; dump < heightSamples[row] - 1; dump++)
					jpeg_read_scanlines(&cinfo, &srcPixelRowZero, 1);
				jpeg_read_scanlines(&cinfo, &srcPixelRowOne, 1);
			}


			dstPixel = (Uint8 *) surface->pixels + row * surface->pitch;
			srcPixel00 = srcPixelRowZero;
			srcPixel01 = srcPixel00 + cinfo.output_components;
			srcPixel10 = srcPixelRowOne;
			srcPixel11 = srcPixel10 + cinfo.output_components;

			for (col = 0; col < maxWidth; col++) {
				srcPixel00 += widthSamples[col] * cinfo.output_components;
				srcPixel01 = srcPixel00 + cinfo.output_components;
				srcPixel10 += widthSamples[col] * cinfo.output_components;
				srcPixel11 = srcPixel10 + cinfo.output_components;



				
				dstVal
					= *srcPixel00 * (CRACK_PRECISION - heightCracks[row])
					* (CRACK_PRECISION - widthCracks[col])
					+ *srcPixel01 * (CRACK_PRECISION - heightCracks[row])
					* (widthCracks[col])
					+ *srcPixel10 * (heightCracks[row])
					* (CRACK_PRECISION - widthCracks[col])
					+ *srcPixel11 * (heightCracks[row])
					* (widthCracks[col]);


				
				
				(dstVal) >>= CRACK_SHIFT;
				(dstVal) >>= CRACK_SHIFT;
				*dstPixel = dstVal & 0xFF;

				srcPixel00++;
				srcPixel01++;
				srcPixel10++;
				srcPixel11++;
				dstPixel++;

				dstVal
					= *srcPixel00 * (CRACK_PRECISION - heightCracks[row])
					* (CRACK_PRECISION - widthCracks[col])
					+ *srcPixel01 * (CRACK_PRECISION - heightCracks[row])
					* (widthCracks[col])
					+ *srcPixel10 * (heightCracks[row])
					* (CRACK_PRECISION - widthCracks[col])
					+ *srcPixel11 * (heightCracks[row])
					* (widthCracks[col]);


				
				
				(dstVal) >>= CRACK_SHIFT;
				(dstVal) >>= CRACK_SHIFT;
				*dstPixel = dstVal & 0xFF;

				srcPixel00++;
				srcPixel01++;
				srcPixel10++;
				srcPixel11++;
				dstPixel++;

				dstVal
					= *srcPixel00 * (CRACK_PRECISION - heightCracks[row])
					* (CRACK_PRECISION - widthCracks[col])
					+ *srcPixel01 * (CRACK_PRECISION - heightCracks[row])
					* (widthCracks[col])
					+ *srcPixel10 * (heightCracks[row])
					* (CRACK_PRECISION - widthCracks[col])
					+ *srcPixel11 * (heightCracks[row])
					* (widthCracks[col]);

				
				
				(dstVal) >>= CRACK_SHIFT;
				(dstVal) >>= CRACK_SHIFT;
				*dstPixel = dstVal & 0xFF;

				srcPixel00 -= 2;
				srcPixel01 -= 2;
				srcPixel10 -= 2;
				srcPixel11 -= 2;
				dstPixel++;
			}
		}


		if (widthSamples != NULL)
			free(widthSamples);
		if (heightSamples != NULL)
			free(heightSamples);
		if (widthCracks != NULL)
			free(widthCracks);
		if (heightCracks != NULL)
			free(heightCracks);
		if (srcPixelRowZero != NULL)
			free(srcPixelRowZero);
		if (srcPixelRowOne != NULL)
			free(srcPixelRowOne);

		
		
	}

	else {
		
	}

	
	if (SDL_MUSTLOCK(surface)) {
		SDL_UnlockSurface(surface);
	}
	
	

	
	jpeg_destroy_decompress(&cinfo);
	fclose(fp);

	return surface;
}
