/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2004. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <assert.h>
#include <string.h>
#include "SDL_rotate.h"


static SDL_bool
RotateSurface0( SDL_Surface* src,
				SDL_Surface* dst )
{
	assert( src != NULL );
	assert( dst != NULL );
	assert( src != dst );
    assert( src->w == dst->w );
    assert( src->h == dst->h );

	if ( SDL_LockSurface(src) ) {
		return SDL_FALSE;
	}
	if ( SDL_LockSurface(dst) ) {
		SDL_UnlockSurface(src);
		return SDL_FALSE;
	}

	
	memcpy( dst->pixels, src->pixels, (dst->pitch * dst->h) );

	SDL_UnlockSurface(dst);
	SDL_UnlockSurface(src);

	return SDL_TRUE;
}


static SDL_bool
RotateSurface90( SDL_Surface* src,
				 SDL_Surface* dst )
{
	int 	i, j;
	int 	bytesPerPixel;
	char*	srcPixels;
	char*	dstPixels;

	assert( src != NULL );
	assert( dst != NULL );
	assert( src != dst );
    assert( src->w == dst->h );
    assert( src->h == dst->w );

	if ( SDL_LockSurface(src) ) {
		return SDL_FALSE;
	}
	if ( SDL_LockSurface(dst) ) {
		SDL_UnlockSurface(src);
		return SDL_FALSE;
	}

	bytesPerPixel = src->format->BytesPerPixel;
	for ( j = 0; j < src->h; j++ ) {
		srcPixels = (char *)src->pixels + (src->pitch * j);
		dstPixels = (char *)dst->pixels + (dst->w - j - 1) * bytesPerPixel;

		for ( i = 0; i < src->w; i++ ) {
			memcpy( dstPixels, srcPixels, bytesPerPixel );
			dstPixels += dst->pitch;
			srcPixels += bytesPerPixel;
		}
	}

	SDL_UnlockSurface(dst);
	SDL_UnlockSurface(src);

	return SDL_TRUE;
}


static SDL_bool
RotateSurface180( SDL_Surface* src,
				  SDL_Surface* dst )
{
	int		i, j;
	int 	bytesPerPixel;
	char*	srcPixels;
	char*	dstPixels;

	assert( src != NULL );
	assert( dst != NULL );
	assert( src != dst );
    assert( src->w == dst->w );
    assert( src->h == dst->h );

	if ( SDL_LockSurface(src) ) {
		return SDL_FALSE;
	}
	if ( SDL_LockSurface(dst) ) {
		SDL_UnlockSurface(src);
		return SDL_FALSE;
	}

	
	bytesPerPixel = src->format->BytesPerPixel;
	srcPixels = (char *)src->pixels;
	for ( j = 0; j < src->h; j++ ) {
		srcPixels = (char *)src->pixels + (src->pitch * j);
		dstPixels = (char *)dst->pixels
			+ dst->pitch * (dst->h - j - 1) 
			+ (dst->w - 1) * bytesPerPixel;

		for ( i = 0; i < src->w; i++ ) {
			memcpy( dstPixels, srcPixels, bytesPerPixel );
			dstPixels -= bytesPerPixel;
			srcPixels += bytesPerPixel;
		}
	}
	
	SDL_UnlockSurface(dst);
	SDL_UnlockSurface(src);

	return SDL_TRUE;
}


static SDL_bool
RotateSurface270( SDL_Surface* src,
				  SDL_Surface* dst )
{
	int 	i, j;
	int 	bytesPerPixel;
	char*	srcPixels;
	char*	dstPixels;

	assert( src != NULL );
	assert( dst != NULL );
	assert( src != dst );
    assert( src->w == dst->h );
    assert( src->h == dst->w );

	if ( SDL_LockSurface(src) ) {
		return SDL_FALSE;
	}
	if ( SDL_LockSurface(dst) ) {
		SDL_UnlockSurface(src);
		return SDL_FALSE;
	}

	
	bytesPerPixel = src->format->BytesPerPixel;
	srcPixels = (char *)src->pixels;
	for ( j = 0; j < src->h; j++ ) {
		srcPixels = (char *)src->pixels + (src->pitch * j);
		dstPixels = (char *)dst->pixels 
			+ dst->pitch * (dst->h - 1)
			+ j * bytesPerPixel;

		for ( i = 0; i < src->w; i++ ) {
			memcpy( dstPixels, srcPixels, bytesPerPixel );
			dstPixels -= dst->pitch;
			srcPixels += bytesPerPixel;
		}
	}

	SDL_UnlockSurface(dst);
	SDL_UnlockSurface(src);

	return SDL_TRUE;
}




SDL_Surface* 
SDL_RotateSurface( SDL_Surface* src,
				   SDL_RotateAngle angle )
{
	SDL_bool	ret = SDL_FALSE;
	SDL_Surface*	dst;
	int				dstWidth, dstHeight;
	
	assert( src != NULL );

	
	if ( angle == SDL_ROTATE_90 || angle == SDL_ROTATE_270 ) {
		dstWidth = src->h;
		dstHeight = src->w;
	} else {
		dstWidth = src->w;
		dstHeight = src->h;
	}

	
	dst = SDL_CreateRGBSurface( SDL_SWSURFACE,
								dstWidth, dstHeight,
								src->format->BitsPerPixel,
								src->format->Rmask,
								src->format->Gmask,
								src->format->Bmask,
								src->format->Amask );
	if ( dst == NULL ) {
		return NULL;
	}

	if ( src->format->palette && src->format->palette->ncolors > 0 ) {
		
		ret = SDL_SetColors( dst, src->format->palette->colors,
							  0, src->format->palette->ncolors );

		if ( ret != 1 ) {
			SDL_FreeSurface( dst );
			return NULL;
		}

		if ((src->flags & SDL_SRCCOLORKEY) == SDL_SRCCOLORKEY) {
			ret = SDL_SetColorKey(dst, SDL_SRCCOLORKEY,
			                      src->format->colorkey);

			if ( ret != 0 ) {
				SDL_FreeSurface(dst);
				return NULL;
			}
		}
	} 

	switch ( angle ) {
	case SDL_ROTATE_0:
		ret = RotateSurface0( src, dst );
		break;

	case SDL_ROTATE_90:
		ret = RotateSurface90( src, dst );
		break;

	case SDL_ROTATE_180:
		ret = RotateSurface180( src, dst );
		break;

	case SDL_ROTATE_270:
		ret = RotateSurface270( src, dst );
		break;		
	}

	if ( ret ) {
		return dst;
	} else {
		SDL_FreeSurface( dst );
		return NULL;
	}
}
