/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2004. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include "SDL_stretch.h"



#define FP_SHIFT		13      
#define	FP_PRECISION	8192    




static inline Uint32
roundFixedPoint( Uint32 in )
{
	Uint32 rem = in % FP_PRECISION;

	if ( rem < (FP_PRECISION >> 1) ) 
		return (in - rem);
	else
		return (in - rem + FP_PRECISION);		
}



static SDL_bool
isRectSizeEqual( const SDL_Rect *rect1, const SDL_Rect *rect2 )
{
	if ( rect1 == rect2 ) {
		if ( rect1 == NULL ) {
            
			return SDL_FALSE;
		} else {
            
			return SDL_TRUE;
		}
	}

	
	
	if ( (rect1 == NULL) || (rect2 == NULL) ) {
		return SDL_FALSE;
	}

	
	if ( (rect1->w == rect2->w) && (rect1->h == rect2->h) ) {
		return SDL_TRUE;
	}

	return SDL_FALSE;
}


static SDL_bool
isPixelFormatEqual( const SDL_PixelFormat *form1,
					const SDL_PixelFormat *form2 )
{
	if ( form1->BitsPerPixel != form2->BitsPerPixel ) 
		return SDL_FALSE;

	if ( form1->BytesPerPixel != form2->BytesPerPixel ) 
		return SDL_FALSE;

	if ( form1->alpha != form2->alpha ) 
		return SDL_FALSE;

	if ( form1->colorkey != form2->colorkey ) 
		return SDL_FALSE;

	if ( form1->Rmask != form2->Rmask || form1->Gmask != form2->Gmask ||
		 form1->Bmask != form2->Bmask || form1->Amask != form2->Amask ) 
		return SDL_FALSE;

	return SDL_TRUE;
}


static void
doCalcNearestSampleVector( Uint16* samples,
						   Uint16 srcSpan, Uint16 dstSpan )
{
	int i;
	Uint32	ratio;

	assert( samples != NULL );
	assert( srcSpan > 0 );
	assert( dstSpan > 0 );

	if ( dstSpan < 2 || srcSpan < 2 ) {
		ratio = 0;
	} else {
		ratio = ( (srcSpan-1) << FP_SHIFT ) / (dstSpan-1);
	}

	
	
	samples[0] = 0;
	samples[dstSpan-1] = srcSpan - 1;

	
	for ( i = 1; i < (dstSpan-1); i++ ) {
		samples[i] = roundFixedPoint(ratio * i) >> FP_SHIFT;
	}
}



static int
doNearestBlit( SDL_Surface* src, SDL_Rect* srcRect, 
			   SDL_Surface* dst, SDL_Rect* dstRect )
{
	int row, col;
	Uint16* widthSamples;
	Uint16* heightSamples;
	Uint8* 	srcPixelRow;
	Uint8*	dstPixel;
	Uint16	maxHeight, maxWidth;

	assert( src != NULL );
	assert( dst != NULL );
	assert( srcRect != NULL );
	assert( dstRect != NULL );
	assert( isPixelFormatEqual(src->format, dst->format) );

	
	if ( srcRect->w < 1 || srcRect->h < 1 ||
		 dstRect->w < 1 || dstRect->h < 1 ) {
		return -1;
	}

	
	widthSamples = (Uint16*)malloc( dstRect->w * sizeof(Uint16) );
	if ( widthSamples == NULL ) {
		return -2;
	}
	heightSamples = (Uint16*)malloc( dstRect->h * sizeof(Uint16) );
	if ( heightSamples == NULL ) {
		free( widthSamples );
		return -2;
	}

	
	doCalcNearestSampleVector( widthSamples, srcRect->w, dstRect->w );
	doCalcNearestSampleVector( heightSamples, srcRect->h, dstRect->h );

	
	for ( col = 0; col < dstRect->w; col++ ) {
		widthSamples[col] *= src->format->BytesPerPixel;
	}

	
	if ( (dstRect->w + dstRect->x) > dst->w ) {
		maxWidth = dst->w - dstRect->x;
	} else {
		maxWidth = dstRect->w;
	}
	if ( (dstRect->h + dstRect->y) > dst->h ) {
		maxHeight = dst->h - dstRect->y;
	} else {
		maxHeight = dstRect->h;
	}

	
	for ( row = 0; row < maxHeight; row++ ) {
		srcPixelRow = (Uint8*)src->pixels + 
			(heightSamples[row] + srcRect->y) * src->pitch +
			srcRect->x * src->format->BytesPerPixel;

		dstPixel = (Uint8*)dst->pixels + 
			(row + dstRect->y) * dst->pitch +
			dstRect->x * dst->format->BytesPerPixel;

			switch (dst->format->BytesPerPixel) {
			case  4:
				for ( col = 0; col < maxWidth; col++ ) {
					*((Uint32 *)dstPixel) = *((Uint32 *)
						(srcPixelRow + widthSamples[col]));
					dstPixel += dst->format->BytesPerPixel;
				}
				break;
			case 2:
				for ( col = 0; col < maxWidth; col++ ) {
					*((Uint16 *)dstPixel) = *((Uint16 *)
						(srcPixelRow + widthSamples[col]));
					dstPixel += dst->format->BytesPerPixel;
				}
				break;
			case 1:
				for ( col = 0; col < maxWidth; col++ ) {
					*((Uint8 *)dstPixel) = *((Uint8 *)
						(srcPixelRow + widthSamples[col]));
					dstPixel += dst->format->BytesPerPixel;
				}
				break;
			default:
			}
	}

	
	free( heightSamples );
	free( widthSamples );
	return 0;
}










#define CRACK_PRECISION  256
#define CRACK_SHIFT      8
#define CRACK_MASK       0xFF

 

 

static void
doCalcBiLinSampleVector( Uint32 *samples, Uint32 *cracks,
			 Uint32  srcSpan, Uint32  dstSpan )
{
	int i;
	Uint32	ratio;

	assert( samples != NULL );
	assert( cracks  != NULL );
	assert( srcSpan > 0 );
	assert( dstSpan > 0 );

	if ( dstSpan < 2 || srcSpan < 2 ) {
		ratio = 0;
	} else {  
		ratio = ( (srcSpan) << CRACK_SHIFT ) / (dstSpan);
	}

	
	samples[0] = 0;
	cracks[0] = 0;

	
	for ( i = 1; i < dstSpan; i++ ) {
		samples[i] = (ratio * i) >> CRACK_SHIFT;
		cracks[i] = (ratio * i) & CRACK_MASK;
	}
	
	
	if (ratio == 0) return;
	i--;
	assert (samples[i] <= srcSpan-1);
	
	while (samples[i] == srcSpan-1) {
		assert(i > 0);
		samples[i] --;
		cracks[i] = CRACK_PRECISION - 1;  
		i--;
	}
 	
}






static int
doBilinearBlit( SDL_Surface* src, SDL_Rect* srcRect, 
				SDL_Surface* dst, SDL_Rect* dstRect,
				SDL_bool directCopy )
{
	
        

        int     row , col;


	Uint8* 	srcPixelRowZero;
	Uint8* 	srcPixelRowOne;
	Uint8* 	srcPixel00;
	Uint8* 	srcPixel01;
	Uint8* 	srcPixel10;
	Uint8* 	srcPixel11;
	Uint8*	dstPixel;

	Uint32 rmask, gmask, bmask, amask;
	Uint32* widthSamples;
	Uint32* heightSamples;
	Uint32* widthCracks;
	Uint32* heightCracks;
	Uint16	maxHeight;
	Uint16  maxWidth;
	Uint32  dstVal;
	SDL_Surface * middleman  = NULL;
	SDL_Rect middlerects, middlerectd;


	assert( src != NULL );
	assert( dst != NULL );
	assert( srcRect != NULL );
	assert( dstRect != NULL );
		
#ifdef DEBUG
	fprintf (stderr, "(sx %d, sy %d, sw %d, sh %d); (dx %d, dy %d, dw %d, dh %d)\n", 
						srcRect->x,
						srcRect->y,
						srcRect->w,
						srcRect->h,

						dstRect->x,
						dstRect->y,
						dstRect->w,
						dstRect->h);
#endif 
	
	

	assert(
			( src->format->BytesPerPixel == 3 ) 
			|| 
			( src->format->BytesPerPixel == 4 ) 
	);


	if ((src->format->BytesPerPixel != 3) &&
		(src->format->BytesPerPixel != 4)) {

		return -1; 
	}

	
	if ( srcRect->w < 1 || srcRect->h < 1 ||
		dstRect->w < 1 || dstRect->h < 1 ) {
		return -1;
	}        

	
	widthSamples = (Uint32*)malloc( dstRect->w * sizeof(Uint32) );
	if ( widthSamples == NULL ) {
		return -2;
	}
	heightSamples = (Uint32*)malloc( dstRect->h * sizeof(Uint32) );
	if ( heightSamples == NULL ) {
		free( widthSamples );
		return -2;
	}
	widthCracks = (Uint32*)malloc( dstRect->w * sizeof(Uint32) );
	if ( widthCracks == NULL ) {
		free( widthSamples );
		free( heightSamples );
		return -2;
	}
	heightCracks = (Uint32*)malloc( dstRect->h * sizeof(Uint32) );
	if ( heightCracks == NULL ) {
		free( widthSamples );
		free( heightSamples);
		free( widthCracks);
		return -2;
	}

	
	doCalcBiLinSampleVector( widthSamples, widthCracks, 
				 srcRect->w, dstRect->w );
	doCalcBiLinSampleVector( heightSamples, heightCracks, 
				 srcRect->h, dstRect->h );

	
	if ( (dstRect->w + dstRect->x) > dst->w ) {
		maxWidth = dst->w - dstRect->x;
	} else {
		maxWidth = dstRect->w;
	}
	if ( (dstRect->h + dstRect->y) > dst->h ) {
		maxHeight = dst->h - dstRect->y;
	} else {
		maxHeight = dstRect->h;
	}


#if SDL_BYTEORDER == SDL_BIG_ENDIAN
        rmask = 0xff000000;
        gmask = 0x00ff0000;
        bmask = 0x0000ff00;
        amask = 0x000000ff;
#else
        rmask = 0x000000ff;
        gmask = 0x0000ff00;
        bmask = 0x00ff0000;
        amask = 0xff000000;
#endif

	if (src->format->BytesPerPixel == 3) {
		amask = 0x00000000;
	} 

	middleman= SDL_CreateRGBSurface(SDL_SWSURFACE,
									maxWidth,
									1,
									src->format->BytesPerPixel * 8,
									src->format->Rmask, 
									src->format->Gmask, 
									src->format->Bmask, 
									src->format->Amask );

	if (middleman == NULL)
		return 0;

	if (directCopy) {
		SDL_SetAlpha(middleman, 0, 0);
	}

	middlerects.x = 0;
	middlerects.y = 0;
	middlerects.w = maxWidth;
	middlerects.h = 1;

	middlerectd.x = dstRect->x;
	middlerectd.y = dstRect->y;
	middlerectd.w = maxWidth;
	middlerectd.h = 1;
 
	

	for ( row = 0; row < maxHeight; row++ ) {
		srcPixelRowZero = (Uint8*)src->pixels + 
		((heightSamples[row] + 0) + srcRect->y) * src->pitch +
		srcRect->x * src->format->BytesPerPixel;

		srcPixelRowOne = (Uint8*)src->pixels + 
		((heightSamples[row] + 1) + srcRect->y) * src->pitch +
		srcRect->x * src->format->BytesPerPixel;

		dstPixel = middleman->pixels;

		for ( col = 0; col < maxWidth; col++ ) {
			srcPixel00 = srcPixelRowZero + widthSamples[col] * src->format->BytesPerPixel;
			srcPixel01 = srcPixel00 + src->format->BytesPerPixel;
			srcPixel10 = srcPixelRowOne  + widthSamples[col] * src->format->BytesPerPixel;
			srcPixel11 = srcPixel10 + src->format->BytesPerPixel;

			
			dstVal 
			= *srcPixel00 * (CRACK_PRECISION - heightCracks[row])  
			* (CRACK_PRECISION - widthCracks[col]) 
			+ *srcPixel01 * (CRACK_PRECISION - heightCracks[row])
			* (widthCracks[col])
			+ *srcPixel10 * (heightCracks[row])
			* (CRACK_PRECISION - widthCracks[col])
			+ *srcPixel11 * (heightCracks[row])
			* (widthCracks[col]);


			
			
			(dstVal) >>= CRACK_SHIFT;
			(dstVal) >>= CRACK_SHIFT;
			*dstPixel = dstVal & 0xFF;

			srcPixel00++;
			srcPixel01++;
			srcPixel10++;
			srcPixel11++;
			dstPixel++;

			dstVal 
			= *srcPixel00 * (CRACK_PRECISION - heightCracks[row])  
			* (CRACK_PRECISION - widthCracks[col]) 
			+ *srcPixel01 * (CRACK_PRECISION - heightCracks[row])
			* (widthCracks[col])
			+ *srcPixel10 * (heightCracks[row])
			* (CRACK_PRECISION - widthCracks[col])
			+ *srcPixel11 * (heightCracks[row])
			* (widthCracks[col]);


			
			
			(dstVal) >>= CRACK_SHIFT;
			(dstVal) >>= CRACK_SHIFT;
			*dstPixel = dstVal & 0xFF;

			srcPixel00++;
			srcPixel01++;
			srcPixel10++;
			srcPixel11++;
			dstPixel++;

			dstVal 
			= *srcPixel00 * (CRACK_PRECISION - heightCracks[row])  
			* (CRACK_PRECISION - widthCracks[col]) 
			+ *srcPixel01 * (CRACK_PRECISION - heightCracks[row])
			* (widthCracks[col])
			+ *srcPixel10 * (heightCracks[row])
			* (CRACK_PRECISION - widthCracks[col])
			+ *srcPixel11 * (heightCracks[row])
			* (widthCracks[col]);

			
			
			(dstVal) >>= CRACK_SHIFT;
			(dstVal) >>= CRACK_SHIFT;
			*dstPixel = dstVal & 0xFF;

			if (src->format->BytesPerPixel == 4){
				
				srcPixel00++;
				srcPixel01++;
				srcPixel10++;
				srcPixel11++;
				dstPixel++;

				dstVal 
				= *srcPixel00 * (CRACK_PRECISION - heightCracks[row])  
				* (CRACK_PRECISION - widthCracks[col]) 
				+ *srcPixel01 * (CRACK_PRECISION - heightCracks[row])
				* (widthCracks[col])
				+ *srcPixel10 * (heightCracks[row])
				* (CRACK_PRECISION - widthCracks[col])
				+ *srcPixel11 * (heightCracks[row])
				* (widthCracks[col]);

				
				
				(dstVal) >>= CRACK_SHIFT;
				(dstVal) >>= CRACK_SHIFT;
				*dstPixel = dstVal & 0xFF;
			} 

		 	dstPixel++;
		}
 
		SDL_BlitSurface(middleman, &middlerects, dst, &middlerectd);
		middlerectd.y ++; 
	}

	
	SDL_FreeSurface(middleman);
	free( heightSamples );
	free( widthSamples );
	free( heightCracks );
	free( widthCracks );

	return 0;
}




int 
SDL_StretchBlit( SDL_Surface *src, SDL_Rect *srcrect,
				 SDL_Surface *dst, SDL_Rect *dstrect,
				 SDL_StretchMode mode )
{
	int 			ret = 0;
	SDL_Rect		realSrcRect, realDstRect;
	SDL_Surface*	tmpSrc = NULL;
	int  			outOfbounds = 0;

	
	if ( isRectSizeEqual(srcrect, dstrect) ) {
		return SDL_BlitSurface( src, srcrect, dst, dstrect );
	}

    if (src->format->BitsPerPixel == 8) { 
		
		

		mode = SDL_STRETCH_NEAREST;	
	}

	
	if ((mode == SDL_STRETCH_NEAREST) &&
	    ( !isPixelFormatEqual( src->format, dst->format ))) {

		
		
		tmpSrc = SDL_ConvertSurface( src, dst->format, SDL_SWSURFACE );
		if ( tmpSrc == NULL ) {
			return -1;
		}
	}
	else {
		tmpSrc = src;
	}

	if ( SDL_LockSurface(tmpSrc) ) {
		if ( tmpSrc != src ) {
			SDL_FreeSurface( tmpSrc );
		}
		return -1;
	}	

	
	if ( srcrect == NULL ) {
		realSrcRect.x = 0; realSrcRect.y = 0;
		realSrcRect.w = tmpSrc->w; realSrcRect.h = tmpSrc->h;
	} else {
		memcpy( &realSrcRect, srcrect, sizeof(SDL_Rect) );
	}

	if ( dstrect == NULL ) {
		realDstRect.x = 0; realDstRect.y = 0;
		realDstRect.w = dst->w; realDstRect.h = dst->h;
	} else {
		memcpy( &realDstRect, dstrect, sizeof(SDL_Rect) );
	}

	if (realDstRect.x + realDstRect.w < 0) outOfbounds = 1;
	if (realDstRect.y + realDstRect.h < 0) outOfbounds = 1;
	if (realDstRect.x > dst->w) outOfbounds = 1;
	if (realDstRect.y > dst->h) outOfbounds = 1;

	if (realSrcRect.x + realSrcRect.w < 0) outOfbounds = 1;
	if (realSrcRect.y + realSrcRect.h < 0) outOfbounds = 1;
	if (realSrcRect.x > src->w) outOfbounds = 1;
	if (realSrcRect.y > src->h) outOfbounds = 1;


	if (!(outOfbounds)) {
		
		if (realSrcRect.x < 0) {realSrcRect.w += realSrcRect.x; realSrcRect.x = 0; }
		if (realSrcRect.y < 0) {realSrcRect.h += realSrcRect.y; realSrcRect.y = 0; }
		if (realSrcRect.x + realSrcRect.w > src->w) {realSrcRect.w = src->w - realSrcRect.x;}
		if (realSrcRect.y + realSrcRect.h > src->h) {realSrcRect.h = src->h - realSrcRect.y;}

		if (realDstRect.x < 0) {
			
			
			int cutoff = realSrcRect.w * (realDstRect.x * (-1)); 
			cutoff /= (float)realDstRect.w; 
			realSrcRect.x += cutoff; 
			realSrcRect.w -= cutoff;
			realDstRect.w += realDstRect.x; 
			realDstRect.x = 0; 
		} 
		if (realDstRect.y < 0) {
			
			
			int cutoff = realSrcRect.h * (realDstRect.y * (-1));
			cutoff /= (float)realDstRect.h; 
			realSrcRect.y += cutoff; 
			realSrcRect.h -= cutoff;

			realDstRect.h += realDstRect.y; 
			realDstRect.y = 0; 
		}

		if (realDstRect.x + realDstRect.w > dst->w) {
			
			
			int cutoff = realDstRect.x + realDstRect.w - dst->w; 
			cutoff *= realSrcRect.w; 
			cutoff /= (float)realDstRect.w; 
			realSrcRect.w -= cutoff; 

			realDstRect.w = dst->w - realDstRect.x; 
		}
		if (realDstRect.y + realDstRect.h > dst->h) {
			
			
			int cutoff = realDstRect.y + realDstRect.h - dst->h;
			cutoff *= realSrcRect.h; 
			cutoff /= (float)realDstRect.h;
			realSrcRect.h -= cutoff;

			realDstRect.h = dst->h - realDstRect.y; 
		}

		if ( mode == SDL_STRETCH_BILINEAR ) {
			
			ret = doBilinearBlit( tmpSrc, &realSrcRect, dst, &realDstRect, SDL_FALSE );
		} else if ( mode == SDL_STRETCH_BILINEAR_LI ) {
			ret = doBilinearBlit( tmpSrc, &realSrcRect, dst, &realDstRect, SDL_TRUE );
		} else {
			
			ret = doNearestBlit( tmpSrc, &realSrcRect, dst, &realDstRect );
		}
	}

	SDL_UnlockSurface(tmpSrc);

	
	if ( tmpSrc != src ) {
		SDL_FreeSurface( tmpSrc );
	}

	if (outOfbounds) {
		fprintf (stderr, "warning: you tried to blit out of bounds!! \n");
		return -1;
	}

	
	if (dstrect != NULL) {
		dstrect->x = realDstRect.x; 		
		dstrect->y = realDstRect.y; 		
		dstrect->w = realDstRect.w; 		
		dstrect->h = realDstRect.h; 		
	}

	return ret;
}
