/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2004. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "SDL_text.h"




typedef struct {
	unsigned int glyph	:2;		
	unsigned int byte	:1;		
	unsigned int bit	:1;		
	unsigned int scan	:2;		
	unsigned int dummy	:2;		
	unsigned int id		:24;	
} format_t;


typedef struct {
	int	type;		
	int	format;		
	int	size;		
	int	offset;		
} header_t;


#define	PCF_ACCELERATORS	0x002	
#define	PCF_METRICS			0x004	
#define	PCF_BITMAPS			0x008	
#define	PCF_BDF_ENCODINGS	0x020	

#define	GLYPHPADOPTIONS			4

#define	PCF_DEFAULT_FORMAT		0
#define PCF_COMPRESSED_METRICS	1


typedef struct {
	short	leftSideBearing;
	short	rightSideBearing;
	short	characterWidth;
	short	ascent;
	short	descent;
	short	attributes;
} metric_t;


typedef struct {
	unsigned char	leftSideBearing;
	unsigned char	rightSideBearing;
	unsigned char	characterWidth;
	unsigned char	ascent;
	unsigned char	descent;
} metric_comp_t ;


typedef struct {
	format_t		format;
	char			noOverlap;			
	char			constantMetrics;	
	char			inkInside;			
	char			inkMetrics;			
	char			drawDirection;		
	char			dummy;				
	int				fontAscent;
	int				fontDescent;
	int				maxOverlap;
	metric_t		minBounds;
	metric_t		maxBounds;
	
} accel_t;


typedef struct {
	accel_t			accelerators;			

	format_t		metricsFormat;			
	int				numMetrics;				
	unsigned int	offsetMetrics;			

	format_t		bitmapFormat;			
	unsigned int	offsetBitmapOffsets;	
	unsigned int	offsetBitmaps;			

	format_t		encodeFormat;			
	unsigned short	firstCol;				
	unsigned short	lastCol;				
	unsigned short	firstRow;				
	unsigned short	lastRow;				
	unsigned int	offsetEncodings;		
} fontinfo_t;




static short
GetEncode( const unsigned char *c, const fontinfo_t *fi, FILE *fp )
{
	short		encode = -1;
	int			encode_pos;

	
	if((*c <= fi->lastCol) && (*c >= fi->firstCol)){
		encode_pos = fi->offsetEncodings + (*c - fi->firstCol) * sizeof(encode); 
		fseek( fp, encode_pos, SEEK_SET );
		fread( &encode, sizeof(encode), 1, fp ); 
	}

	if( encode == -1 ){
		
		encode_pos = fi->offsetEncodings + ('?' - fi->firstCol) * sizeof(encode); 
		fseek( fp, encode_pos, SEEK_SET );
		fread( &encode, sizeof(encode), 1, fp ); 
	}

	return encode;
}



static int
GetMetric( metric_t *m, const char *c, const fontinfo_t *fi, FILE *fp )
{
	

	short			encode;
	metric_comp_t	cmetric;
	int				metric_pos;

	encode = GetEncode( c, fi, fp );
	if( encode == -1 ){
		return encode;
	}

	if( fi->metricsFormat.id == PCF_COMPRESSED_METRICS ) {
		metric_pos = fi->offsetMetrics + encode * sizeof(cmetric);
		fseek( fp, metric_pos, SEEK_SET );
		fread( &cmetric, sizeof(cmetric), 1, fp ); 

		
		m->leftSideBearing = (short)cmetric.leftSideBearing - 0x80;
		m->rightSideBearing = (short)cmetric.rightSideBearing - 0x80;
		m->ascent = (short)cmetric.ascent - 0x80;
		m->descent = (short)cmetric.descent - 0x80;
		m->characterWidth = (short)cmetric.characterWidth - 0x80;
	} else {
		metric_pos = fi->offsetMetrics + encode * sizeof(metric_t);
		fseek( fp, metric_pos, SEEK_SET );
		fread( &m, sizeof(metric_t), 1, fp ); 
	}

	return 0;
}



static int
GetCharWidth( const metric_t *metric )
{
	return metric->characterWidth;
}



static int
GetTextWidth( const char *text, const fontinfo_t *fi, FILE *fp )
{
	int	i, width = 0;
	metric_t	metric;

	

	
	GetMetric( &metric, &text[0], fi, fp );
	width += -metric.leftSideBearing + GetCharWidth( &metric );;
	
	
	for( i = 1; text[i]; i++ ){
		GetMetric( &metric, &text[i], fi, fp );

		width += GetCharWidth( &metric );
	}

	
	if( metric.rightSideBearing > metric.characterWidth ){
		width += metric.rightSideBearing - metric.characterWidth;
	}

	return width;
}



static int
AnalyzeHeader( FILE* fp, fontinfo_t* fontinfo )
{
	char		buf[4];	
	int			num_sections;
	int			i;
	long		last_pos;
	header_t	hdr;

	
	rewind( fp );

	
	memset( fontinfo, 0, sizeof(*fontinfo) );

	
	if( fread( buf, 4, 1, fp ) != 1 ){
		return -1; 
	}

	
	if( fread( &num_sections, 4, 1, fp ) != 1 ){
		return -1; 
	}

	
	last_pos = ftell( fp );
	for( i = 0 ; i < num_sections; i++ ){
		fseek( fp, last_pos, SEEK_SET );
		fread( &hdr, sizeof(hdr), 1, fp );

		
		last_pos = ftell( fp );

		
		fseek( fp, hdr.offset, SEEK_SET );

		
		switch(hdr.type){
		case PCF_ACCELERATORS:
			fread( &fontinfo->accelerators, sizeof(fontinfo->accelerators),
																	1, fp );
			break;
		case PCF_METRICS:
			fread( &fontinfo->metricsFormat, 4, 1, fp );
			if( fontinfo->metricsFormat.id == PCF_COMPRESSED_METRICS ){
				fread( &(fontinfo->numMetrics), sizeof(short), 1, fp );
			} else {
				fread( &(fontinfo->numMetrics), sizeof(int), 1, fp );
			}
			fontinfo->offsetMetrics = ftell( fp );
			break;
		case PCF_BITMAPS:
			fread( &fontinfo->bitmapFormat, 4, 1, fp );
			fread( &(fontinfo->numMetrics), sizeof(int), 1, fp );
			fontinfo->offsetBitmapOffsets = ftell( fp );
			fontinfo->offsetBitmaps = fontinfo->offsetBitmapOffsets + 
									sizeof(int) * fontinfo->numMetrics +
									sizeof(int) * GLYPHPADOPTIONS;
			break;
		case PCF_BDF_ENCODINGS:
			fread( &fontinfo->encodeFormat, 4, 1, fp );
			fread( &fontinfo->firstCol, sizeof(short), 1, fp );
			fread( &fontinfo->lastCol, sizeof(short), 1, fp );
			fread( &fontinfo->firstRow, sizeof(short), 1, fp );
			fread( &fontinfo->lastRow, sizeof(short), 1, fp );
			fread( &buf, sizeof(short), 1, fp );	
			fontinfo->offsetEncodings = ftell( fp );
			break;
		default:
			break;
		}
	}

	
	if( fontinfo->numMetrics == 0 ||
		fontinfo->offsetMetrics == 0 ||
		fontinfo->offsetBitmapOffsets == 0 ||
		fontinfo->offsetBitmaps == 0 ||
		fontinfo->offsetEncodings == 0 )
	{
		return -1;
	}

	return 0;
}



static unsigned int
GetBitmapOffset( const char *c, const fontinfo_t *fi, FILE *fp )
{
	short			encode;
	unsigned int	bo_pos, offset = 0;

	encode = GetEncode( c, fi, fp );
	if( encode != -1 ){
		bo_pos = fi->offsetBitmapOffsets + encode * sizeof( offset );
		fseek( fp, bo_pos, SEEK_SET );
		fread( &offset, sizeof( offset ), 1, fp );
		offset += fi->offsetBitmaps;
	}

	return offset;
}



static void
bitOrderInvert( char *c )
{
	static const char invert[16] =
			{ 0, 8, 4, 12, 2, 10, 6, 14, 1, 9, 5, 13, 3, 11, 7, 15 };
    *c = (invert[*c & 15] << 4) | invert[(*c >> 4) & 15];
}



static void
DrawBits( const unsigned char src, unsigned char **dst ) 
{
	int		k;

	for( k = 7; k >= 0; k-- ){
		if( src & (1 << k) ){
			**dst = 1;
		}
		(*dst)++;
	}
}



static int
DrawText( SDL_Surface *s, const char *text, const fontinfo_t *fi, FILE *fp )
{
	int				i, x, y, cur_pos_x = 0;
	Uint32			bitmapoffset;
	metric_t		metric;
	unsigned char	src, *dst;
	int				baseline, linebytes, widthbytes, init_origin;

	baseline = fi->accelerators.fontAscent;
	linebytes = 1 << fi->bitmapFormat.glyph;

	
	GetMetric( &metric, &text[0], fi, fp );
	init_origin = -metric.leftSideBearing;

	for( i = 0; i < strlen( text ); i++ ){
		GetMetric( &metric, &text[i], fi, fp );
		bitmapoffset = GetBitmapOffset( &text[i], fi, fp );
		fseek( fp, bitmapoffset, SEEK_SET );

		widthbytes = (metric.rightSideBearing - metric.leftSideBearing + 7) / 8;

		for( y = 0; y < metric.ascent + metric.descent; y++ ){
			dst = (char *)s->pixels + init_origin + metric.leftSideBearing +
					(y + baseline - metric.ascent ) * s->pitch + cur_pos_x;
			for( x = 0; x < linebytes; x++ ){
				fread( &src, 1, 1, fp );
				if( x > widthbytes ){
					continue;
				}
				bitOrderInvert( &src );
				DrawBits( src, &dst );
			}
		}
			
		cur_pos_x += metric.characterWidth;
	}

	return 0;
}


SDL_Surface*
SDL_CreateText( const char *text, const char *fontPath )
{
	SDL_Surface		*surface = NULL;
	fontinfo_t		fi;
	FILE			*fp;
	int				width, height;
	SDL_Palette		*palette;

	Uint8			video_bpp = 8;
	Uint32			videoflags = SDL_SWSURFACE;

	if( fontPath == NULL ){
		return NULL;
	}

	
	fp = fopen( fontPath, "rb" );
	if( fp == NULL ){
		return NULL;
	}
	
	
	if( AnalyzeHeader( fp, &fi ) ){
		return NULL;
	}

	width = GetTextWidth( text, &fi, fp );
	height = fi.accelerators.fontAscent + fi.accelerators.fontDescent;

	surface=SDL_AllocSurface(videoflags, width, height , video_bpp, 0, 0, 0, 0);
	if ( surface != NULL ) {

		palette = surface->format->palette;
		palette->colors[0].r = 0 ;
		palette->colors[0].g = 0 ;
		palette->colors[0].b = 0 ;
		palette->colors[1].r = 255;
		palette->colors[1].g = 255;
		palette->colors[1].b = 255;
		SDL_SetColorKey( surface, SDL_SRCCOLORKEY|SDL_RLEACCEL, 0 );

		DrawText( surface, text, &fi, fp );
		SDL_UpdateRect( surface, 0, 0, 0, 0 );
    }
	
	fclose( fp );

	return surface;
}
