/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2004. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdio.h>
#include <assert.h>
#include <jpeglib.h>
#include <setjmp.h>
#include <string.h>

#include "SDL_jpeg.h"
#include "SDL_thumbnail.h"



const char	SOI[] =	{ 0xff, 0xd8 };     
const char	DQT[] = { 0xff, 0xdb };     
const char	APP1[] = { 0xff, 0xe1 };    
                                        
const char EXIFID[] = { 0x45, 0x78, 0x69, 0x66 };   
                                                    


struct my_error_mgr {
	struct jpeg_error_mgr pub;

	jmp_buf setjmp_buffer;
};
typedef struct my_error_mgr * my_error_ptr;



static void
my_error_exit( j_common_ptr cinfo )
{
	my_error_ptr myerr = (my_error_ptr)cinfo->err;

	(*cinfo->err->output_message)( cinfo );

	
	longjmp( myerr->setjmp_buffer, 1 );
}



static void
my_output_message( j_common_ptr cinfo )
{
#ifndef NDEBUG
  char buffer[JMSG_LENGTH_MAX];

  
  (*cinfo->err->format_message)( cinfo, buffer );

  
  fprintf( stderr, "%s\n", buffer );
#endif
}



long
SDL_HasThumbnail( const char *filename )
{
	SDL_bool		ret = SDL_FALSE;
	int				i;
	unsigned short	s;
	char			buf[4], c;
	size_t			n;

	FILE			*fp;
	fp = fopen( filename, "rb" );
	if ( fp == NULL ) {
		return 0;
	}

	
	n = fread(buf, sizeof(SOI), 1, fp);
	if( (n != 1) || (buf[0] != SOI[0]) || (buf[1] != SOI[1]) ){
		goto bail;
	}

	
	while( !feof(fp) ){
		c = fgetc(fp);
		if( c != APP1[0] ){	
			continue;
		}

		
		c = fgetc(fp);

		
		if( c == DQT[1] ){
			break;
		} 

		if( ((unsigned char)c < 0xe0) || ((unsigned char)c > 0xef) ){
			
			continue;
		} else {
			
			fread( &s, sizeof(APP1), 1, fp );
			s = (s << 8) | ( s >> 8 );

			if( c == APP1[1] ){
				
				
				n = fread(buf, sizeof(EXIFID), 1, fp);
				if ( ( n > 0 ) && strncmp( buf, EXIFID, sizeof(EXIFID) ) ){
					goto bail;
				}

				
				for( i = 0; i < s ; i++){
					c = fgetc(fp);
					if( c != SOI[0] ){
						continue;
					}
					if( (c = fgetc(fp)) == SOI[1] ){
						
						
						fseek( fp, -2, SEEK_CUR );
						ret = SDL_TRUE;
						goto bail;
					} else {
						ungetc((unsigned char)c, fp);
					}
				}
				
				
				goto bail;
			} else {
				
				
				fseek( fp, (long)s - 2, SEEK_CUR );
			}
		}
	}

bail:
	if (ret == SDL_TRUE) {
		 long offset = ftell(fp);
		fclose(fp);
		return offset; 
	}
	fclose(fp);
	return -1;
}



SDL_Surface*
SDL_LoadThumbnail( const char* file, 
                   const SDL_JpegScale scale, 
                   const long offset)
{
	struct jpeg_decompress_struct cinfo;
	struct my_error_mgr jerr;
	JSAMPROW srow;
	SDL_Surface* surface;
	Uint32 rmask, gmask, bmask, amask;
	FILE* fp;
	int i;

	assert( file );
	assert( offset >= 0 );

	fp = fopen( file, "rb" );
	if ( fp == NULL ) {
		return NULL;
	}

	if (fseek(fp, offset, SEEK_SET)) {
		fclose(fp);
		return NULL;
	}

	cinfo.err = jpeg_std_error( &jerr.pub );
	jerr.pub.error_exit = my_error_exit;
	jerr.pub.output_message = my_output_message;

	
	if( setjmp(jerr.setjmp_buffer) ) {
		jpeg_destroy_decompress( &cinfo );
		fclose(fp);
		return NULL;
	}

	jpeg_create_decompress( &cinfo );
	jpeg_stdio_src( &cinfo, fp );

	jpeg_read_header(&cinfo, TRUE);

	
	cinfo.out_color_space = JCS_RGB;
	cinfo.dct_method = JDCT_ISLOW;

    
	if ( scale >= SDL_JPEG_SCALE_HALF && scale <= SDL_JPEG_SCALE_EIGHTH ) {
		cinfo.scale_num = 1;
		cinfo.scale_denom = (unsigned int)scale;
	}

	jpeg_calc_output_dimensions(&cinfo);

	if ( (cinfo.image_width != 160) || (cinfo.image_height != 120) ) {
		jpeg_destroy_decompress( &cinfo );
		fclose(fp);
		return NULL;		
	}

#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    rmask = 0xff000000;
    gmask = 0x00ff0000;
    bmask = 0x0000ff00;
    amask = 0x00000000;
#else
    rmask = 0x000000ff;
    gmask = 0x0000ff00;
    bmask = 0x00ff0000;
    amask = 0x00000000;
#endif

	
	surface = SDL_CreateRGBSurface( SDL_SWSURFACE,
									cinfo.output_width,
									cinfo.output_height,
									24, 
									rmask, gmask, bmask, amask );
	if ( surface == NULL ) {
		jpeg_destroy_decompress( &cinfo );
		fclose(fp);
		return NULL;		
	}

	
	jpeg_start_decompress( &cinfo );

	
	if ( SDL_MUSTLOCK(surface) ) {
		SDL_LockSurface(surface);
	}

	
	srow = (JSAMPLE*)surface->pixels;
	for ( i = 0; i < cinfo.output_height; i++ ) {
		jpeg_read_scanlines( &cinfo, &srow, 1 );
		srow += surface->pitch;
	}

	
	if ( SDL_MUSTLOCK(surface) ) {
		SDL_UnlockSurface(surface);
	}

	
	jpeg_destroy_decompress( &cinfo );
	fclose(fp);

	return surface;
}
