/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2004. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <assert.h>
#include <string.h>
#include "SDL_jpeg.h"
#include "SDL_png.h"
#include "SDL_gif.h"
#include "SDL_util.h"

typedef enum {
	IMAGETYPE_UNKNOWN,	
	IMAGETYPE_BMP,		
	IMAGETYPE_JPG,		
	IMAGETYPE_PNG,		
	IMAGETYPE_GIF		
} FileFormat;


static FileFormat
getImageType( const char* filename )
{
	FileFormat f = IMAGETYPE_UNKNOWN;
	const char* cur;

	assert(filename != 0);

	
	cur = filename + strlen(filename) - 1;
	while( *cur != '.' && cur > filename )
		cur--;

	
	if ( cur == filename )
		return IMAGETYPE_UNKNOWN;

	
	cur++;

	if (strcasecmp(cur, "bmp") == 0) {
		f = IMAGETYPE_BMP;
	} else if (strcasecmp(cur, "jpg") == 0 || 
			   strcasecmp(cur, "jpeg") == 0 ||
			   strcasecmp(cur, "jps") == 0 ) {
		f = IMAGETYPE_JPG;
	} else if (strcasecmp(cur, "png") == 0) {
		f = IMAGETYPE_PNG;
	} else if (strcasecmp(cur, "gif") == 0) {
		f = IMAGETYPE_GIF;
	}

	return f;
}






static SDL_Surface*
doLoadImage(const char* file,
	    SDL_Rect* dstrect,
	    SDL_RotateAngle angle,
	    SDL_StretchMode mode,
	    SDL_bool keepRatio)
{
	assert(file != 0);
	
	
	switch(getImageType(file)) {
	case IMAGETYPE_BMP:
		return SDL_LoadBMP(file);
		break;

	case IMAGETYPE_JPG:
		return SDL_LoadJPGLowMem(file, dstrect, angle, mode, keepRatio);
		break;

	case IMAGETYPE_PNG:
		return SDL_LoadPNG(file);
		break;

	case IMAGETYPE_GIF:
		return SDL_LoadGIF(file);
		break;

	default:
		return NULL;
		break;
	}
}



SDL_Surface*
SDL_LoadImage( const char* file,
			   SDL_Rect* dstrect,
			   SDL_RotateAngle angle,
			   SDL_StretchMode mode,
			   SDL_bool keepRatio)
{
	SDL_Rect rect;
	SDL_Surface* surf;
	SDL_Surface* rotatedSurf;
	SDL_Surface* scaledSurf;
	
	assert(file != 0);

	surf = doLoadImage(file, dstrect, angle, mode, keepRatio);
	if (surf == NULL) {
		return NULL;
	}

	if ( angle == SDL_ROTATE_0 ) {
		
		rotatedSurf = surf;
	} else {
		
		rotatedSurf = SDL_RotateSurface(surf, angle);
		SDL_FreeSurface(surf);
		if (rotatedSurf == NULL) {
			return NULL;
		}
	}

	if (dstrect == NULL) {
		
		return rotatedSurf;
	} else if (((dstrect->w == rotatedSurf->w) &&
			(dstrect->h == rotatedSurf->h)) ||
			(dstrect->w == 0 || dstrect->h == 0)) {
		
		return rotatedSurf;
	} else if (keepRatio == SDL_TRUE){
		
		
		rect.x = rect.y = 0;
#if 0
		if (rotatedSurf->w > rotatedSurf->h) {
			if (dstrect->w > dstrect->h) {
			 	rect.w = dstrect->w;
				rect.h = (dstrect->w * rotatedSurf->h) / rotatedSurf->w;
			} else {
			 	rect.w = dstrect->w;
				rect.h = (dstrect->w * rotatedSurf->h) / rotatedSurf->w;
			}
		} else {
			if (dstrect->w > dstrect->h) {
				rect.w = (dstrect->h * rotatedSurf->w) / rotatedSurf->h; 
				rect.h = dstrect->h;
			} else {
			 	rect.w = dstrect->w;
				rect.h = (dstrect->w * rotatedSurf->h) / rotatedSurf->w;
			}
		}
#else
		if (((float)(rotatedSurf->w) / (float)(rotatedSurf->h)) >
			((float)(dstrect->w) / (float)(dstrect->h))) {
		 	rect.w = dstrect->w;
			rect.h = (dstrect->w * rotatedSurf->h) / rotatedSurf->w;
		} else {
			rect.w = (dstrect->h * rotatedSurf->w) / rotatedSurf->h; 
			rect.h = dstrect->h;
		}
#endif
	} else {
		rect.x = 0; rect.y = 0;
		rect.w = dstrect->w;
		rect.h = dstrect->h;
	}

	if ((rect.w == rotatedSurf->w) && (rect.h == rotatedSurf->h)) {
	  return rotatedSurf;  
	}

	
	scaledSurf = SDL_CreateRGBSurface(SDL_SWSURFACE, rect.w, rect.h,
					  rotatedSurf->format->BitsPerPixel,
					  rotatedSurf->format->Rmask,
					  rotatedSurf->format->Gmask,
					  rotatedSurf->format->Bmask,
					  rotatedSurf->format->Amask);
	

	if ( rotatedSurf->format->palette && 
		 rotatedSurf->format->palette->ncolors > 0 )
	{
		int ret;

		
		ret = SDL_SetColors( scaledSurf, rotatedSurf->format->palette->colors,
							  0, rotatedSurf->format->palette->ncolors );

		if ( ret != 1 ) {
			SDL_FreeSurface( scaledSurf );
			SDL_FreeSurface( rotatedSurf );
			return NULL;
		}

		if ((rotatedSurf->flags & SDL_SRCCOLORKEY) == SDL_SRCCOLORKEY) {
			ret = SDL_SetColorKey(scaledSurf, SDL_SRCCOLORKEY, 
			                      rotatedSurf->format->colorkey);
			if ( ret != 0 ) {
				SDL_FreeSurface( scaledSurf );
				SDL_FreeSurface( rotatedSurf );
				return NULL;
			}
		}
	} 

	if ( (mode == SDL_STRETCH_BILINEAR) && (scaledSurf->format->Amask != 0) ) {
		mode = SDL_STRETCH_BILINEAR_LI;
	}
	
	
	if (SDL_StretchBlit(rotatedSurf, NULL, scaledSurf, NULL, mode) == -1) {	   
		SDL_FreeSurface(scaledSurf);
		scaledSurf = NULL;
	}
	SDL_FreeSurface(rotatedSurf);


	return scaledSurf;
}





int
SDL_LoadAndBlit( const char* file, 
				 SDL_Surface* dst,
				 SDL_Rect* dstrect,
				 SDL_RotateAngle angle,
				 SDL_StretchMode mode )
{
	int ret;
	SDL_Rect rect;
	SDL_Surface* surf;
	SDL_Surface* rotatedSurf;
	SDL_bool keepRatio = SDL_FALSE ; 
	SDL_Surface* formattedSurf;	
	assert(file != 0);
	assert(dst != 0);
	assert(dstrect != 0);
	

	surf = doLoadImage(file, dstrect, angle, mode, keepRatio) ; 
	if (surf == NULL) {
		return -1;
	}

	if ( angle == SDL_ROTATE_0 ) {
		
		rotatedSurf = surf;
	} else {
		
		rotatedSurf = SDL_RotateSurface(surf, angle);
		SDL_FreeSurface(surf);
		if (rotatedSurf == NULL) {
			return -1;
		}
	}

	if (dstrect == NULL) {
		
		ret = SDL_BlitSurface(rotatedSurf, NULL, dst, NULL);
		SDL_FreeSurface(rotatedSurf);	
		return ret;	   

	} else {
		rect.x = dstrect->x; 
		rect.y = dstrect->y;
		rect.w = dstrect->w;
		rect.h = dstrect->h;
	}

	if (rotatedSurf->format->BytesPerPixel != 1) {
		formattedSurf = rotatedSurf;
	} else {
		formattedSurf = SDL_ConvertSurface(rotatedSurf, dst->format,
		                                   SDL_SWSURFACE);
		SDL_FreeSurface(rotatedSurf);
		if (formattedSurf == NULL) {
			return -1;
		}

	}
	rotatedSurf = NULL;

	ret = SDL_StretchBlit(formattedSurf, NULL, dst, &rect, mode);
	SDL_FreeSurface(formattedSurf);	

	return ret;
}



int SDL_TilingBlit( SDL_Surface* tile,
					SDL_Surface* dst,
					SDL_Rect* dstrect)
{
	 int i, j;
	 int fx, fy;
	 int result;
	 SDL_Rect rect;
	 SDL_Rect drawable;
	if (!tile || !dst) {
		return -1;
	}

	if (!dstrect) {
		drawable.x = drawable.y = 0;
		drawable.w = dst->w;
		drawable.h = dst->h;
	} else {
		memcpy(&drawable, dstrect, sizeof(drawable));
		
		SDL_SetClipRect(dst, dstrect);
	}

	fx = dstrect->w / tile->w;
	if (fx == 0) {
		fx = 1;
	}

	fy = dstrect->h / tile->h;
	if (fy == 0) {
		fy = 1;
	}

	rect.w = tile->w;
	rect.h = tile->h;
	rect.y = dstrect->y;
	for (j = 0; j < fx; j++) {
		rect.x = dstrect->x;
		for (i = 0; i < fx; i++) {
			result = SDL_BlitSurface(tile, NULL, dst, &rect);
			if (result != 0) {
				SDL_SetClipRect(dst, NULL);
				return result;
			}
			rect.x += tile->w;
		}
		rect.y += tile->h;
	}
	SDL_SetClipRect(dst, NULL);
	return 0;
}


int SDL_GetImageDimension( const char *fname, int *width, int *height)
{
	assert(fname);
	assert(width);
	assert(height);

	
	switch(getImageType(fname)) {
		
	case IMAGETYPE_BMP:
		
		return -1;
		break;

	case IMAGETYPE_JPG:
		return SDL_JPGDimension(fname, width, height);
		break;

	case IMAGETYPE_PNG:
		
		return -1;
		break;

	default:
		break;
	}

	return -1;
}



