/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2004. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <unistd.h>
#include <assert.h>
#include <stdlib.h>
#include "SDL.h"

#include "SDL_jpeg.h"
#include "SDL_rotate.h"
#include "SDL_stretch.h"
#include "SDL_util.h"

SDL_Surface*	g_video;

static char*	stretchModeStrings[] = { "nearest-neighbour", "bilinear" };


static void
TestRotateAngles( SDL_Surface* src )
{
	SDL_Surface*	dstRot;
	SDL_RotateAngle	angle;
	SDL_Rect		srect;

	srect.x = 0; srect.y = 0;
	srect.w = g_video->w; srect.h = g_video->h;

	printf( "Testing each rotation angle... " );
	fflush( stdout );

	for ( angle = SDL_ROTATE_0; angle <= SDL_ROTATE_270; angle++ ) {
		dstRot = SDL_RotateSurface( src, angle );
		SDL_BlitSurface( dstRot, NULL, g_video, &srect );
		SDL_UpdateRect( g_video, 0, 0, 0, 0 );
		SDL_FreeSurface( dstRot );
		sleep(1);
	}
	printf( "successful\n" );
}


static void
TestRotateSpeed( SDL_Surface* src, SDL_RotateAngle angle )
{
	int				i;
	Uint32			tickStart, tickEnd;
	SDL_Surface*	dstRotNew;
	SDL_Surface*	dstRotOld;	
	char*			rotStrings[] = { "0", "90", "180", "270" };

	assert( src != NULL );

	printf( "Testing %s-degree surface rotate speed... ", rotStrings[angle]);
	fflush( stdout );

	dstRotOld = SDL_RotateSurface( src, angle );
	tickStart = SDL_GetTicks();

	for ( i = 0; i < 100; i++ ) {
		dstRotNew = SDL_RotateSurface( dstRotOld, angle );
		SDL_FreeSurface( dstRotOld );
		dstRotOld = dstRotNew;
	}

	tickEnd = SDL_GetTicks();
	SDL_FreeSurface( dstRotNew );

	printf( "%.2f ms/rotate\n", (tickEnd-tickStart)/100.0 );
}

static void
TestStretchBlit( SDL_Surface* src, SDL_StretchMode mode )
{
	const int	steps = 100;
	int			i;
	int			ret = 0;
	SDL_Rect	drect;

	printf( "Testing horizontal stretching (using %s algorithm)... ", 
			stretchModeStrings[mode] ); fflush( stdout );

	drect.x = 0; drect.y = 0;
	drect.w = g_video->w; drect.h = g_video->h;
	for ( i = 1; i < steps; i++ ) {
		drect.w = i * g_video->w / steps;
		if ( drect.w < 2 ) drect.w = 2;

		ret |= SDL_StretchBlit( src, NULL, g_video, &drect, mode );
		SDL_UpdateRect( g_video, 0, 0, 0, 0 );

		if ( ret != 0 ) {
			printf(" (i=%d) ", i );
			ret = 0;
		}
	}
	if ( ret != 0 ) {
		printf( "failed\n");
	} else {
		printf( "successful\n" );
	}


	
	SDL_FillRect( g_video, NULL, 0 );

	printf( "Testing vertical stretching (using %s algorithm)... ", 
			stretchModeStrings[mode] ); fflush( stdout );

	ret = 0;
	drect.x = 0; drect.y = 0;
	drect.w = g_video->w; drect.h = g_video->h;
	for ( i = 1; i < steps; i++ ) {
		drect.h = i * g_video->h / steps ;
		if ( drect.h < 2 ) drect.h = 2;

		ret |= SDL_StretchBlit( src, NULL, g_video, &drect, mode );
		SDL_UpdateRect( g_video, 0, 0, 0, 0 );

		if ( ret != 0 ) {
			printf(" (i=%d) ", i );
			ret = 0;
		}
	}
	if ( ret != 0 ) {
		printf( "failed\n");
	} else {
		printf( "successful\n" );
	}


	
	SDL_FillRect( g_video, NULL, 0 );

	printf( "Testing diagonal stretching (using %s algorithm)... ", 
			stretchModeStrings[mode] ); fflush( stdout );

	ret = 0;
	drect.x = 0; drect.y = 0;
	for ( i = 1; i < steps; i++ ) {
		drect.h = i * g_video->h / steps;
		if ( drect.h < 2 ) drect.h = 2;

		drect.w = i * g_video->w / steps;
		if ( drect.w < 2 ) drect.w = 2;

		ret |= SDL_StretchBlit( src, NULL, g_video, &drect, mode );
		SDL_UpdateRect( g_video, 0, 0, 0, 0 );

		if ( ret != 0 ) {
			printf(" (i=%d) ", i );
			ret = 0;
		}
	}
	if ( ret != 0 ) {
		printf( "failed\n");
	} else {
		printf( "successful\n" );
	}
}

static void
TestStretchBlitCoords( SDL_Surface* src, SDL_StretchMode mode )
{
	int			i;
	int			ret = 0;
	SDL_Rect	drect, srect;

	ret = 0;
	drect.x = 0; drect.y = 0;
	drect.w = (g_video->w/2);
	drect.h = src->h * drect.w / src->w;

	srect.x = 0; srect.y = 0;
	srect.w = 0; srect.h = 0;

	printf( "Testing coordinate stretching (using %s algorithm)... ", 
			stretchModeStrings[mode] ); fflush( stdout );

	
	for ( i = 0; i < (g_video->w - drect.w); i++ ) {
		drect.x = i;

		srect.w = i * src->w / (g_video->w - drect.w);
        if ( srect.w < 1 ) srect.w = 1;

		srect.h = i * src->h / (g_video->w - drect.w);
        if ( srect.h < 1 ) srect.h = 1;

		SDL_FillRect( g_video, NULL, 0 );
		ret |= SDL_StretchBlit( src, &srect, g_video, &drect, mode );
		SDL_UpdateRect( g_video, 0, 0, 0, 0 );
	}

	
	for ( i = 0; i < (g_video->h - drect.h); i++ ) {
		drect.y = i;
		SDL_FillRect( g_video, NULL, 0 );
		ret |= SDL_StretchBlit( src, &srect, g_video, &drect, mode );
		SDL_UpdateRect( g_video, 0, 0, 0, 0 );
	}

	
	for ( i = (g_video->w - drect.w); i >= 0; i-- ) {
		drect.x = i;
		SDL_FillRect( g_video, NULL, 0 );
		ret |= SDL_StretchBlit( src, &srect, g_video, &drect, mode );
		SDL_UpdateRect( g_video, 0, 0, 0, 0 );
	}

	
	for ( i = (g_video->h - drect.h); i >= 0; i-- ) {
		drect.y = i;
		SDL_FillRect( g_video, NULL, 0 );
		ret |= SDL_StretchBlit( src, &srect, g_video, &drect, mode );
		SDL_UpdateRect( g_video, 0, 0, 0, 0 );
	}

	if ( ret != 0 ) {
		printf( "failed\n");
	} else {
		printf( "successful\n" );
	}
}

static void
TestLoadImage(const char* filename)
{
	SDL_Surface* s;
	SDL_Rect r;

	r.x = 0; r.y = 0;
	r.w = g_video->w; r.h = g_video->h;
	s = SDL_LoadImage(filename, &r, SDL_ROTATE_0, SDL_STRETCH_BILINEAR, SDL_FALSE);
	if ( s == NULL ) {
		fprintf(stderr, "SDL_LoadImage: Failed to load %s\n", filename);
		return;
	} else {
		fprintf(stderr, "SDL_LoadImage: Load %s successfully\n", filename);
	}
	SDL_FillRect( g_video, NULL, 0 );
	SDL_BlitSurface( s, NULL, g_video, NULL );
	SDL_UpdateRect( g_video, 0, 0, 0, 0 );

	SDL_FreeSurface(s);
}


static void
TestLoadAndBlitImage(const char* filename)
{
	int ret;
	SDL_Rect r;

	r.x = 0; r.y = 0;
	r.w = g_video->w; r.h = g_video->h;
	SDL_FillRect( g_video, NULL, 0 );
	ret = SDL_LoadAndBlit(filename, g_video, &r, SDL_ROTATE_0, SDL_STRETCH_BILINEAR);
	if ( ret == -1 ) {
		fprintf(stderr, "SDL_LoadAndBlit: Failed to load %s\n", filename);
		return;
	} else {
		fprintf(stderr, "SDL_LoadAndBlit: Load %s successfully\n", filename);
	}
	SDL_UpdateRect( g_video, 0, 0, 0, 0 );
}

int
main( int argc, char *argv[] )
{
	int i;

	if ( argc <= 1 ) {
		printf("Usage: %s <filename>\n", argv[0]);
		return -1;
	}

	SDL_Init(SDL_INIT_VIDEO);
	atexit(SDL_Quit);

	g_video = SDL_SetVideoMode( 640, 480, 32, SDL_HWSURFACE );
	if ( g_video == NULL ) {
		return -2;
	}

	for ( i = 1; i < argc; i++ ) {
		SDL_Surface*	test;
		SDL_Surface*	converted;

		
		TestLoadImage( argv[i] );

		
		TestLoadAndBlitImage( argv[i] );

		
		printf( "Testing file loading (%s)... ", argv[i] );
		fflush( stdout );
		test = SDL_LoadImage( argv[i], NULL, SDL_ROTATE_0, SDL_STRETCH_NEAREST, SDL_FALSE);
		if ( test ) {
			printf( "successful\n" );
		} else {
			printf( "failed\n" );
			return -3;
		}

		
		converted = SDL_ConvertSurface( test, g_video->format, SDL_SWSURFACE );
		if ( converted == NULL ) {
			printf("Could not convert image surface\n");
			return -4;
		}

		SDL_FreeSurface( test );
		test = converted;

		
		TestRotateSpeed( test, SDL_ROTATE_0 );
		TestRotateSpeed( test, SDL_ROTATE_90 );
		TestRotateSpeed( test, SDL_ROTATE_180 );
		TestRotateSpeed( test, SDL_ROTATE_270 );

		
		SDL_FillRect( g_video, NULL, 0 );
		TestRotateAngles( test );
			
		
		SDL_FillRect( g_video, NULL, 0 );
		TestStretchBlit( test, SDL_STRETCH_NEAREST );

		
		SDL_FillRect( g_video, NULL, 0 );
		TestStretchBlitCoords( test, SDL_STRETCH_NEAREST );

		SDL_FreeSurface( test );
	}

	return 0;
}
