/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2004. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdlib.h>
#include <string.h>
#include "config.h"


#include "SDL_util.h"

#include "UIRsrcMgr.h"


UIRsrcMgr*	UIRsrcMgr::instance = NULL;


#define DEFAULT_LANG	LangEnglish

#include "UIRsrc.cpp"


const char *LANG_IDS[] = { "_e", "_j", "_f", "_g", "_i", "_s", "_p", "_c", "_k"};

const unsigned int DEFAULT_BOARD_ID = 0x5; 



UIRsrcMgr::
UIRsrcMgr	()
{
	m_lang = LangEnglish;
	m_using_module = ModNone;

	 char *value = getenv("BOARDID");
	if (value) {
		m_unBoardID = strtoul(value, 0, 16);
	} else {
		m_unBoardID = DEFAULT_BOARD_ID;
	}
}



UIRsrcMgr::
~UIRsrcMgr	()
{
	Dispose();
}

int UIRsrcMgr::
Load		(rsrc_file_t& rsrc_file)
{
	if (rsrc_file.surface)
		return 0;

	 char tmp[FILENAME_MAX + 1];
	assert(sizeof(tmp) > (sizeof(RESOURCE_DIR) + strlen(rsrc_file.file)));
	memset(tmp, 0, sizeof(tmp));
	strcpy(tmp, RESOURCE_DIR);

	if (rsrc_file.isLocalizable) {
		 char *p = strchr(rsrc_file.file, '.');
		if (!p) {
			return -1;
		}
		strcat(tmp, LANG_IDS[m_lang]);
		strcat(tmp, "/");		
		strncat(tmp, rsrc_file.file, p - rsrc_file.file);
		strcat(tmp, LANG_IDS[m_lang]);
		strcat(tmp, p);
	} else {
		strcat(tmp, rsrc_file.file);
	}

	DebugF("Load %s\n", tmp);
	rsrc_file.surface = ::SDL_LoadImage(tmp, NULL, SDL_ROTATE_0, SDL_STRETCH_BILINEAR, SDL_TRUE);
	if (!rsrc_file.surface) {
		DebugF("cannot load %s.\n", tmp);
		return -1;
	}
	return 0;
}


int UIRsrcMgr::
Init		(Module using_module)
{
	int failed = 0;
	int result;

	m_using_module = using_module;
	for (int i = 0; i < (int)(sizeof(rsrc_files)/sizeof(rsrc_file_t)); i++) {
		if (!(rsrc_files[i].align & using_module) || rsrc_files[i].surface ||
			rsrc_files[i].isLocalizable == true) {
			continue;
		}

		result = Load(rsrc_files[i]);
		if (result != 0) {
			DebugF("%s(%d): Unable to load resource: %s\n", 
				   __FILE__, __LINE__, rsrc_files[i].file);
			failed = result;
		}
	}
	for (int i = 0; i < (int)(sizeof(rsrc_files)/sizeof(rsrc_file_t)); i++) {
		if (!(rsrc_files[i].align & using_module) || rsrc_files[i].surface ||
			rsrc_files[i].isLocalizable == false) {
			continue;
		}

		result = Load(rsrc_files[i]);
		if (result != 0) {
			DebugF("%s(%d): Unable to load resource: %s\n", 
				   __FILE__, __LINE__, rsrc_files[i].file);
			failed = result;
		}
	}
	return failed;
}


void UIRsrcMgr::
DisposeLocalizeRsrc		()
{
	for (int i = 0; i < (int)(sizeof(rsrc_files)/sizeof(rsrc_file_t)); i++) {
		if (rsrc_files[i].surface && rsrc_files[i].isLocalizable) {
			SDL_FreeSurface(rsrc_files[i].surface);
			rsrc_files[i].surface = NULL;
		}
	}
}

void UIRsrcMgr::
Dispose		()
{
	for (int i = 0; i < (int)(sizeof(rsrc_files)/sizeof(rsrc_file_t)); i++) {
		if (rsrc_files[i].surface) {
			SDL_FreeSurface(rsrc_files[i].surface);
			rsrc_files[i].surface = NULL;
		}
	}
}


::SDL_Surface* UIRsrcMgr::
GetResource	(Rsrc id)
{
	
	if (m_unBoardID == 0x6) {	
		switch (id) {
		case navIconEnter:
			id = navIconEnterCardType; 
			break;
		case navIconEscape:
			id = navIconEscapeCardType;
			break;
		default:
			break;
		}
	}

	if (id >= (Rsrc)(sizeof(rsrc_files)/sizeof(rsrc_file_t))) {
		return NULL;
	}

	if (rsrc_files[id].surface == NULL) {
		Load(rsrc_files[id]);
	}

	return rsrc_files[id].surface;
}

int UIRsrcMgr::
SetLang	(const Language& lang)
{
	if (m_lang == lang) {
		return 0;
	}
	
	m_lang = lang;

	
	DisposeLocalizeRsrc();

	
	 int result = Init(m_using_module);
	if (result != 0) {
		return result;
	}

	return 0;
}

Language UIRsrcMgr::
E21Lang2UILang(Uint8 lang)
{
	const static struct lang_conv {
		Uint8 lang;
		Language ui_lang;
	} langs[] = {
		{0x01, LangEnglish},
		{0x02, LangJapanese},
		{0x03, LangFrench},
		{0x04, LangGermany},
		{0x05, LangItaly},
		{0x06, LangSpanish},
		{0x07, LangPortuguese},
		{0x08, LangChinese},
		{0x09, LangKorean}
	};
	for (unsigned int i = 0; i < (sizeof(langs)/sizeof(lang_conv)); i++) {
		if (langs[i].lang == lang) {
			return langs[i].ui_lang;
		}
	}

	return DEFAULT_LANG;
}

int UIRsrcMgr::
SetLang	(const Uint8 lang)
{
	return SetLang(E21Lang2UILang(lang));
}
