/* codec_faac.c
 * Copyright (C) 2001 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_faac.c,v 1.1.1.1 2003/11/28 06:03:45 ahu Exp $
 */


#include <stdlib.h>
#include <string.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"

#include <faad.h>
#ifdef USEFAAC
#include <faac.h>
#endif

/* Some static global variables */
static int Initialised = 0;           /* Initialising counter */

#define IMHERE fprintf(stderr, "DEBUG: line %d\n", __LINE__)

/* Each instance of the codec should have its own Param structure allocated */
typedef struct {
  int use_esds;
  int decoder_initialized;
  faacDecHandle hDecoder;
  unsigned long samplerate;
  unsigned char channels;
  /* TODO: Encoding is work in progress and as such it does not work at all */
#ifdef USEFAAC
  int encoder_initialized;
  faacEncHandle hEncoder;
  unsigned long samplesInput; unsigned long maxBytesOutput;
  int bitrate;
  int bandwidth;
#endif
} Param;

static const char    Signature[5]  = "FAAC";
#ifdef USEFAAC
static const char   *Name = "Freeware Advanced Audio Codec";
#else
static const char   *Name = "Freeware Advanced Audio Decoder";
#endif
static const char    AltSignature[5]  = "mp4a";
static const char   *AltName = "MPEG-4 Audio Decoder";
static const char   *Version  = "0.1";
static const char   *Copyright  = "2003 OpenQuicktime Team";
static const char   *ModuleAuthor  = "Jozef Hatala";
static const char   *CodecAuthor  = "Copyright (C) 2001,2002 Menno Bakker";

#ifdef USEFAAC
static const oqt_codec_param_t params[] =
  {
    { "bitrate", OQT_PARAMETER_INTEGER, "The bitrate of the target encoded stream"},
    { "badwidth", OQT_PARAMETER_INTEGER, "The cutoff frequency"}
  };

#define PARAM_COUNT (sizeof(params)/sizeof(params[0]))

static const oqt_codec_param_t* list_params_FAAC(int *o_numparams)
{
  *o_numparams = PARAM_COUNT;
	
  return params;
}

static int set_param_FAAC(oqt_t* file,
			  int          track,
			  const char*  key,
			  const void*  value)
{
  Param *p = (Param*)(file->vtracks[track].codec_private);

  if(!strcasecmp(key, "bitrate")) {
    p->bitrate = *(int*)value;
  } else if(!strcasecmp(key, "bandwidth")) {
    p->bandwidth = *(int*)value;
  } else {
    fprintf(stderr,"set_param_FAAC: unknown parameter named '%s'\n",key);
    return OQT_PARAMETER_UNKNOWN;
  }
  return 0;
}


static int get_param_FAAC(oqt_t* file,
			  int          track,
			  const char*  key,
			  void*        value)
{
  Param *p = (Param*)(file->vtracks[track].codec_private);

  if(!strcasecmp(key, "bitrate")) {
    *(int*)value = p->bitrate;
    return OQT_PARAMETER_INTEGER;
  } else if(!strcasecmp(key, "bandwidth")) {
    *(int*)value = p->bandwidth;
    return OQT_PARAMETER_INTEGER;
  } else {
    fprintf(stderr,"get_param_FAAC: unknown parameter named '%s'\n",key);
    return OQT_PARAMETER_UNKNOWN;
  }
}
#endif

static
int decode_faac(oqt_t *file, int track,
		oqt_uint32_t inputsize,
		oqt_uint8_t  *input,
		oqt_uint32_t outputsize,
		oqt_uint8_t  *output)
{
  Param *p = (Param*)(file->atracks[track].codec_private);
  int inputpos = 0, outputpos = 0;
  faacDecFrameInfo frameInfo;

  if(!p->decoder_initialized) {
    p->decoder_initialized = 1;
    p->hDecoder = faacDecOpen();

    if(p->use_esds) {
      oqt_int32_t entries = \
	file->atracks[track].track->mdia.minf.stbl.stsd.total_entries;
      oqt_stsd_table_t *table = \
	file->atracks[track].track->mdia.minf.stbl.stsd.table;

      if(entries < 1 || !table || !table->esds.decoderConfigLen) {
	fprintf(stderr, "decode_faac: missing esds atom for mp4a\n");
	p->samplerate = oqt_get_audio_samplerate(file, track);
	p->channels = oqt_get_audio_channels(file, track);

      } else {
	if(faacDecInit2(p->hDecoder, table->esds.decoderConfig,
			table->esds.decoderConfigLen,
			&p->samplerate, &p->channels) < 0) {
	  fprintf(stderr, "decode_faac: failed to parse decoder config for mp4a\n");
	  p->samplerate = oqt_get_audio_samplerate(file, track);
	  p->channels = oqt_get_audio_channels(file, track);
	}
      }

      if(oqt_get_audio_samplerate(file, track) != p->samplerate ||
	 oqt_get_audio_channels(file, track) != p->channels) {
	fprintf(stderr, "decode_faac: decoder config does not match movie header\n");
      }

    } else {
      int res;

      res = faacDecInit(p->hDecoder, input, &p->samplerate, &p->channels);
      if(res < 0) {
	fprintf(stderr, "decode_faac: failed to parse decoder config\n");
      }

      if(oqt_get_audio_samplerate(file, track) != p->samplerate ||
	 oqt_get_audio_channels(file, track) != p->channels) {
	fprintf(stderr, "decode_faac: decoder config does not match movie header\n");
      }

      inputpos = res;
    }
  }

  while(inputpos < inputsize && outputpos < outputsize) {
    int len;
    void *sample_buffer;

    /* this seems to help a lot */
    while(inputpos < inputsize && input[inputpos] == 0) {
      ++inputpos;
    }

    sample_buffer = faacDecDecode(p->hDecoder, &frameInfo, input+inputpos);
    if (frameInfo.error > 0) {
      fprintf(stderr, "decode_faac error: %s\n",
	      faacDecGetErrorMessage(frameInfo.error));

      return outputpos;
    }

    inputpos += frameInfo.bytesconsumed;
    len = p->channels*frameInfo.samples;
    if(outputpos+len > outputsize) {
      fprintf(stderr, "decode_faac: would overflow sample buffer!\n");
      return outputpos;
    }
    memcpy(output+outputpos, sample_buffer, len);
    outputpos += len;
  }

  return outputpos;
}

#ifdef USEFAAC
static int prepare_FAAC(oqt_t *file, int track)
{
  Param *p = (Param*)(file->atracks[track].codec_private);

  IMHERE;

  if(!p->encoder_initialized) {
    faacEncConfigurationPtr format;

    p->encoder_initialized = 1;
    p->samplerate = oqt_get_audio_samplerate(file, track);
    p->channels= oqt_get_audio_channels(file, track);
    p->hEncoder = faacEncOpen(p->samplerate, p->channels,
			      &p->samplesInput, &p->maxBytesOutput);
    if(!p->hEncoder) {
      fprintf(stderr, "prepare_FAAC: failed to create encoder\n");
      return -1;
    }
    format = faacEncGetCurrentConfiguration(p->hEncoder);

#if 0
    format->aacObjectType = LOW;
    format->mpegVersion = MPEG4;
    format->useTns = 0;
    format->allowMidside = 1;
    format->outputFormat = 0;
    format->bitRate = p->bitrate;
    format->bandWidth = p->bandwidth;
#endif

    if(faacEncSetConfiguration(p->hEncoder, format) == 0) {
      fprintf(stderr, "prepare_FAAC: failed to set format\n");
    }
  }

  return p->samplesInput;
}


static int
encode_FAAC(oqt_t *file, int track,
	   oqt_uint32_t num_samples,
	   oqt_uint8_t *input,
	   oqt_uint32_t out_size,
	   oqt_uint8_t *output)
{
  Param *p = (Param*)(file->atracks[track].codec_private);
  int sample_size = oqt_get_audio_bits(file, track)/8;
  int res;

  if(!p->encoder_initialized) {
    fprintf(stderr, "encode_FAAC: encoder not initialized\n");
    return 0;
  }

  res = faacEncEncode(p->hEncoder, input, num_samples * sample_size,
		      output, out_size);

  if(res <= 0) {
    fprintf(stderr, "encode_FAAC: encoding failed: %d\n", res);
    return 0;
  }

  return res;
}
#endif

static int
init_faac(oqt_t *file, int track)
{
  oqt_audio_map_t *atrack = &file->atracks[track];
  Param *p = malloc(sizeof(Param));

  memset(p, 0, sizeof(Param));
  atrack->codec_private = p;

  ++Initialised;
	
  if (!p) {
    fprintf(stderr, "Error: malloc failed when creating codec_private in init_faac\n");
    return 0;
  }

	
  return Initialised;
}

static int
init_mp4a(oqt_t *file, int track)
{
  oqt_audio_map_t *atrack = &file->atracks[track];
  Param *p = malloc(sizeof(Param));

  memset(p, 0, sizeof(Param));
  atrack->codec_private = p;

  ++Initialised;
	
  if (!p) {
    fprintf(stderr, "Error: malloc failed when creating codec_private in init_faac\n");
    return 0;
  }

  p->use_esds = 1;

#if USEFAAC
  // default
  p->bitrate = 96000;
  p->bandwidth = 18000;
#endif

  return Initialised;
}


static int
delete_faac(oqt_t *file, int track)
{
  oqt_audio_map_t *atrack = &file->atracks[track];
  Param *p = (Param*)(atrack->codec_private);
	
  if (p) {
    if(p->hDecoder) {
      faacDecClose(p->hDecoder);
    }
#ifdef USEFAAC
    if(p->hEncoder) {
      faacEncClose(p->hEncoder);
    }
#endif
    free(p);
    atrack->codec_private=NULL;
  }
	
  return --Initialised;
}



static void
set_standard_values(oqt_audio_codec_t *codec ) {

  codec->info.version = Version;
  codec->info.copyright = Copyright;
  codec->info.module_author = ModuleAuthor;
  codec->info.codec_author = CodecAuthor;
  codec->info.interface_age = OQT_INTERFACE_AGE;
	
  codec->delete_codec = delete_faac;
  codec->decode = decode_faac;
	
  codec->info.merit = 80;
}



oqt_codec_info_t*
oqt_plugin_register( void )
{
  oqt_audio_codec_t *codec,*alt;

  // Allocate the memory for the codec record
  codec = oqt_allocate_audio_codec();
  alt = oqt_allocate_audio_codec();
  if (!codec || !alt) return NULL;

  // Set the name and signature of the codec
  memcpy(codec->info.fourcc, Signature, 4);
  codec->info.name = Name;

  memcpy(alt->info.fourcc, AltSignature, 4);
  alt->info.name = AltName;
	
  alt->info.parent = &codec->info;
  codec->info.has_children = 1;
	
  // Set the callbacks
  set_standard_values(codec);
  codec->init_codec = init_faac;
#ifdef USEFAAC
  codec->encode = encode_FAAC;
  codec->list_params = list_params_FAAC;
  codec->set_param = set_param_FAAC;
  codec->get_param = get_param_FAAC;
  codec->prepare_encoder = prepare_FAAC;
#endif

  set_standard_values(alt);
  alt->init_codec = init_mp4a;
		
  // We only have to register the alternate codec
  if (oqt_register_codec(&alt->info)) {
    fprintf(stderr, "Failed to register the mp4a codec.\n");
  }
	
  return &codec->info;
}
