/* codec_divx.c
 * Copyright (C) 2002 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_divx.c,v 1.1.1.1 2003/11/28 06:03:46 ahu Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

// workaround
#define LINUX
#include <decore.h>
#undef BIG_ENDIAN

#include <encore2.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"


static int              Initialised   = 0;      

static const char   Signature[5]  = "DX50"; 
static const char   AltSignature4[5]  = "DIV4"; 
static const char   AltSignature3[5]  = "DIV3"; 
static const char   AltSignatureX[5]  = "DIVX"; 
static const char   AltSignatureMP4[5]  = "mp4v"; 
static const char   *Name = "DivX 5.xx"; 
static const char   *AltName4 = "DivX 4.13 compatible decompressor";
static const char   *AltName3 = "DivX ;-) 3.11 compatible decompressor";
static const char   *AltNameX = "OpenDIVX compatible decompressor";
static const char   *AltNameMP4 = "MPEG-4 Video compatible decompressor";
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 OpenQuicktime Team"; 
static const char   *ModuleAuthor  = "Jozef Hatala"; 
static const char   *CodecAuthor  = "DivXNetworks"; 

static const oqt_codec_param_t params[] =
{
  { "bitrate", OQT_PARAMETER_INTEGER, "The bitrate of the target encoded stream"},
  { "max_key_interval", OQT_PARAMETER_INTEGER, "The maximum interval between key frames"},
  { "quality", OQT_PARAMETER_INTEGER, "The quality of compression (1-fastest, 5-best)"}
};

#define PARAM_COUNT (sizeof(params)/sizeof(params[0]))

static int next_decore_handle = 1;
 
/* Each instance of the codec should have its own Private structure allocated */
typedef struct {
  // decoder
  int decore_initialized;
  int decore_handle;
  int use_esds;
  DEC_PARAM dec_param;
  unsigned char *dec_buffer;

  // encoder
  int encore_initialized;
  int encore_handle;
  ENC_PARAM enc_param;
  unsigned char *enc_buffer;
} Private;  

static const oqt_codec_param_t* list_params_DIV5(int *o_numparams)
{
	*o_numparams = PARAM_COUNT;
	
	return params;
}

static int set_param_DIV5(oqt_t* file,
			  int          track,
			  const char*  key,
			  const void*  value)
{
  Private *p = (Private*)(file->vtracks[track].codec_private);

  if(!strcasecmp(key, "bitrate"))
    p->enc_param.bitrate = *(int*)value;
  else if(!strcasecmp(key, "max_key_interval"))
    p->enc_param.max_key_interval = *(int*)value;
  else if(!strcasecmp(key, "quality"))
    p->enc_param.quality = *(int*)value;
  else {
    fprintf(stderr,"set_param_DIVX: unknown parameter named '%s'\n",key);
    return OQT_PARAMETER_UNKNOWN;
  }
  return 0;
}


static int get_param_DIV5(oqt_t* file,
			  int          track,
			  const char*  key,
			  void*        value)
{
  Private *p = (Private*)(file->vtracks[track].codec_private);

  if(!strcasecmp(key, "bitrate")) {
    *(int*)value = p->enc_param.bitrate;
    return OQT_PARAMETER_INTEGER;
  } else if(!strcasecmp(key, "max_key_interval")) {
    *(int*)value = p->enc_param.max_key_interval;
    return OQT_PARAMETER_INTEGER;
  } else if(!strcasecmp(key, "quality")) {
    *(int*)value = p->enc_param.quality;
    return OQT_PARAMETER_INTEGER;
  } else {
    fprintf(stderr,"get_param_DIVX: unknown parameter named '%s'\n",key);
    return OQT_PARAMETER_UNKNOWN;
  }
}

static int reads_colormodel_DIVX(oqt_t *file, 
		int track,
		int colormodel )
{
  return ( colormodel == BC_YUV420P );
}

static int preferred_colormodel_DIVX(oqt_t *file, 
		int track )
{
  return BC_YUV420P;
}


static int decode_DIVX(oqt_t *file,
		       int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t  *input,
		       oqt_uint8_t  **output,
		       int color_model)
{
  int res;
  Private *p = (Private*)(file->vtracks[track].codec_private);
  int width 	= oqt_get_video_width(file, track);
  int height 	= oqt_get_video_height(file, track);
  DEC_FRAME dec_frame;
  oqt_uint8_t *vol_buffer = NULL;

  dec_frame.bitstream = input;
  dec_frame.length = inputsize;
  dec_frame.stride = width;

  if(color_model == BC_NONE) {
    dec_frame.render_flag = 0;
  } else if(color_model == BC_YUV420P) {
    dec_frame.render_flag = 1;
  } else {
    fprintf(stderr,"decode_DIVX: unsupported color model\n");
    return 1;
  }

  if(!p->decore_initialized) {
    DEC_MEM_REQS dec_mem_reqs;

    p->decore_handle = next_decore_handle++;
    p->dec_param.x_dim = width;
    p->dec_param.y_dim = height;
    p->dec_param.output_format = DEC_YV12;
    p->dec_param.time_incr = 0;
    p->dec_param.build_number = 0;

    res = decore(p->decore_handle, DEC_OPT_MEMORY_REQS,
		 &p->dec_param, &dec_mem_reqs);
    if(res != DEC_OK) {
      fprintf(stderr, "decode_DIVX: DEC_OPT_MEMORY_REQS failed: %d\n", res);
      return 1;
    }

    p->dec_param.buffers.mp4_display_buffers =
      calloc(1, dec_mem_reqs.mp4_display_buffers_size);
    p->dec_param.buffers.mp4_edged_for_buffers =
      calloc(1, dec_mem_reqs.mp4_edged_for_buffers_size);
    p->dec_param.buffers.mp4_edged_back_buffers =
      calloc(1, dec_mem_reqs.mp4_edged_back_buffers_size);
    p->dec_param.buffers.mp4_edged_ref_buffers =
      calloc(1, dec_mem_reqs.mp4_edged_ref_buffers_size);
    p->dec_param.buffers.mp4_reference =
      calloc(1, dec_mem_reqs.mp4_reference_size);
    p->dec_param.buffers.mp4_state =
      calloc(1, dec_mem_reqs.mp4_state_size);
    p->dec_param.buffers.mp4_stream =
      calloc(1, dec_mem_reqs.mp4_stream_size);
    p->dec_param.buffers.mp4_tables =
      calloc(1, dec_mem_reqs.mp4_tables_size);

    res = decore(p->decore_handle, DEC_OPT_INIT, &p->dec_param, NULL);
    if(res != DEC_OK) {
      fprintf(stderr, "decode_DIVX: DEC_OPT_INIT failed: %d\n", res);
      free(p->dec_param.buffers.mp4_display_buffers);
      free(p->dec_param.buffers.mp4_edged_for_buffers);
      free(p->dec_param.buffers.mp4_edged_back_buffers);
      free(p->dec_param.buffers.mp4_edged_ref_buffers);
      free(p->dec_param.buffers.mp4_reference);
      free(p->dec_param.buffers.mp4_state);
      free(p->dec_param.buffers.mp4_stream);
      free(p->dec_param.buffers.mp4_tables);
      return 1;
    }

    p->dec_buffer = malloc(width*height + width*height/2);

    p->decore_initialized = 1;

    if(p->use_esds) {
      oqt_int32_t entries = \
        file->vtracks[track].track->mdia.minf.stbl.stsd.total_entries;
      oqt_stsd_table_t *table = \
        file->vtracks[track].track->mdia.minf.stbl.stsd.table;
      
      if(entries < 1 || !table) {
        fprintf(stderr, "decode_DIVX: missing esds atom for mp4v\n");

      } else {
        oqt_esds_t *esds = &(table->esds);

        p->use_esds = 0;
        vol_buffer = malloc(esds->decoderConfigLen + inputsize);
        memcpy(vol_buffer, esds->decoderConfig, esds->decoderConfigLen);
        memcpy(vol_buffer + esds->decoderConfigLen, input, inputsize);
        dec_frame.bitstream = vol_buffer;
        dec_frame.length = esds->decoderConfigLen + inputsize;
      }
    }
  }

  if(!dec_frame.render_flag) {
    dec_frame.bmp = NULL;
  } else if(output[2] == output[0] + width*height &&
     output[1] == output[2] + width*height/4) {
    dec_frame.bmp = output[0];
  } else {
    dec_frame.bmp = p->dec_buffer;
  }

  res = decore(p->decore_handle, DEC_OPT_FRAME, &dec_frame, NULL);
  if(res != DEC_OK) {
    fprintf(stderr, "decode_DIVX: DEC_OPT_FRAME failed: %d\n", res);
    return 1;
  }

  if(vol_buffer) {
    free(vol_buffer);
  }

  if(dec_frame.bmp == p->dec_buffer) {
    memcpy(output[0], p->dec_buffer, width*height);
    memcpy(output[2], p->dec_buffer + width*height,
	   width*height/4);
    memcpy(output[1], p->dec_buffer + width*height + width*height/4,
	   width*height/4);
  }

  return 0;
}

/* OpenDIVX has reversed U and V planes */
static int decode_OpenDIVX(oqt_t *file,
                       int track,
                       oqt_uint32_t inputsize,
                       oqt_uint8_t *input,
                       oqt_uint8_t **output,
                       int color_model)
{
  if(color_model == BC_NONE) {
    return decode_DIVX(file, track, inputsize, input, output, color_model);
  } else if(color_model == BC_YUV420P) {
    unsigned char *tmp_output[3] = { output[0], output[2], output[1] };
    return decode_DIVX(file, track, inputsize, input, tmp_output, color_model);
  } else {
    fprintf(stderr,"decode_OpenDIVX: unsupported color model (%d)\n", color_model);
    return 1;
  }
}


static int writes_colormodel_DIV5(oqt_t *file, 
	int track,
	int colormodel)
{ 
  return ( colormodel == BC_YUV420P );
}

static int encode_DIV5(oqt_t *file, 
				int track,
				oqt_uint8_t **input,
				oqt_uint32_t out_size,
				oqt_uint8_t *output,
				int *IsAKeyFrame,
				int color_model)
{
  int res;
  Private *p = (Private*)(file->vtracks[track].codec_private);
  int width = file->vtracks[track].track->tkhd.track_width;
  int height = file->vtracks[track].track->tkhd.track_height;
  ENC_FRAME enc_frame;
  ENC_RESULT enc_result;

  if(color_model != BC_YUV420P) {
    fprintf(stderr,"encode_DIVX: unsupported color model (%d)\n", color_model);
    return 0;
  }

  if(!p->encore_initialized) {
    p->enc_param.handle = NULL;
    p->enc_param.x_dim = width;
    p->enc_param.y_dim = height;
    p->enc_param.framerate = 
      (float)file->vtracks[track].track->mdia.mdhd.time_scale /
      file->vtracks[track].track->mdia.minf.stbl.stts.table[0].sample_duration;

    res = encore(NULL, ENC_OPT_INIT, &p->enc_param, NULL);
    if(res != ENC_OK) {
      fprintf(stderr, "encode_DIVX: ENC_OPT_INIT failed: %d\n", res);
      return 0;
    }
    
    p->encore_handle = p->enc_param.handle;

    p->enc_buffer = malloc(width*height + width*height/2);

    p->encore_initialized = 1;
  }

  enc_frame.bitstream = output;
  enc_frame.length = 0;
  enc_frame.colorspace = ENC_CSP_YV12;

  if(input[2] == input[0] + width*height &&
     input[1] == input[2] + width*height/4) {
    enc_frame.image = input[0];

  } else {
    memcpy(p->enc_buffer, input[0], width*height);
    memcpy(p->enc_buffer + width*height,
	   input[2], width*height/4);
    memcpy(p->enc_buffer + width*height + width*height/4,
	   input[1], width*height/4);

    enc_frame.image = p->enc_buffer;
  }

  res = encore(p->encore_handle, ENC_OPT_ENCODE, &enc_frame, &enc_result);
  if(res != ENC_OK) {
    fprintf(stderr, "encode_DIVX: ENC_OPT_ENCODE failed: %d\n", res);
    return 0;
  }

  *IsAKeyFrame = enc_result.is_key_frame;

  return enc_frame.length;
}


static int delete_DIVX(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Private *p = (Private*)(vtrack->codec_private);
	
	if (p) {
		if(p->encore_initialized) {
			encore(p->encore_handle, ENC_OPT_RELEASE, NULL, NULL);
		}
		if(p->decore_initialized) {
			decore(p->decore_handle, DEC_OPT_RELEASE, NULL, NULL);
			free(p->dec_param.buffers.mp4_display_buffers);
			free(p->dec_param.buffers.mp4_edged_for_buffers);
			free(p->dec_param.buffers.mp4_edged_back_buffers);
			free(p->dec_param.buffers.mp4_edged_ref_buffers);
			free(p->dec_param.buffers.mp4_reference);
			free(p->dec_param.buffers.mp4_state);
			free(p->dec_param.buffers.mp4_stream);
			free(p->dec_param.buffers.mp4_tables);
		}
		
		if(p->dec_buffer)
			free(p->dec_buffer);
		if(p->enc_buffer)
			free(p->enc_buffer);
		
		free(p);
		vtrack->codec_private = NULL;
	}
	
	return --Initialised;
}


static int init_DIV5(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Private *p = malloc(sizeof(Private));
	vtrack->codec_private = p;
	memset(p, 0, sizeof(Private));
	
	// defaults
	p->dec_param.codec_version = 500;
	p->enc_param.bitrate = 900000;
	p->enc_param.quality = 5;
	
	return ++Initialised;
}

static int init_DIV4(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Private *p = malloc(sizeof(Private));
	vtrack->codec_private = p;
	memset(p, 0, sizeof(Private));
	
	// defaults
	p->dec_param.codec_version = 412;
	p->enc_param.bitrate = 900000;
	p->enc_param.quality = 5;
	
	return ++Initialised;
}

static int init_DIV3(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Private *p = malloc(sizeof(Private));
	vtrack->codec_private = p;
	memset(p, 0, sizeof(Private));
	
	// defaults
	p->dec_param.codec_version = 311;
	p->enc_param.bitrate = 900000;
	p->enc_param.quality = 5;
	
	return ++Initialised;
}

static int init_mp4v(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Private *p = malloc(sizeof(Private));
	vtrack->codec_private = p;
	memset(p, 0, sizeof(Private));
	p->use_esds = 1;
	
	// defaults
	p->dec_param.codec_version = 500;
	p->enc_param.bitrate = 900000;
	p->enc_param.quality = 5;
	
	return ++Initialised;
}

static void fill_codec_struct(oqt_video_codec_t *codec)
{
  codec->info.version = Version;
  codec->info.copyright = Copyright;
  codec->info.module_author = ModuleAuthor;
  codec->info.codec_author = CodecAuthor;
  codec->info.interface_age = OQT_INTERFACE_AGE;
	
  codec->decode = decode_DIVX;
  codec->delete_codec = delete_DIVX;
  codec->reads_colormodel = reads_colormodel_DIVX;
  codec->preferred_colormodel = preferred_colormodel_DIVX;
}

oqt_codec_info_t*
oqt_plugin_register( void )
{
  oqt_video_codec_t *codec;
  oqt_video_codec_t *alt4;
  oqt_video_codec_t *alt3;
  oqt_video_codec_t *altX;
  oqt_video_codec_t *altMP4;

#if ENCORE_MAJOR_VERSION < 5000
# error "Expecting DivX version code >= 5000"
#endif

  codec = oqt_allocate_video_codec();
  if (!codec) return NULL;
  fill_codec_struct(codec);
  memcpy( &codec->info.fourcc, Signature, 4); 
  codec->info.has_children = 1;
  codec->info.name = Name;
  codec->init_codec = init_DIV5;
  codec->encode = encode_DIV5;
  codec->writes_colormodel = writes_colormodel_DIV5;
  codec->list_params = list_params_DIV5;
  codec->set_param = set_param_DIV5;
  codec->get_param = get_param_DIV5;

  alt4 = oqt_allocate_video_codec();
  if (!alt4) return NULL;
  fill_codec_struct(alt4);
  memcpy( &alt4->info.fourcc, AltSignature4, 4); 
  alt4->info.name = AltName4;
  alt4->info.parent = &codec->info;
  alt4->info.merit = 90;
  alt4->init_codec = init_DIV4;

  alt3 = oqt_allocate_video_codec();
  if (!alt3) return NULL;
  fill_codec_struct(alt3);
  memcpy( &alt3->info.fourcc, AltSignature3, 4); 
  alt3->info.name = AltName3;
  alt3->info.parent = &codec->info;
  alt3->info.merit = 90;
  alt3->init_codec = init_DIV3;

  altX = oqt_allocate_video_codec();
  if (!altX) return NULL;
  fill_codec_struct(altX);
  memcpy( &altX->info.fourcc, AltSignatureX, 4); 
  altX->info.name = AltNameX;
  altX->info.parent = &codec->info;
  altX->info.merit = 85;
  altX->init_codec = init_DIV4;
  altX->decode = decode_OpenDIVX;

  altMP4 = oqt_allocate_video_codec();
  if (!altMP4) return NULL;
  fill_codec_struct(altMP4);
  memcpy( &altMP4->info.fourcc, AltSignatureMP4, 4); 
  altMP4->info.name = AltNameMP4;
  altMP4->info.parent = &codec->info;
  altMP4->info.merit = 80;
  altMP4->init_codec = init_mp4v;
  altMP4->decode = decode_DIVX;

  if (oqt_register_codec(&alt3->info)) {
    fprintf(stderr, "Failed to register DIV3 video codec\n");
    return NULL;
  }

  if (oqt_register_codec(&alt4->info)) {
    fprintf(stderr, "Failed to register DIV4 video codec\n");
    return NULL;
  }

  if (oqt_register_codec(&altX->info)) {
    fprintf(stderr, "Failed to register DIVX video codec\n");
    return NULL;
  }

  if (oqt_register_codec(&altMP4->info)) {
    fprintf(stderr, "Failed to register mp4v video codec\n");
    return NULL;
  }

  return &codec->info;
}
