/* codec_xvid.c
 * Copyright (C) 2002 QT4Linux and OpenQuicktime Teams
 *
 * This file is part of OpenQuicktime, a free QuickTime library.
 *
 * Based on QT4Linux by Adam Williams.
 *
 * OpenQuicktime is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * OpenQuicktime is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: codec_xvid.c,v 1.1.1.1 2003/11/28 06:03:46 ahu Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <xvid.h>

#include "openquicktime/openquicktime.h"
#include "openquicktime/colormodels.h"
#include "openquicktime/structs.h"
#include "openquicktime/plugin.h"


static int              Initialised   = 0;      

static const char   Signature[5]  = "XVID"; 
static const char   Alt4Signature[5]  = "DIV4"; 
static const char   Alt5Signature[5]  = "DX50"; 
static const char   AltXSignature[5]  = "DIVX"; 
static const char   AltMP4Signature[5]  = "mp4v"; 
static const char   *Name = "XviD MPEG-4 video codec"; 
static const char   *Alt4Name = "DivX 4 compatible decompressor"; 
static const char   *Alt5Name = "DivX 5 compatible decompressor"; 
static const char   *AltXName = "OpenDIVX compatible decompressor"; 
static const char   *AltMP4Name = "MPEG-4 Video compatible decompressor"; 
static const char   *Version  = "0.1"; 
static const char   *Copyright  = "2002 OpenQuicktime Team"; 
static const char   *ModuleAuthor  = "Jozef Hatala"; 
static const char   *CodecAuthor  = "Michael Militzer, Peter Ross, Christoph Lampert"; 

static const oqt_codec_param_t params[] =
{
  { "bitrate", OQT_PARAMETER_INTEGER, "The bitrate of the target encoded stream"}
};

#define PARAM_COUNT (sizeof(params)/sizeof(params[0]))

typedef struct {
  // decoder
  int decoder_initialized;
  int use_esds;
  void *dechandle;
  unsigned char *dec_buffer;

  // encoder
  int encoder_initialized;
  void *enchandle;
  int bitrate;
  unsigned char *enc_buffer;
} Private;  

static const oqt_codec_param_t* list_params_XVID(int *o_numparams)
{
	*o_numparams = PARAM_COUNT;
	
	return params;
}

static int set_param_XVID(oqt_t* file,
			  int          track,
			  const char*  key,
			  const void*  value)
{
  Private *p = (Private*)(file->vtracks[track].codec_private);

  if(!strcasecmp(key, "bitrate"))
    p->bitrate = *(int*)value;
  else {
    fprintf(stderr,"set_param_XVID: unknown parameter named '%s'\n",key);
    return OQT_PARAMETER_UNKNOWN;
  }
  return 0;
}


static int get_param_XVID(oqt_t* file,
			  int          track,
			  const char*  key,
			  void*        value)
{
  Private *p = (Private*)(file->vtracks[track].codec_private);

  if(!strcasecmp(key, "bitrate")) {
    *(int*)value = p->bitrate;
    return OQT_PARAMETER_INTEGER;
  } else {
    fprintf(stderr,"get_param_XVID: unknown parameter named '%s'\n",key);
    return OQT_PARAMETER_UNKNOWN;
  }
}

static int reads_colormodel_XVID(oqt_t *file, 
		 int track,
		 int colormodel)
{
  return ( colormodel == BC_YUV420P );
}


static int decode_XVID(oqt_t *file,
		       int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t  *input,
		       oqt_uint8_t  **output,
		       int color_model)
{
  int xerr;
  Private *p = (Private*)(file->vtracks[track].codec_private);
  int width = file->vtracks[track].track->tkhd.track_width;
  int height = file->vtracks[track].track->tkhd.track_height;
  XVID_DEC_FRAME xframe;
  oqt_uint8_t *vol_buffer = NULL;

  xframe.bitstream = input;
  xframe.length = inputsize;
  xframe.stride = width;

  if(!p->decoder_initialized) {
    XVID_INIT_PARAM xinit;
    XVID_DEC_PARAM xparam;

    xinit.cpu_flags = 0;
    xvid_init(NULL, 0, &xinit, NULL);
    xparam.width = width;
    xparam.height = height;

    xerr = xvid_decore(NULL, XVID_DEC_CREATE, &xparam, NULL);
    if(xerr) {
      fprintf(stderr, "decode_XVID: XVID_DEC_CREATE failed: %d\n", xerr);
      return 1;
    }

    p->dec_buffer = malloc(width*height + width*height/2);

    p->dechandle = xparam.handle;
    p->decoder_initialized = 1;

    if(p->use_esds) {
      oqt_int32_t entries = \
        file->vtracks[track].track->mdia.minf.stbl.stsd.total_entries;
      oqt_stsd_table_t *table = \
        file->vtracks[track].track->mdia.minf.stbl.stsd.table;
      
      if(entries < 1 || !table) {
        fprintf(stderr, "decode_XVID: missing esds atom for mp4v\n");

      } else {
        oqt_esds_t *esds = &(table->esds);

        p->use_esds = 0;
        vol_buffer = malloc(esds->decoderConfigLen + inputsize);
        memcpy(vol_buffer, esds->decoderConfig, esds->decoderConfigLen);
        memcpy(vol_buffer + esds->decoderConfigLen, input, inputsize);
        xframe.bitstream = vol_buffer;
        xframe.length = esds->decoderConfigLen + inputsize;
      }
    }
  }

  if(color_model == BC_NONE) {
    xframe.colorspace = XVID_CSP_NULL;
    xframe.image = NULL;
  } else if(color_model == BC_YUV420P) {
    xframe.colorspace = XVID_CSP_YV12;

    if(output[2] == output[0] + width*height &&
       output[1] == output[2] + width*height/4) {
      xframe.image = output[0];
    } else {
      xframe.image = p->dec_buffer;
    }
  } else {
    fprintf(stderr,"decode_XVID: unsupported color model\n");
    return 1;
  }

  xerr = xvid_decore(p->dechandle, XVID_DEC_DECODE, &xframe, NULL);
  if(xerr) {
    fprintf(stderr, "decode_XVID: XVID_DEC_DECODE failed: %d\n", xerr);
    return 1;
  }

  if(vol_buffer) {
    free(vol_buffer);
  }

  if(xframe.image == p->dec_buffer) {
    memcpy(output[0], p->dec_buffer, width*height);
    memcpy(output[2], p->dec_buffer + width*height,
	   width*height/4);
    memcpy(output[1], p->dec_buffer + width*height + width*height/4,
	   width*height/4);
  }

  return 0;
}


/* OpenDIVX has reversed U and V planes */
static int decode_OpenDIVX(oqt_t *file,
		       int track,
		       oqt_uint32_t inputsize,
		       oqt_uint8_t  *input,
		       oqt_uint8_t  **output,
		       int color_model)
{
  if(color_model == BC_NONE) {
    return decode_XVID(file, track, inputsize, input, output, color_model);
  } else if(color_model == BC_YUV420P) {
    unsigned char *tmp_output[3] = { output[0], output[2], output[1] };
    return decode_XVID(file, track, inputsize, input, tmp_output, color_model);
  } else {
    fprintf(stderr,"decode_OpenDIVX: unsupported color model\n");
    return 1;
  }
}


static int writes_colormodel_XVID(oqt_t *file, 
		 int track,
		 int colormodel)
{ 
  return ( colormodel == BC_YUV420P );
}

static int preferred_colormodel_XVID(oqt_t *file, 
		 int track)
{ 
  return BC_YUV420P;
}

static int encode_XVID(oqt_t *file, 
		       int track,
		       oqt_uint8_t **input,
		       oqt_uint32_t out_size,
		       oqt_uint8_t *output,
		       int *IsAKeyFrame,
		       int color_model)
{
  int xerr;
  Private *p = (Private*)(file->vtracks[track].codec_private);
  int width = file->vtracks[track].track->tkhd.track_width;
  int height = file->vtracks[track].track->tkhd.track_height;
  XVID_ENC_FRAME xframe;

  if(color_model != BC_YUV420P) {
    fprintf(stderr,"encode_XVID: unsupported color model\n");
    return 0;
  }

  if(!p->encoder_initialized) {
    XVID_INIT_PARAM xinit;
    XVID_ENC_PARAM xparam;
    float framerate = 
      (float)file->vtracks[track].track->mdia.mdhd.time_scale /
      file->vtracks[track].track->mdia.minf.stbl.stts.table[0].sample_duration;
    
    xinit.cpu_flags = 0;
    xvid_init(NULL, 0, &xinit, NULL);

    xparam.width = width;
    xparam.height = height;
    if(framerate == (int)framerate) {
      xparam.fincr = 1;
      xparam.fbase = framerate;
    } else {
      xparam.fincr = 1001;
      xparam.fbase = (int)(1001 * framerate);
    }
    xparam.rc_bitrate = p->bitrate;
    // TODO: turn these into parameters
    xparam.rc_reaction_delay_factor = 16;
    xparam.rc_averaging_period = 100;
    xparam.rc_buffer = 100;
    xparam.min_quantizer = 2;
    xparam.max_quantizer = 31;
    xparam.max_key_interval = (int)(10*framerate);

    xerr = xvid_encore(NULL, XVID_ENC_CREATE, &xparam, NULL);
    if(xerr) {
      fprintf(stderr, "encode_XVID: XVID_ENC_CREATE failed: %d\n", xerr);
      return 1;
    }

    p->enc_buffer = malloc(width*height + width*height/2);

    p->enchandle = xparam.handle;
    p->encoder_initialized = 1;
  }

  // TODO: parameters
  xframe.general = XVID_H263QUANT;
  xframe.general |= XVID_HALFPEL;
  xframe.general |= XVID_INTER4V;
  xframe.general |= PMV_HALFPELREFINE16 | PMV_EARLYSTOP16 |
    PMV_HALFPELDIAMOND8 | PMV_EARLYSTOP8;

  xframe.bitstream = output;
  xframe.length = -1;
  xframe.colorspace = XVID_CSP_YV12;
  xframe.intra = -1;
  xframe.quant = 0;

  if(input[2] == input[0] + width*height &&
     input[1] == input[2] + width*height/4) {
    xframe.image = input[0];

  } else {
    memcpy(p->enc_buffer, input[0], width*height);
    memcpy(p->enc_buffer + width*height,
	   input[2], width*height/4);
    memcpy(p->enc_buffer + width*height + width*height/4,
	   input[1], width*height/4);

    xframe.image = p->enc_buffer;
  }

  xerr = xvid_encore(p->enchandle, XVID_ENC_ENCODE, &xframe, NULL);
  if(xerr) {
    fprintf(stderr, "encode_XVID: XVID_ENC_ENCODE failed: %d\n", xerr);
    return 0;
  }

  *IsAKeyFrame = xframe.intra;
  return xframe.length;
}

static int delete_XVID(oqt_t *file, int track)
{
  oqt_video_map_t *vtrack = &file->vtracks[track];
  Private *p = (Private*)(vtrack->codec_private);

  if (p) {
    if(p->encoder_initialized) {
      xvid_encore(p->enchandle, XVID_ENC_DESTROY, NULL, NULL);
    }
    if(p->decoder_initialized) {
      xvid_decore(p->dechandle, XVID_DEC_DESTROY, NULL, NULL);
    }

    if(p->dec_buffer)
      free(p->dec_buffer);
    if(p->enc_buffer)
      free(p->enc_buffer);

    free(p);
    vtrack->codec_private = NULL;
   }

  return --Initialised;
}

static int init_XVID(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Private *p = malloc(sizeof(Private));
	vtrack->codec_private = p;
	memset(p, 0, sizeof(Private));
	
	// defaults
	p->bitrate = 900000;
	
	return ++Initialised;
}

static int init_mp4v(oqt_t *file, int track)
{
	oqt_video_map_t *vtrack = &file->vtracks[track];
	Private *p = malloc(sizeof(Private));
	vtrack->codec_private = p;
	memset(p, 0, sizeof(Private));
	
	// defaults
	p->bitrate = 900000;
	p->use_esds = 1;
	
	return ++Initialised;
}

static void fill_codec_struct(oqt_video_codec_t *codec)
{
  codec->info.version = Version;
  codec->info.copyright = Copyright;
  codec->info.module_author = ModuleAuthor;
  codec->info.codec_author = CodecAuthor;
  codec->info.interface_age = OQT_INTERFACE_AGE;	

  codec->init_codec = init_XVID;
  codec->decode = decode_XVID;
  codec->delete_codec = delete_XVID;
  codec->reads_colormodel = reads_colormodel_XVID;
  codec->preferred_colormodel = preferred_colormodel_XVID;
}

oqt_codec_info_t*
oqt_plugin_register( void )
{
  oqt_video_codec_t *codec;
  oqt_video_codec_t *alt4;
  oqt_video_codec_t *alt5;
  oqt_video_codec_t *altX;
  oqt_video_codec_t *altMP4;

  codec = oqt_allocate_video_codec();
  if (!codec) return NULL;
  fill_codec_struct(codec);

  codec->encode = encode_XVID;
  codec->writes_colormodel = writes_colormodel_XVID;
  codec->list_params = list_params_XVID;
  codec->set_param = set_param_XVID;
  codec->get_param = get_param_XVID;

  codec->info.name = Name;
  memcpy( &codec->info.fourcc, Signature, 4); 
  codec->info.has_children = 1;

  alt4 = oqt_allocate_video_codec();
  if (!alt4) return NULL;
  fill_codec_struct(alt4);

  alt4->info.name = Alt4Name;
  memcpy( &alt4->info.fourcc, Alt4Signature, 4); 
  alt4->info.parent = &codec->info;
  alt4->info.merit = 80;

  alt5 = oqt_allocate_video_codec();
  if (!alt5) return NULL;
  fill_codec_struct(alt5);

  alt5->info.name = Alt5Name;
  memcpy( &alt5->info.fourcc, Alt5Signature, 4); 
  alt5->info.parent = &codec->info;
  alt5->info.merit = 80;

  altX = oqt_allocate_video_codec();
  if (!altX) return NULL;
  fill_codec_struct(altX);
  altX->decode = decode_OpenDIVX;

  altX->info.name = AltXName;
  memcpy( &altX->info.fourcc, AltXSignature, 4); 
  altX->info.parent = &codec->info;
  altX->info.merit = 80;

  altMP4 = oqt_allocate_video_codec();
  if (!altMP4) return NULL;
  fill_codec_struct(altMP4);
  altMP4->init_codec = init_mp4v;

  altMP4->info.name = AltMP4Name;
  memcpy( &altMP4->info.fourcc, AltMP4Signature, 4); 
  altMP4->info.parent = &codec->info;
  altMP4->info.merit = 85;

  if (oqt_register_codec(&alt4->info)) {
    fprintf(stderr, "Failed to register DIV4 video codec\n");
    return NULL;
  }

  if (oqt_register_codec(&alt5->info)) {
    fprintf(stderr, "Failed to register DX50 video codec\n");
    return NULL;
  }

  if (oqt_register_codec(&altX->info)) {
    fprintf(stderr, "Failed to register DIVX video codec\n");
    return NULL;
  }

  if (oqt_register_codec(&altMP4->info)) {
    fprintf(stderr, "Failed to register mp4v video codec\n");
    return NULL;
  }

  return &codec->info;
}
