/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2004. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This file is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This file is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this file; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/



#include <new>
#include "debug.h"
#include "ImageArray.h"



namespace {
    
    inline int Max(int a, int b)
    {
        return a < b ? b : a;
    }
}


ImageArray::ImageArray(SDL_Surface *screen, const SDL_Rect& rect)
    : Grain(screen,rect)
{
	m_UpdateArea = drawable_area;
}



ImageArray::~ImageArray()
{
    
    CellArray::iterator iter;
    for(iter = m_cells.begin(); iter != m_cells.end(); iter++) {
        if (*iter) {
            delete (*iter);
        }
    }
}



int
ImageArray::Init(int size)
{
    assert( m_cells.size() == 0 );

    
    if (size > 0) {
        m_cells.reserve(size);
        for (int i = 0; i < size; i++) {
            m_cells.push_back(0);
        }
    }

    return 0;
}



int
ImageArray::Draw()
{
    int result;
    SDL_Rect dstArea = {0, 0, 0, 0};
    dstArea.x = m_UpdateArea.x;

    
    if ( m_OrigImage == 0 ) {
        int result;
        result = StoreOrigImage();
        if (result != 0) {
            return result;
        }
	}

    
    RecomputeUpdateArea();

    
    CellArray::const_iterator iter;
    for(iter = m_cells.begin(); iter != m_cells.end(); iter++) {

        if ((*iter) && (*iter)->image) {
            
            dstArea.y = m_UpdateArea.y + 
                (m_UpdateArea.h - (*iter)->image->h) / 2;

            
            result = SDL_BlitSurface((*iter)->image, NULL, m_screen, &dstArea);
            if ( result != 0 ) {
                return result;
            }

            
            dstArea.x += (*iter)->width;
        }
    }

    return 0;
}



int
ImageArray::GetWidth() const
{
    int totalWidth = 0;

    CellArray::const_iterator  iter;
    for(iter = m_cells.begin(); iter != m_cells.end(); iter++) {
        if (*iter) {
            totalWidth += (*iter)->width;
        }
    }

    return totalWidth;
}



int
ImageArray::GetHeight() const
{
    int totalHeight = 0;

    CellArray::const_iterator  iter;
    for(iter = m_cells.begin(); iter != m_cells.end(); iter++) {
        if ( (*iter) && (*iter)->image ) {
            totalHeight = Max(totalHeight, (*iter)->image->h);
        }
    }

    return totalHeight;
}



void
ImageArray::SetHeight(int height)
{
    assert( height >= GetHeight() );

    m_UpdateArea.h = height;
}




void
ImageArray::SetPosition(int x, int y)
{
    assert( x >= 0 );
    assert( y >= 0 );

    
    if ( x == m_UpdateArea.x && y == m_UpdateArea.y ) {
        return;
    }

    
    RecomputeUpdateArea();

    
    RestoreOrigImage();

    m_UpdateArea.x = x;
    m_UpdateArea.y = y;
    drawable_area = m_UpdateArea;

    
    StoreOrigImage();
}



int
ImageArray::AddCell(SDL_Surface* image, int width)
{
    assert( image != 0 );
    assert( width >= image->w );

    Cell* newCell = CreateNewCell(image,width);
    if ( newCell == 0 ) {
        return -1;
    }

    
    m_cells.push_back(newCell);

    return 0;
}



int
ImageArray::RemoveCell(unsigned int index)
{
    if ( index >= m_cells.size() ) {
        return -1;
    }

    
    CellArray::iterator iter = m_cells.begin();
    for ( unsigned int i = 0; i < index; i++ ) {
        iter++;
    }

    
    Cell* tmp = (*iter);
    m_cells.erase(iter);

    
    delete tmp;

    return 0;
}



SDL_Surface*
ImageArray::SetCell(SDL_Surface* image, int width, unsigned int index)
{
    assert( image != 0 );
    assert( width >= image->w );
    assert( index < m_cells.size() );

    Cell* tmpCell = m_cells[index];
    if ( tmpCell == 0 ) {
        
        tmpCell = CreateNewCell(image,width);
        assert( tmpCell != 0 );
        
        
        m_cells[index] = tmpCell;
        return 0;
    }

    
    SDL_Surface* oldImage = 0;
    if ( tmpCell && tmpCell->image ) {
        oldImage = tmpCell->image;
    }

    tmpCell->image = image;
    tmpCell->width = width;

    return oldImage;
}



int
ImageArray::GetCellCount() const
{
    return m_cells.size();
}



int
ImageArray::GetCellWidth(unsigned int index) const
{
    assert( index < m_cells.size() );

    int width = 0;
    if ( m_cells[index] ) {
        width = m_cells[index]->width;
    }

    return width;
}



ImageArray::Cell* 
ImageArray::CreateNewCell(SDL_Surface* image, int width)
{
    assert( image != 0 );
    assert( width >= image->w );

    
    Cell* newCell = new(std::nothrow) Cell;
    if ( newCell == 0 ) {
        return 0;
    }

    
    newCell->image = image;

    
    newCell->width = width;

    return newCell;
}



void
ImageArray::RecomputeUpdateArea()
{
    int newWidth = GetWidth();
    int newHeight = GetHeight();

    
    if ( newWidth != m_UpdateArea.w ||
         newHeight > m_UpdateArea.h ) {

        int result;

        
        result = RestoreOrigImage();
        assert( result == 0 );

        
        m_UpdateArea.w = newWidth;
        if ( newHeight > m_UpdateArea.h ) {
            m_UpdateArea.h = newHeight;
        }
        drawable_area = m_UpdateArea;

        
        result = StoreOrigImage();
        assert( result == 0 );
    }
}



int
ImageArray::StoreOrigImage()
{
    assert( m_screen != 0 );

    if ( m_OrigImage ) {
        SDL_FreeSurface(m_OrigImage);
        m_OrigImage = 0;
    }

    

    SDL_PixelFormat* format = m_screen->format;
    SDL_Surface* surface;

    surface = SDL_CreateRGBSurface(SDL_SWSURFACE, 
                                   drawable_area.w, drawable_area.h,
                                   format->BitsPerPixel,
                                   format->Rmask, format->Gmask,
                                   format->Bmask, format->Amask);
    if ( surface == 0 ) {
        DebugPrint("ImageArray::StoreOrigImage: Create surface failed\n");
        return -1;
    }

    int result = SDL_BlitSurface(m_screen, &drawable_area, surface, NULL);
    if ( result != 0 ) {
        DebugPrint("ImageArray::StoreOrigImage: Blit failed\n");
        SDL_FreeSurface(surface);
        return result;
    }

    m_OrigImage = surface;

    return 0;
}
