/*
 *  Copyright (C) 2000, 2001 Const Kaplinsky.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * tight.c - handle ``tight'' encoding.
 *
 * This file shouldn't be compiled directly. It is included multiple
 * times by rfbproto.c, each time with a different definition of the
 * macro BPP. For each value of BPP, this file defines a function
 * which handles a tight-encoded rectangle with BPP bits per pixel.
 *
 */

#define TIGHT_MIN_TO_COMPRESS 12
#define LZO_DEFAULT_BUFFER_SIZE	(100 * 1024)

#define UintBPP CONCAT2E(Uint,BPP)
#define filterPtrBPP CONCAT2E(filterPtr,BPP)

#define HandleTightBPP CONCAT2E(HandleTight,BPP)
#define InitFilterCopyBPP CONCAT2E(InitFilterCopy,BPP)
#define InitFilterPaletteBPP CONCAT2E(InitFilterPalette,BPP)
#define InitFilterGradientBPP CONCAT2E(InitFilterGradient,BPP)
#define FilterCopyBPP CONCAT2E(FilterCopy,BPP)
#define FilterPaletteBPP CONCAT2E(FilterPalette,BPP)
#define FilterGradientBPP CONCAT2E(FilterGradient,BPP)

#if BPP != 8
#define DecompressJpegRectBPP CONCAT2E(DecompressJpegRect,BPP)
#endif

#define DecompressLZORectBPP CONCAT2E(DecompressLZORect,BPP)

#if BPP == 16
#define DecompressRKRectBPP CONCAT2E(DecompressRKRect,BPP)
#endif


#ifndef RGB_TO_PIXEL
#define RGB_TO_PIXEL(bpp,r,g,b)		\
  ((Uint##bpp)SDL_MapRGB(dpy->format, r, g, b))

#define RGB24_TO_PIXEL(bpp,r,g,b)   \
  ((Uint##bpp)SDL_MapRGB(dpy->format, r, g, b))

#define RGB24_TO_PIXEL32(r,g,b)						\
  (((Uint32)(r) & 0xFF) << myFormat.redShift |				\
   ((Uint32)(g) & 0xFF) << myFormat.greenShift |			\
   ((Uint32)(b) & 0xFF) << myFormat.blueShift)

#endif


#ifndef DebugAlgo

#if !defined(NDEBUG) && defined(DEBUG)
#define PRINT_ALGO
#endif

#ifdef PRINT_ALGO
#define DebugAlgo(fmt , args...)\
    fprintf(stderr , "Tight algorithm: " fmt , ## args)
#else
#define DebugAlgo(fmt , args...)
#endif

#endif /* DebugAlgo */


/* Type declarations */

typedef void (*filterPtrBPP)(int, UintBPP *);


/* Prototypes */

static int InitFilterCopyBPP (int rw, int rh);
static int InitFilterPaletteBPP (int rw, int rh);
static int InitFilterGradientBPP (int rw, int rh);
static void FilterCopyBPP (int numRows, UintBPP *destBuffer);
static void FilterPaletteBPP (int numRows, UintBPP *destBuffer);
static void FilterGradientBPP (int numRows, UintBPP *destBuffer);
#if BPP != 8
static SDL_bool DecompressJpegRectBPP(int x, int y, int w, int h);
#endif
static SDL_bool DecompressLZORectBPP(int rx, int ry, int rw, int rh);
#if BPP == 16
static SDL_bool DecompressRKRectBPP(int rx, int ry, int rw, int rh);
#endif

/* Definitions */

static SDL_bool
HandleTightBPP (int rx, int ry, int rw, int rh)
{
    SDL_Rect rect;
    Uint32 colour;
    UintBPP fill_colour;
    Uint8 comp_ctl;
    Uint8 filter_id;
    filterPtrBPP filterFn;
    z_streamp zs;
    char *buffer2;
    int err, stream_id, compressedLen, bitsPixel;
    int bufferSize, rowSize, numRows, portionLen, rowsProcessed, extraBytes;

    if (!ReadFromRFBServer((char *)&comp_ctl, 1))
        return SDL_FALSE;

    /* Flush zlib streams if we are told by the server to do so. */
    for (stream_id = 0; stream_id < 4; stream_id++) {
        if ((comp_ctl & 1) && zlibStreamActive[stream_id]) {
            if (inflateEnd (&zlibStream[stream_id]) != Z_OK &&
                zlibStream[stream_id].msg != NULL)
                fprintf(stderr, "inflateEnd: %s\n", zlibStream[stream_id].msg);
            zlibStreamActive[stream_id] = SDL_FALSE;
        }
        comp_ctl >>= 1;
    }


	switch ( comp_ctl ) {
    /* Handle solid rectangles. */
    case rfbTightFill:
		DebugAlgo("Fill compression\n");
#if BPP == 32
        if (myFormat.depth == 24 && myFormat.redMax == 0xFF &&
            myFormat.greenMax == 0xFF && myFormat.blueMax == 0xFF) {
            if (!ReadFromRFBServer(buffer, 3))
                return SDL_FALSE;
            fill_colour = RGB24_TO_PIXEL32(buffer[0], buffer[1], buffer[2]);
        } else {
            if (!ReadFromRFBServer((char*)&fill_colour, sizeof(fill_colour)))
                return SDL_FALSE;
        }
#else
        if (!ReadFromRFBServer((char*)&fill_colour, sizeof(fill_colour)))
            return SDL_FALSE;
#endif

#if (BPP == 8)
        colour = (appData.useBGR233) ?
            BGR233ToPixel[fill_colour] : fill_colour;
#else
        colour = fill_colour;
#endif

        rect.x = rx;
        rect.y = ry;
        rect.w = rw;
        rect.h = rh;
		SDL_FillRect(ovl, &rect, colour);
        return SDL_TRUE;
		break;

#if ( BPP != 8 )
	case rfbTightJpeg:
		DebugAlgo("JPEG compression\n");
		return DecompressJpegRectBPP(rx, ry, rw, rh);
		break;

	case rfbTightStretchJpeg:
		{
			Uint16 sw, sh;

			DebugAlgo("StretchJPEG compression\n");

			/* Read the requested scaled width */
			if (!ReadFromRFBServer((char *)&sw, sizeof(sw))) {
				return SDL_FALSE;
			}

			/* Read the requested scaled height */
			if (!ReadFromRFBServer((char *)&sh, sizeof(sh))) {
				return SDL_FALSE;
			}
        
			appData.useZoomRect = SDL_TRUE;
			appData.zoomRect.x = rx;
			appData.zoomRect.y = ry;
			appData.zoomRect.w = rw;
			appData.zoomRect.h = rh;

			return DecompressJpegRectBPP(rx, ry, rw, rh);
		}
		break;
#endif

#if BPP == 16
	case rfbTightRK:
		DebugAlgo("RK compression\n");
		return DecompressRKRectBPP(rx,ry,rw,rh);
		break;
#endif

	case rfbTightLZO:
		DebugAlgo("LZO compression\n");
		return DecompressLZORectBPP(rx,ry,rw,rh);
		break;

	default:
		/* Must be other encoding so fall through */		
		DebugAlgo("Fall through\n");
		break;
	}

    /*
     * Here primary compression mode handling begins.
     * Data was processed with optional filter + zlib compression.
     */

    /* First, we should identify a filter to use. */
    if ((comp_ctl & rfbTightExplicitFilter) != 0) {
        if (!ReadFromRFBServer((char*)&filter_id, 1))
            return SDL_FALSE;

        switch (filter_id) {
        case rfbTightFilterCopy:
            filterFn = FilterCopyBPP;
            bitsPixel = InitFilterCopyBPP(rw, rh);
            break;
        case rfbTightFilterPalette:
            filterFn = FilterPaletteBPP;
            bitsPixel = InitFilterPaletteBPP(rw, rh);
            break;
        case rfbTightFilterGradient:
            filterFn = FilterGradientBPP;
            bitsPixel = InitFilterGradientBPP(rw, rh);
            break;
        default:
            fprintf(stderr, "Tight encoding: unknown filter code received.\n");
            return SDL_TRUE;
        }
    } else {
        filterFn = FilterCopyBPP;
        bitsPixel = InitFilterCopyBPP(rw, rh);
    }
    if (bitsPixel == 0) {
        fprintf(stderr, "Tight encoding: error receiving palette.\n");
        return SDL_TRUE;
    }

    /* Determine if the data should be decompressed or just copied. */
    rowSize = (rw * bitsPixel + 7) / 8;
    if (rh * rowSize < TIGHT_MIN_TO_COMPRESS) {
        if (!ReadFromRFBServer((char*)buffer, rh * rowSize))
            return SDL_FALSE;

        buffer2 = &buffer[TIGHT_MIN_TO_COMPRESS * 4];
        filterFn(rh, (UintBPP *)buffer2);
        CopyDataToScreen(buffer2, rx, ry, rw, rh);

        return SDL_TRUE;
    }

    /* Read the length (1..3 bytes) of compressed data following. */
    compressedLen = (int)ReadCompactLen();
    if (compressedLen <= 0) {
        fprintf(stderr, "Incorrect data received from the server.\n");
        return SDL_TRUE;
    }

    /* Now let's initialize compression stream if needed. */
    stream_id = comp_ctl & 0x03;
    zs = &zlibStream[stream_id];
    if (!zlibStreamActive[stream_id]) {
        zs->zalloc = Z_NULL;
        zs->zfree = Z_NULL;
        zs->opaque = Z_NULL;
        err = inflateInit(zs);
        if (err != Z_OK) {
            if (zs->msg != NULL)
                fprintf(stderr, "InflateInit error: %s.\n", zs->msg);
            return SDL_TRUE;
        }
        zlibStreamActive[stream_id] = SDL_TRUE;
    }

    /* Read, decode and draw actual pixel data in a loop. */

    bufferSize = BUFFER_SIZE * bitsPixel / (bitsPixel + BPP) & 0xFFFFFFFC;
    buffer2 = &buffer[bufferSize];
    if (rowSize > bufferSize) {
        /* Should be impossible when BUFFER_SIZE >= 16384 */
        fprintf(stderr, "Internal error: incorrect buffer size.\n");
        return SDL_TRUE;
    }

    rowsProcessed = 0;
    extraBytes = 0;

    while (compressedLen > 0) {
        if (compressedLen > ZLIB_BUFFER_SIZE)
            portionLen = ZLIB_BUFFER_SIZE;
        else
            portionLen = compressedLen;

        if (!ReadFromRFBServer((char*)zlib_buffer, portionLen))
            return SDL_FALSE;

        compressedLen -= portionLen;

        zs->next_in = (Bytef *)zlib_buffer;
        zs->avail_in = portionLen;

        do {
            zs->next_out = (Bytef *)&buffer[extraBytes];
            zs->avail_out = bufferSize - extraBytes;

            err = inflate(zs, Z_SYNC_FLUSH);
            if (err == Z_BUF_ERROR)   /* Input exhausted -- no problem. */
                break;
            if (err != Z_OK && err != Z_STREAM_END) {
                if (zs->msg != NULL) {
                    fprintf(stderr, "Inflate error: %s.\n", zs->msg);
                } else {
                    fprintf(stderr, "Inflate error: %d.\n", err);
                }
                return SDL_TRUE;
            }

            numRows = (bufferSize - zs->avail_out) / rowSize;

            filterFn(numRows, (UintBPP *)buffer2);

            extraBytes = bufferSize - zs->avail_out - numRows * rowSize;
            if (extraBytes > 0)
                memcpy(buffer, &buffer[numRows * rowSize], extraBytes);

            CopyDataToScreen(buffer2, rx, ry + rowsProcessed, rw, numRows);
            rowsProcessed += numRows;
        }
        while (zs->avail_out == 0);
    }

    if (rowsProcessed != rh) {
        fprintf(stderr, "Incorrect number of scan lines after decompression.\n");
        return SDL_TRUE;
    }

    return SDL_TRUE;
}

/*----------------------------------------------------------------------------
 *
 * Filter stuff.
 *
 */

/*
  The following variables are defined in rfbproto.c:
  static SDL_bool cutZeros;
  static int rectWidth, rectColors;
  static Uint8 tightPalette[256*4];
  static Uint8 tightPrevRow[2048*3*sizeof(Uint16)];
*/

static int
InitFilterCopyBPP (int rw, int rh)
{
    rectWidth = rw;

#if BPP == 32
    if (myFormat.depth == 24 && myFormat.redMax == 0xFF &&
        myFormat.greenMax == 0xFF && myFormat.blueMax == 0xFF) {
        cutZeros = SDL_TRUE;
        return 24;
    } else {
        cutZeros = SDL_FALSE;
    }
#endif

    return BPP;
}

static void
FilterCopyBPP (int numRows, UintBPP *dst)
{

#if BPP == 32
    int x, y;

    if (cutZeros) {
        for (y = 0; y < numRows; y++) {
            for (x = 0; x < rectWidth; x++) {
                dst[y*rectWidth+x] =
                    RGB24_TO_PIXEL32(buffer[(y*rectWidth+x)*3],
                                     buffer[(y*rectWidth+x)*3+1],
                                     buffer[(y*rectWidth+x)*3+2]);
            }
        }
        return;
    }
#endif

    memcpy (dst, buffer, numRows * rectWidth * (BPP / 8));
}

static int
InitFilterGradientBPP (int rw, int rh)
{
    int bits;

    bits = InitFilterCopyBPP(rw, rh);
    if (cutZeros)
        memset(tightPrevRow, 0, rw * 3);
    else
        memset(tightPrevRow, 0, rw * 3 * sizeof(Uint16));

    return bits;
}

#if BPP == 32

static void
FilterGradient24 (int numRows, Uint32 *dst)
{
    int x, y, c;
    Uint8 thisRow[2048*3];
    Uint8 pix[3];
    int est[3];

    for (y = 0; y < numRows; y++) {

        /* First pixel in a row */
        for (c = 0; c < 3; c++) {
            pix[c] = tightPrevRow[c] + buffer[y*rectWidth*3+c];
            thisRow[c] = pix[c];
        }
        dst[y*rectWidth] = RGB24_TO_PIXEL32(pix[0], pix[1], pix[2]);

        /* Remaining pixels of a row */
        for (x = 1; x < rectWidth; x++) {
            for (c = 0; c < 3; c++) {
                est[c] = (int)tightPrevRow[x*3+c] + (int)pix[c] -
                    (int)tightPrevRow[(x-1)*3+c];
                if (est[c] > 0xFF) {
                    est[c] = 0xFF;
                } else if (est[c] < 0x00) {
                    est[c] = 0x00;
                }
                pix[c] = (Uint8)est[c] + buffer[(y*rectWidth+x)*3+c];
                thisRow[x*3+c] = pix[c];
            }
            dst[y*rectWidth+x] = RGB24_TO_PIXEL32(pix[0], pix[1], pix[2]);
        }

        memcpy(tightPrevRow, thisRow, rectWidth * 3);
    }
}

#endif

static void
FilterGradientBPP (int numRows, UintBPP *dst)
{
    int x, y, c;
    UintBPP *src = (UintBPP *)buffer;
    Uint16 *thatRow = (Uint16 *)tightPrevRow;
    Uint16 thisRow[2048*3];
    Uint16 pix[3];
    Uint16 max[3];
    int shift[3];
    int est[3];

#if BPP == 32
    if (cutZeros) {
        FilterGradient24(numRows, dst);
        return;
    }
#endif

    max[0] = myFormat.redMax;
    max[1] = myFormat.greenMax;
    max[2] = myFormat.blueMax;

    shift[0] = myFormat.redShift;
    shift[1] = myFormat.greenShift;
    shift[2] = myFormat.blueShift;

    for (y = 0; y < numRows; y++) {

        /* First pixel in a row */
        for (c = 0; c < 3; c++) {
            pix[c] = (Uint16)((src[y*rectWidth] >> shift[c]) + (thatRow[c] & max[c]));
            thisRow[c] = pix[c];
        }
        dst[y*rectWidth] = RGB_TO_PIXEL(BPP, pix[0], pix[1], pix[2]);

        /* Remaining pixels of a row */
        for (x = 1; x < rectWidth; x++) {
            for (c = 0; c < 3; c++) {
                est[c] = (int)thatRow[x*3+c] + (int)pix[c] - (int)thatRow[(x-1)*3+c];
                if (est[c] > (int)max[c]) {
                    est[c] = (int)max[c];
                } else if (est[c] < 0) {
                    est[c] = 0;
                }
                pix[c] = (Uint16)((src[y*rectWidth+x] >> shift[c]) + (est[c] & max[c]));
                thisRow[x*3+c] = pix[c];
            }
            dst[y*rectWidth+x] = RGB_TO_PIXEL(BPP, pix[0], pix[1], pix[2]);
        }
        memcpy(thatRow, thisRow, rectWidth * 3 * sizeof(Uint16));
    }
}

static int
InitFilterPaletteBPP (int rw, int rh)
{
    Uint8 numColors;
#if BPP == 32
    int i;
    UintBPP *palette = (UintBPP *)tightPalette;
#endif

    rectWidth = rw;

    if (!ReadFromRFBServer((char*)&numColors, 1))
        return 0;

    rectColors = (int)numColors;
    if (++rectColors < 2)
        return 0;

#if BPP == 32
    if (myFormat.depth == 24 && myFormat.redMax == 0xFF &&
        myFormat.greenMax == 0xFF && myFormat.blueMax == 0xFF) {
        if (!ReadFromRFBServer((char*)&tightPalette, rectColors * 3))
            return 0;
        for (i = rectColors - 1; i >= 0; i--) {
            palette[i] = RGB24_TO_PIXEL32(tightPalette[i*3],
                                          tightPalette[i*3+1],
                                          tightPalette[i*3+2]);
        }
        return (rectColors == 2) ? 1 : 8;
    }
#endif

    if (!ReadFromRFBServer((char*)&tightPalette, rectColors * (BPP / 8)))
        return 0;

    return (rectColors == 2) ? 1 : 8;
}

static void
FilterPaletteBPP (int numRows, UintBPP *dst)
{
    int x, y, b, w;
    Uint8 *src = (Uint8 *)buffer;
    UintBPP *palette = (UintBPP *)tightPalette;

    if (rectColors == 2) {
        w = (rectWidth + 7) / 8;
        for (y = 0; y < numRows; y++) {
            for (x = 0; x < rectWidth / 8; x++) {
                for (b = 7; b >= 0; b--)
                    dst[y*rectWidth+x*8+7-b] = palette[src[y*w+x] >> b & 1];
            }
            for (b = 7; b >= 8 - rectWidth % 8; b--) {
                dst[y*rectWidth+x*8+7-b] = palette[src[y*w+x] >> b & 1];
            }
        }
    } else {
        for (y = 0; y < numRows; y++)
            for (x = 0; x < rectWidth; x++)
                dst[y*rectWidth+x] = palette[(int)src[y*rectWidth+x]];
    }
}

#if BPP != 8

/*----------------------------------------------------------------------------
 *
 * JPEG decompression.
 *
 */

/*
  The following variables are defined in rfbproto.c:
  static SDL_bool jpegError;
  static struct jpeg_source_mgr jpegSrcManager;
  static JOCTET *jpegBufferPtr;
  static size_t *jpegBufferLen;
*/

static SDL_bool
DecompressJpegRectBPP(int x, int y, int w, int h)
{
    struct jpeg_decompress_struct cinfo;
    struct my_error_mgr jerr;
    int compressedLen;
    Uint8 *compressedData;
    UintBPP *pixelPtr;
    JSAMPROW rowPointer[2];
    int doFree = FALSE;

    compressedLen = (int)ReadCompactLen();
    if (compressedLen <= 0) {
        fprintf(stderr, "Incorrect data received from the server.\n");
        return SDL_TRUE;
    }

    if ( compressedLen > (BUFFER_SIZE / 2) ) {
        fprintf(stderr, "Compressed Buffer Size: %d\n", compressedLen);
        compressedData = malloc(compressedLen);
        if (compressedData == NULL) {
            fprintf(stderr, "Memory allocation error.\n");
            return SDL_FALSE;
        }
        doFree = TRUE;
    } else {
        compressedData = (Uint8*)&buffer[BUFFER_SIZE / 2];
    }

    if (!ReadFromRFBServer((char*)compressedData, compressedLen)) {
        if ( doFree ) free(compressedData);
        return SDL_FALSE;
    }

	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = my_error_exit;

	if (setjmp(jerr.setjmp_buffer)) {
		fprintf(stderr, "JPEG error. Ignoring...\n");

		/* If we get here, the JPEG code has signaled an error.  We
		 * need to clean up the JPEG object and return.  Returning
		 * TRUE here will cause VNC to seek to next message
		 */
		jpeg_destroy_decompress(&cinfo);
		if ( doFree ) free(compressedData);
		return SDL_TRUE;
	}
	
    jpeg_create_decompress(&cinfo);

    JpegSetSrcManager(&cinfo, compressedData, compressedLen);

    jpeg_read_header(&cinfo, TRUE);

#if BPP == 16
    cinfo.out_color_space = JCS_RGB16_565;
#else
    cinfo.out_color_space = JCS_RGB;
#endif
    cinfo.dct_method = JDCT_FASTEST;
    cinfo.do_fancy_upsampling = FALSE;
    cinfo.do_block_smoothing = FALSE;

    jpeg_start_decompress(&cinfo);
    if (cinfo.output_width != w || cinfo.output_height != h ||
        cinfo.output_components != 3) {
        fprintf(stderr, "Tight Encoding: Wrong JPEG data received.\n");
        jpeg_destroy_decompress(&cinfo);
        if ( doFree ) free(compressedData);
        return SDL_TRUE;
    }


#if BPP == 16
    pixelPtr = (Uint16*)((Uint8*)ovl->pixels + 
                         y * ovl->pitch + 
                         x * ovl->format->BytesPerPixel);

    rowPointer[0] = (JSAMPROW)pixelPtr;
    rowPointer[1] = (JSAMPROW)((Uint8*)rowPointer[0] + ovl->pitch);

    while (cinfo.output_scanline < cinfo.output_height) {
        jpeg_read_scanlines(&cinfo, rowPointer, 2);
        if (jpegError) {
            break;
        }

        rowPointer[0]  = (JSAMPROW)((Uint8*)rowPointer[1] + ovl->pitch);
        rowPointer[1]  = (JSAMPROW)((Uint8*)rowPointer[0] + ovl->pitch);
    }
#else
    rowPointer[0] = (JSAMPROW)buffer;

    pixelPtr = (UintBPP*)((Uint8*)ovl->pixels + 
                          y * ovl->pitch + 
                          x * ovl->format->BytesPerPixel);

    while (cinfo.output_scanline < cinfo.output_height) {
        int byteoffset;
        int dx;
        
        jpeg_read_scanlines(&cinfo, rowPointer, 1);
        if (jpegError) {
            break;
        }

        byteoffset = 0;
        for (dx = 0; dx < w; dx++) {
            pixelPtr[dx] =
                RGB24_TO_PIXEL(BPP, 
                               buffer[byteoffset], 
                               buffer[byteoffset + 1], 
                               buffer[byteoffset + 2]);

            byteoffset += 3;
        }

        pixelPtr = (UintBPP*)((Uint8*)pixelPtr + ovl->pitch);
    }
#endif

    if (!jpegError)
        jpeg_finish_decompress(&cinfo);

    jpeg_destroy_decompress(&cinfo);
    if ( doFree ) free(compressedData);

    return !jpegError;
}

#endif


/*----------------------------------------------------------------------------
 *
 * LZO decompression.
 *
 */
static SDL_bool
DecompressLZORectBPP(int rx, int ry, int rw, int rh)
{
	static SDL_bool isLZOInitDone = SDL_FALSE;
	Uint32 in_bytes;
	Uint32 out_bytes;
	char*  out_buf;
	int    inflateResult = LZO_E_OK;

	if ( !isLZOInitDone ) {
		if ( lzo_init() != LZO_E_OK )
		{
			fprintf(stderr,"LZO library init failed!/n");
			return SDL_TRUE;
		}
		isLZOInitDone = SDL_TRUE;
	}

	/* First make sure we have a large enough raw buffer to hold the
	 * decompressed data.  In practice, with a fixed BPP, fixed frame
	 * buffer size and the first update containing the entire frame
	 * buffer, this buffer allocation should only happen once, on the
	 * first update.
	 */
	if ( raw_buffer_size < (( rw * rh ) * ( BPP / 8 ))) {

		if ( raw_buffer != NULL ) {
			free( raw_buffer );
		}

		raw_buffer_size = (( rw * rh ) * ( BPP / 8 ));
		raw_buffer = (char*) malloc( raw_buffer_size );
	}
	out_buf = raw_buffer;

    /* Read the length (1..3 bytes) of compressed data following. */
    in_bytes = (int)ReadCompactLen();
    if (in_bytes <= 0) {
        fprintf(stderr, "Incorrect data received from the server.\n");
        return SDL_TRUE;
    }
#ifdef DEBUG
	DebugPrint("LZO: data size: %d\n", in_bytes);
#endif

	/* Process buffer full of data until no more to process, or
	 * some type of inflater error, or Z_STREAM_END.
	 */
    if ( in_bytes > lzo_buffer_size ) {
		int new_bufsize = ((in_bytes < LZO_DEFAULT_BUFFER_SIZE) ? 
						   LZO_DEFAULT_BUFFER_SIZE : in_bytes);

        DebugPrint("LZO: Increasing buffer to %d bytes\n", new_bufsize);

        // increase the buffer size
        if ( lzo_buffer ) {
            free(lzo_buffer);
        }
        lzo_buffer = (char*)malloc(new_bufsize);
        lzo_buffer_size = new_bufsize;
    }

    /* Fill the buffer, obtaining data from the server. */
    if (!ReadFromRFBServer(lzo_buffer,in_bytes)) {
        DebugPrint("LZO: ReadFromRFBServer failed to read %d bytes\n", 
                   in_bytes);
        return SDL_FALSE;
    }
#ifdef DEBUG
    DebugPrint("LZO: ReadFromRFBServer read %d bytes\n", in_bytes);
#endif

    /* decompress the entire buffer in one-shot (as required by LZO) */
    inflateResult = lzo1x_decompress(lzo_buffer, in_bytes,
                                     out_buf, &out_bytes,
                                     NULL);
    if ( inflateResult == LZO_E_OK ) {
#ifdef DEBUG
        DebugPrint("LZO: lzo1x_decompress %d bytes in %d bytes out\n",
                   in_bytes, out_bytes);
#endif
		CopyDataToScreen(raw_buffer, rx, ry, rw, rh);
	} 

	return SDL_TRUE;
}


#if BPP == 16
/*----------------------------------------------------------------------------
 *
 * RK decompression.
 *
 */
static int
RKdecode(int x_start, int y_start, int rw, int rh, unsigned char *dstbuf)
{
	short color;                           /* $B?'%3!<%I!J#1#6%S%C%H$rA[Dj!K(B */
	short x, y;                            /* $B:BI8CM(B */
	short run_count;                       /* $B%i%s%+%&%s%H?t(B */
	unsigned short cachecolor0, cachecolor1; /* $B%-%c%C%7%e%+%i!<#0!"#1(B */
	short next_dot_cachecolor;              /* $BNY$N%+%i!<CM(B */
	short x_end, y_end;                    /* $B#X!$#Y$N=*N;CM(B */
	short currentcolor;                    /* $B8=:_$N%+%i!<CM(B */
	short width;                           /* $B#1%i%$%s$NI}(B */
	short masked_colour;
	short* outpixel = (short*)dstbuf;
	unsigned char* srcbuf;
	int   remaining;
	int   bytes_avail;

	/* $BJQ?t=i4|2=(B */
	cachecolor0 = 0;
	cachecolor1 = 0xffff;
	next_dot_cachecolor = 3;
	run_count = 0;
	currentcolor = 0;
	width = 0;
	bytes_avail = 0;
	srcbuf = buffer;

    /* Read the length (1..3 bytes) of compressed data following. */
    remaining = (int)ReadCompactLen();
    if (remaining <= 0) {
        fprintf(stderr, "Incorrect data received from the server.\n");
        return SDL_TRUE;
    }
#ifdef DEBUG
	DebugPrint("RK: data size: %d\n", remaining);
#endif

	/* $B%X%C%@FI$_=P$7(B */
	/* X$B!"(BY$B$rFI$_=P$9!J%S%C%/%(%s%G%#%"%s!K(B */
	if (!ReadFromRFBServer((char *)&x_end, 2)) {
		DebugPrint("RK: Failed to read x_end\n");
		return SDL_FALSE;
	}
	if (!ReadFromRFBServer((char *)&y_end, 2)) {
		DebugPrint("RK: Failed to read y_end\n");
		return SDL_FALSE;
	}
	remaining -= 4;


	x_end = x_start + rw;
	y_end = y_start + rh;
#ifdef DEBUG
	DebugPrint("RK[read]: x_end, y_end: %d,%d\n", x_end, y_end);
#endif

	x = x_start;
	y = y_start;

	while( y < y_end ) {

		if ( bytes_avail < 8192 && remaining > 0 ) {
			int in_bytes;

			if ( bytes_avail > 0 ) {
				// move the remaining bytes to the beginning of the buffer
				memmove(buffer, srcbuf, bytes_avail);
				srcbuf = buffer;
			}

			in_bytes = MIN(remaining, 8192);

			/* Read an approximate line worth of data from the server */
			if (!ReadFromRFBServer(srcbuf + bytes_avail, in_bytes)) {
				DebugPrint("RK: Failed to read RK data\n");				
				return SDL_FALSE;
			}

#ifdef DEBUG
			DebugPrint("RK: from server: %d bytes\n", in_bytes);
#endif
			bytes_avail += in_bytes;
			remaining -= in_bytes;
		}

		while( x < x_end ) {

			if ( next_dot_cachecolor == 3 ) {
				/* $BB(CM%+%i!<(B */

				color = *(short *)srcbuf;			 /* $B?'%G!<%?(B */
				srcbuf += 2;
				bytes_avail -= 2;

				cachecolor1 = cachecolor0;
				cachecolor0 = color & ~0x0020;
				if(( color & 0x0020) != 0) { /* $B#1%I%C%H%S%C%H$,M-8z$N>l9g(B */
					*outpixel = cachecolor0;
					outpixel++;
					x++;
					next_dot_cachecolor = 3; /* $B<!$N%I%C%H$OB(CM%+%i!<(B */
				} else {
					next_dot_cachecolor = 0; /* $B<!$N%I%C%H$O%-%c%C%7%e%+%i!<(B0 */
				}
				
			} else if ( next_dot_cachecolor == 0 || next_dot_cachecolor == 1 ) {
				/* $B%-%c%C%7%e%+%i!<(B0/1 */
				unsigned short usecachecolor = 
					(next_dot_cachecolor == 0) ? cachecolor0 : cachecolor1;

				next_dot_cachecolor = ( *srcbuf >> 6) & 0x03;
					
				if (next_dot_cachecolor == 2) { /* $B%i%s%+%&%s%H?t$,#6#3$rD6$($F$$$k$+$r%A%'%C%/$9$k(B */
					currentcolor = usecachecolor;
					continue;
				}
				run_count = *srcbuf & 0x3f;
				masked_colour = usecachecolor & ~0x0020;

				srcbuf += 1;
				bytes_avail -= 1;
				 
				if( run_count != 0) { /* $BDL>o$N%i%s=hM}(B */
					x += run_count;
					do {
						*outpixel++ = masked_colour;
					}while( --run_count);
				} else { /* $B1&C<$^$G$N%i%s=hM}(B */
					do {
						*outpixel++ = masked_colour;
						x++;
					}while( x < x_end);
				}

			} else if ( next_dot_cachecolor == 2 ) {
				/* $B%i%s%+%&%s%H?t$,#6#3$rD6$($?>l9g$N=hM}(B */
				run_count = (short)(*(srcbuf++)) & 0x003f; /* $B%i%s%+%&%s%H?t$N2<0L(B6$B%S%C%H$r<hF@$9$k(B */
				run_count |= (((short)(*srcbuf) & 0x003f) << 6);	/* $B%i%s%+%&%s%H?t$N>e0L(B6$B%S%C%H$r<hF@$9$k(B */
					
				next_dot_cachecolor = ( *srcbuf >> 6) & 0x03;

				srcbuf += 1;
				bytes_avail -= 2;

				if( run_count != 0) { /* $BDL>o$N%i%s=hM}(B */
					masked_colour = currentcolor & ~0x0020;

					x += run_count;
					do {
						*outpixel++ = masked_colour;
					}while( --run_count);
				}

			} else {
				DebugPrint("RK: ERROR!!!\n");
				return SDL_TRUE;
			}
		}

		/* $B9T$N=*N;$KEv$?$C$F$N3F<oJQ?t=i4|2=(B */		
		x = x_start;
		y++;
		cachecolor0 = 0;
		cachecolor1 = 0xffff;
		next_dot_cachecolor = 3;
		run_count = 0;
	}

#ifdef DEBUG
	DebugPrint("RK: Bytes available: %d\n", bytes_avail);
	DebugPrint("RK: Bytes remaining: %d\n", remaining);
	DebugPrint("RK: decoded lines: %d/%d\n", y,rh);
#endif

	return SDL_TRUE;	
}

static SDL_bool
DecompressRKRectBPP(int rx, int ry, int rw, int rh)
{
	/* First make sure we have a large enough raw buffer to hold the
	 * decompressed data.  In practice, with a fixed BPP, fixed frame
	 * buffer size and the first update containing the entire frame
	 * buffer, this buffer allocation should only happen once, on the
	 * first update.
	 */
	if ( raw_buffer_size < (( rw * rh ) * ( BPP / 8 ))) {

		if ( raw_buffer != NULL ) {
			free( raw_buffer );
		}

		raw_buffer_size = (( rw * rh ) * ( BPP / 8 ));
		raw_buffer = (char*) malloc( raw_buffer_size );
	}

	if ( RKdecode(rx, ry, rw, rh, raw_buffer) ) {
		CopyDataToScreen(raw_buffer, rx, ry, rw, rh);
	} else {
		DebugPrint("RKdecode ERROR\n");
	}

	return SDL_TRUE;
}
#endif
