/*
 *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * desktop.c - functions to deal with "desktop" window.
 */
#include <stddef.h>
#include "libkeyinput.h"
#include "imconfig.h"
#include <vncviewer.h>
#include <SDL.h>

static SDL_Cursor* CreateDotCursor();
static void CopyBGR233ToScreen(Uint8 *buf, int x, int y, int width,int height);
static void PrintVideoInfo();

SDL_Surface* dpy;
SDL_Surface* ovl;

static SDL_Overlay* hw_ovl;
static Uint8 hw_ovl_BytesPerPixel;
static SDL_Rect hw_screen_rect;

/*
 * DesktopInitBeforeRealization creates the "desktop" widget and the viewport
 * which controls it.
 */

void
DesktopInitBeforeRealization()
{
    /* Don't show the local cursor */
    SDL_ShowCursor(0);

#if defined(mips)
    /* Finally try to get the input focus.  With some WMs we might
       have to grab the keyboard, but this seems to be OK with the
       ones I've tried. */
    SDL_WM_GrabInput(SDL_GRAB_ON);
#endif

    if (!dpy) {
        SDL_Surface* image;
		SDL_Overlay* overlay;
        int flags;
		unsigned int ovlformat = 0;
		Uint32 Rmask, Gmask, Bmask, Amask;

        DebugPrint( "Best video mode:\n");
        PrintVideoInfo();

#if defined(mips)
        flags = SDL_FULLSCREEN|SDL_HWSURFACE;
#else
        flags = SDL_HWSURFACE;
#endif
        image = SDL_SetVideoMode(LCD_WIDTH, LCD_HEIGHT, 32, flags);
        if ( image == NULL ) {
			DebugPrint( "Failed to create display surface\n");
            Cleanup(ERROR_INIT);
        }
		SDL_SetAlpha(image, SDL_SRCALPHA, SDL_ALPHA_OPAQUE);

        /* Set the colormap */
        if ( image->format->palette ) {
            SDL_SetPalette(image, SDL_LOGPAL|SDL_PHYSPAL, cmap.colors,
                           0, cmap.ncolors);
        }

        DebugPrint( "Using video mode:\n");
        PrintVideoInfo();
        DebugPrint( "\tSurface BPP: %d\n", image->format->BitsPerPixel);
        DebugPrint( "\tSurface Dimensions: %d x %d\n", image->w, image->h);
        DebugPrint( "\tSurface pitch: %d bytes\n", image->pitch);


		/* Create the YUV overlay (actually its RGB... but whatever) */
		if ( si.format.depth == 32 ) {
			hw_ovl_BytesPerPixel = 4;
			ovlformat = SDL_ARGB_OVERLAY;
			Amask = 0x00000000;
			Rmask = 0x00FF0000;
			Gmask = 0x0000FF00;
			Bmask = 0x000000FF;
		}
		else if ( si.format.depth == 16 ) {
			hw_ovl_BytesPerPixel = 2;
			ovlformat = SDL_RGB565_OVERLAY;
			Amask = 0x00000000;
			Rmask = 0x0000F800;
			Gmask = 0x000007E0;
			Bmask = 0x0000001F;
		} else {
            Cleanup(ERROR_INIT);
			return;		/* this return silences a GCC warning */
		}
        overlay = SDL_CreateYUVOverlay(si.framebufferWidth,
									   si.framebufferHeight,
									   ovlformat, image);
		if ( overlay == NULL ) {
			DebugPrint( "Failed to create YUV overlay\n");
            Cleanup(ERROR_INIT);
		}
		DebugPrint( "Created YUV overlay:\n");
		DebugPrint( "\tformat: %d\n", overlay->format);
		DebugPrint( "\twidth: %d\n", overlay->w);
		DebugPrint( "\theight: %d\n", overlay->h);
		DebugPrint( "\tplanes: %d\n", overlay->planes);
		DebugPrint( "\tpitch plane 0: %d\n", overlay->pitches[0]);
		DebugPrint( "\tpixels plane 0: 0x%8.8X\n", (int)overlay->pixels[0]);

		/* create the overlay shadow buffer */
		ovl = SDL_CreateRGBSurface(SDL_SWSURFACE, 
								   overlay->w, 
								   overlay->h,
								   hw_ovl_BytesPerPixel * 8,	/* bits per pixel */
								   Rmask, Gmask, Bmask, Amask);
		if ( ovl == NULL ) {
			DebugPrint( "Failed to create overlay shadow buffer\n");
            Cleanup(ERROR_INIT);
		}

		/* initialize the global surface pointers */
        dpy = image;
		hw_ovl = overlay;
    }
}


/*
 * DesktopInitAfterRealization does things which require the X windows to
 * exist.  It creates some GCs and sets the dot cursor.
 */

void
DesktopInitAfterRealization()
{
    Uint32 colour;
    SDL_Cursor* dotCursor;

    /* Fill the GUI background with black */
    colour = SDL_MapRGBA(dpy->format, 0, 0, 0, SDL_ALPHA_OPAQUE);
    SDL_FillRect(dpy, NULL, colour);
    SDL_UpdateRect(dpy, 0, 0, 0, 0);
	DebugPrint("GUI memory cleared\n");

    /* Fill the overlay background with black */
    colour = SDL_MapRGB(ovl->format, 0, 0, 0);
	SDL_FillRect(ovl, NULL, colour);
    UpdateOverlayRect(ovl, 0, 0, 0, 0);
	DebugPrint("Overlay memory cleared\n");

    /* Set the cursor */
    dotCursor = CreateDotCursor();
    SDL_SetCursor(dotCursor);

	/* Check and adjust for screen aspect */
	hw_screen_rect.x = 0;
	hw_screen_rect.y = 0;
	hw_screen_rect.w = LCD_WIDTH;
	hw_screen_rect.h = LCD_HEIGHT;
   
	/* the ratio of width to height is compared to 4:3
	 * (LCD_WIDTH:LCD_HEIGHT)
	 */
	if ( (si.framebufferWidth * 3) > (si.framebufferHeight * 4) ) {
		hw_screen_rect.h = LCD_WIDTH * si.framebufferHeight / si.framebufferWidth;
		hw_screen_rect.y = (LCD_HEIGHT - hw_screen_rect.h) / 2;
	} else if ((si.framebufferWidth * 3) < (si.framebufferHeight * 4)) {
		hw_screen_rect.w = LCD_HEIGHT * si.framebufferWidth / si.framebufferHeight;
		hw_screen_rect.x = (LCD_WIDTH - hw_screen_rect.w) / 2;
	}

	/* Initialize the UI resource manager */
	InitUI();

	/* update the black border area */
	SetVideoArea(&hw_screen_rect);

	/* Show the YUV overlay */
	SDL_DisplayYUVOverlay(hw_ovl, &hw_screen_rect);
}


/*
 * HandleBasicDesktopEvent - deal with expose and leave events.
 */

void
HandleBasicDesktopEvent(SDL_Event* event)
{
    switch (event->type) {

    case SDL_VIDEOEXPOSE:
        SendFramebufferUpdateRequest(0, 0, ovl->w, ovl->h, SDL_FALSE);
        break;

    case SDL_ACTIVEEVENT:
        {
            if ( event->active.gain ) {
                SendFramebufferUpdateRequest(0, 0, ovl->w, ovl->h, SDL_FALSE);
            }
        }
        break;
    }
}


/*
 * SendRFBEvent is an action which sends an RFB event.  It can be used in two
 * ways.  Without any parameters it simply sends an RFB event corresponding to
 * the X event which caused it to be called.  With parameters, it generates a
 * "fake" RFB event based on those parameters.  The first parameter is the
 * event type, either "ptr", "keydown", "keyup" or "key" (down&up).  For a
 * "key" event the second parameter is simply a keysym string as understood by
 * XStringToKeysym().  For a "ptr" event, the following three parameters are
 * just X, Y and the button mask (0 for all up, 1 for button1 down, 2 for
 * button2 down, 3 for both, etc).
 */

void
SendRFBEvent(SDL_Event* ev, String *params, int *num_params)
{
    if (appData.fullScreen && ev->type == SDL_MOUSEMOTION) {
        if (BumpScroll(ev))
            return;
    }

    switch (ev->type) {
        Uint8 buttons;
        Uint32 keys;

    case SDL_MOUSEMOTION:
        while (SDL_PeepEvents(ev, 1, SDL_GETEVENT, SDL_MOUSEMOTIONMASK) == 1)
            ;	/* discard all queued motion notify events */

        SendPointerEvent(ev->motion.x, ev->motion.y, ev->motion.state);
        break;

    case SDL_MOUSEBUTTONDOWN:
        buttons = ev->button.state << ev->button.button;
        SendPointerEvent(ev->button.x, ev->button.y, buttons);
        break;

    case SDL_MOUSEBUTTONUP:
        buttons = ev->button.state << ev->button.button;
        SendPointerEvent(ev->button.x, ev->button.y, buttons);
        break;

    case SDL_KEYDOWN:
    case SDL_KEYUP:
        keys = ev->key.keysym.sym;
        SendKeyEvent(keys, (ev->type == SDL_KEYDOWN));
        break;

    default:
        DebugPrint("Invalid event passed to SendRFBEvent\n");
        break;
    }
}


/*
 * CreateDotCursor.
 */

static SDL_Cursor*
CreateDotCursor()
{
    SDL_Cursor* cursor;
    static char srcBits[] = { 0, 14,14,14, 0 };
    static char mskBits[] = { 14,31,31,31,14 };

    cursor = SDL_CreateCursor(srcBits, mskBits, 8, 5, 5, 2);
    return cursor;
}


/*
 * CopyDataToScreen.
 */

void
CopyDataToScreen(char *buf, int x, int y, int width, int height)
{
    if (!appData.useBGR233) {
        int widthInBytes = width * (si.format.bitsPerPixel / 8);
        int scrnWidthInBytes = ovl->pitch;
        char* scrn;

#ifdef ARGB1555
		int w;
		Uint16* src;
		Uint16* dst;
#endif

        scrn = ((Uint8*)ovl->pixels +
               y * scrnWidthInBytes +
               x * ovl->format->BytesPerPixel);

        while( height-- ) {
#ifdef ARGB1555
            src = (Uint16*)buf;
            dst = (Uint16*)scrn;
            for( w = 0; w < width; w++ ) {
                dst[w] = (src[w] & 0x001F) | ((src[w] >> 1) & 0x7FE0);
            }
#else
            memcpy(scrn, buf, widthInBytes);
#endif

            buf += widthInBytes;
            scrn += scrnWidthInBytes;
        }

    } else {
        CopyBGR233ToScreen((Uint8 *)buf, x, y, width, height);
    }
}


/*
 * CopyBGR233ToScreen.
 */

static void
CopyBGR233ToScreen(Uint8 *buf, int x, int y, int width, int height)
{
    int p, q;
    int xoff = 7 - (x & 7);
    int xcur;
    int fbwb = si.framebufferWidth / 8;
    Uint8 *scr1 = ((Uint8 *)ovl->pixels) + y * fbwb + x / 8;
    Uint8 *scrt;
    Uint8 *scr8 = ((Uint8 *)ovl->pixels) + y * si.framebufferWidth + x;
    Uint16 *scr16 = ((Uint16 *)ovl->pixels) + y * si.framebufferWidth + x;
    Uint32 *scr32 = ((Uint32 *)ovl->pixels) + y * si.framebufferWidth + x;

    switch (myFormat.bitsPerPixel) {

        /* thanks to Chris Hooper for single bpp support */

    case 1:
        for (q = 0; q < height; q++) {
            xcur = xoff;
            scrt = scr1;
            for (p = 0; p < width; p++) {
                *scrt = ((*scrt & ~(1 << xcur))
                         | (BGR233ToPixel[*(buf++)] << xcur));

                if (xcur-- == 0) {
                    xcur = 7;
                    scrt++;
                }
            }
            scr1 += fbwb;
        }
        break;

    case 8:
        for (q = 0; q < height; q++) {
            for (p = 0; p < width; p++) {
                *(scr8++) = BGR233ToPixel[*(buf++)];
            }
            scr8 += si.framebufferWidth - width;
        }
        break;

    case 16:
        for (q = 0; q < height; q++) {
            for (p = 0; p < width; p++) {
                *(scr16++) = BGR233ToPixel[*(buf++)];
            }
            scr16 += si.framebufferWidth - width;
        }
        break;

    case 32:
        for (q = 0; q < height; q++) {
            for (p = 0; p < width; p++) {
                *(scr32++) = BGR233ToPixel[*(buf++)];
            }
            scr32 += si.framebufferWidth - width;
        }
        break;
    }
}

static void
PrintVideoInfo()
{
#ifndef NDEBUG
    const SDL_VideoInfo* vinfo;
    char  driver[256];

    vinfo = SDL_GetVideoInfo();
    SDL_VideoDriverName(driver, 256);

    DebugPrint( "Video Info for driver: %s\n", driver);
    DebugPrint( "\tHW available: %d\n", vinfo->hw_available);
    DebugPrint( "\tHW blit: %d\n", vinfo->blit_hw);
    DebugPrint( "\tHW blit CC: %d\n", vinfo->blit_hw_CC);
    DebugPrint( "\tHW blit Alpha: %d\n", vinfo->blit_hw_A);
    DebugPrint( "\tSW blit: %d\n", vinfo->blit_sw);
    DebugPrint( "\tSW blit CC: %d\n", vinfo->blit_sw_CC);
    DebugPrint( "\tSW blit Alpha: %d\n", vinfo->blit_sw_A);
    DebugPrint( "\tBytes per pixel: %d\n", vinfo->vfmt->BytesPerPixel);
    DebugPrint( "\tBits per pixel: %d\n", vinfo->vfmt->BitsPerPixel);
#endif
}

void 
UpdateOverlayRect(SDL_Surface* src, 
				  Sint16 x, Sint16 y,
				  Uint16 w, Uint16 h)
{
	Uint8* src_row;
	Uint8* dst_row;
	
	assert( hw_ovl != NULL && src != NULL );
	assert( src->format->BytesPerPixel == hw_ovl_BytesPerPixel );

	if ( x == 0 && y == 0 && w == 0 && h == 0 ) {
		w = src->w;
		h = src->h;
	}

	src_row = (Uint8*)src->pixels + y * src->pitch +
		x * src->format->BytesPerPixel;
	dst_row = (Uint8*)hw_ovl->pixels[0] + y * hw_ovl->pitches[0] +
		x * hw_ovl_BytesPerPixel;

	SDL_LockYUVOverlay(hw_ovl);

	while( h-- ) {
		memcpy(dst_row, src_row, w * hw_ovl_BytesPerPixel);
		src_row += src->pitch;
		dst_row += hw_ovl->pitches[0];
	}

	SDL_UnlockYUVOverlay(hw_ovl);
}

void ZoomOverlay(SDL_Rect* zoomrect)
{
	static SDL_Rect current_zoom_rect = {0, 0, 0, 0};

	if ( zoomrect == NULL &&
		 current_zoom_rect.x == 0 &&
		 current_zoom_rect.y == 0 &&
		 current_zoom_rect.w == 0 &&
		 current_zoom_rect.h == 0 ) {
		// already zoomed out to fullscreen
		return;
	} else if ( zoomrect &&
				current_zoom_rect.x == zoomrect->x &&
				current_zoom_rect.y == zoomrect->y &&
				current_zoom_rect.w == zoomrect->w &&
				current_zoom_rect.h == zoomrect->h ) {
		// already zoomed into same area
		return;
	}

	if ( zoomrect == NULL ) {
		/* update the black border area */
		SetVideoArea(&hw_screen_rect);

		/* Reset the YUV overlay position */
		SDL_DisplayYUVOverlay(hw_ovl, &hw_screen_rect);
		SDL_ZoomOverlay(hw_ovl, NULL);

		current_zoom_rect.x = 0;
		current_zoom_rect.y = 0;
		current_zoom_rect.w = 0;
		current_zoom_rect.h = 0;

	} else {
		SDL_Rect new_hw_rect = {0, 0, LCD_WIDTH, LCD_HEIGHT };
		SDL_Rect new_zoom_rect = *zoomrect;

		// save the requested zoom rectangle
		current_zoom_rect = new_zoom_rect;

#ifndef NDEBUG
		fprintf(stderr, "Prescale zoom rect: %4dx%4d+%4d+%4d\n",				
				new_zoom_rect.w, new_zoom_rect.h,
				new_zoom_rect.x, new_zoom_rect.y );
#endif

		// recalculate screen dimensions if aspect is not 4:3
		if ((new_zoom_rect.w * 3) != (new_zoom_rect.h * 4)) {
			if ((new_zoom_rect.w * 3) > (new_zoom_rect.h * 4) ) {
				new_hw_rect.h = LCD_WIDTH * new_zoom_rect.h / new_zoom_rect.w;
				new_hw_rect.y = (LCD_HEIGHT - new_hw_rect.h) / 2;
			} else {
				new_hw_rect.w = LCD_HEIGHT * new_zoom_rect.w / new_zoom_rect.h;
				new_hw_rect.x = (LCD_WIDTH - new_hw_rect.w) / 2;
			}
		}

		// scale the zoom rect to match the display size
		new_zoom_rect.x = new_zoom_rect.x * new_hw_rect.w / ovl->w + new_hw_rect.x; 
		new_zoom_rect.w = new_zoom_rect.w * new_hw_rect.w / ovl->w; 
		new_zoom_rect.y = new_zoom_rect.y * new_hw_rect.h / ovl->h + new_hw_rect.y; 
		new_zoom_rect.h = new_zoom_rect.h * new_hw_rect.h / ovl->h; 

#ifndef NDEBUG
		fprintf(stderr, "HW Rect: %4dx%4d+%4d+%4d\tZoom rect: %4dx%4d+%4d+%4d\n",
				new_hw_rect.w, new_hw_rect.h,
				new_hw_rect.x, new_hw_rect.y,
				new_zoom_rect.w, new_zoom_rect.h,
				new_zoom_rect.x, new_zoom_rect.y );
#endif

		/* update the black border area */
		SetVideoArea(&new_hw_rect);

		SDL_DisplayYUVOverlay(hw_ovl, &new_hw_rect);
		SDL_ZoomOverlay(hw_ovl, &new_zoom_rect);
	}
}

void CleanupVideoSurfaces()
{
	if ( ovl ) {
		SDL_FreeSurface(ovl);
		ovl = NULL;
	}

	if ( hw_ovl ) {
		SDL_FreeYUVOverlay(hw_ovl);
		hw_ovl = NULL;
	}
}
