/*
 *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * fullscreen.c - functions to deal with full-screen mode.
 */

#include <vncviewer.h>

static SDL_TimerID timer;
static SDL_bool timerSet = SDL_FALSE;
static SDL_bool scrollLeft, scrollRight, scrollUp, scrollDown;
static int desktopX, desktopY;

static SDL_bool DoBumpScroll();
static Uint32 BumpScrollTimerCallback(Uint32 interval, void* param);



/*
 * FullScreenOn goes into full-screen mode.  It makes the toplevel window
 * unmanaged by the window manager and sets its geometry appropriately.
 *
 * We have toplevel -> form -> viewport -> desktop.  "form" must always be the
 * same size as "toplevel".  "desktop" should always be fixed at the size of
 * the VNC desktop.  Normally "viewport" is the same size as "toplevel" (<=
 * size of "desktop"), and "viewport" deals with any difference by putting up
 * scrollbars.
 *
 * When we go into full-screen mode, we allow "viewport" and "form" to be
 * different sizes, and we effectively need to work out all the geometries
 * ourselves.  There are two cases to deal with:
 *
 * 1. When the desktop is smaller than the display, "viewport" is simply the
 *    size of the desktop and "toplevel" (and "form") are the size of the
 *    display.  "form" is visible around the edges of the desktop.
 *
 * 2. When the desktop is bigger than the display in either or both dimensions,
 *    we force "viewport" to have scrollbars.
 *
 *    If the desktop width is bigger than the display width, then the width of
 *    "viewport" is the display width plus the scrollbar width, otherwise it's
 *    the desktop width plus the scrollbar width.  The width of "toplevel" (and
 *    "form") is then either the same as "viewport", or just the display width,
 *    respectively.  Similarly for the height of "viewport" and the height of
 *    "toplevel".
 *
 *    So if the desktop is bigger than the display in both dimensions then both
 *    the scrollbars will be just off the screen.  If it's bigger in only one
 *    dimension then that scrollbar _will_ be visible, with the other one just
 *    off the screen.  We treat this as a "feature" rather than a problem - you
 *    can't easily get around it if you want to use the Athena viewport for
 *    doing the scrolling.
 *
 * In either case, we position "viewport" in the middle of "form".
 *
 * We store the calculated size of "viewport" and the scrollbars in global
 * variables so that FullScreenOff can use them.
 */

void
FullScreenOn()
{
    appData.fullScreen = SDL_TRUE;

    /* Toggle the WM to fullscreen mode */
    SDL_WM_ToggleFullScreen(dpy);

    /* Don't show the local cursor */
    SDL_ShowCursor(0);

    /* Finally try to get the input focus.  With some WMs we might have to grab
       the keyboard, but this seems to be OK with the ones I've tried. */
    SDL_WM_GrabInput(SDL_GRAB_ON);    
}


/*
 * FullScreenOff leaves full-screen mode.  It makes the toplevel window
 * managed by the window manager and sets its geometry appropriately.
 *
 * We also want to reestablish the link between the geometry of "form" and
 * "viewport".  We do this similarly to the way we broke it in FullScreenOn, by
 * making "viewport" unmanaged, changing certain resources on it and asking
 * "form" to manage it again.
 *
 * There seems to be a slightly strange behaviour with setting forceBars back
 * to false, which results in "desktop" being stretched by the size of the
 * scrollbars under certain circumstances.  Resizing both "toplevel" and
 * "viewport" to the full-screen viewport size minus the scrollbar size seems
 * to fix it, though I'm not entirely sure why. */

void
FullScreenOff()
{
    appData.fullScreen = SDL_FALSE;

    /* Toggle the WM to fullscreen mode */
    SDL_WM_ToggleFullScreen(dpy);

    /* Show the local cursor */
    SDL_ShowCursor(1);

    /* Finally try to get the input focus.  With some WMs we might have to grab
       the keyboard, but this seems to be OK with the ones I've tried. */
    SDL_WM_GrabInput(SDL_GRAB_OFF);
}


/*
 * ToggleFullScreen is an action which toggles in and out of full-screen mode.
 */

void
ToggleFullScreen()
{
    if (appData.fullScreen) {
        FullScreenOff();
    } else {
        FullScreenOn();
    }
}


/*
 * BumpScroll is called when in full-screen mode and the mouse is against one
 * of the edges of the screen.  It returns true if any scrolling was done.
 */

SDL_bool
BumpScroll(SDL_Event* ev)
{
    scrollLeft = scrollRight = scrollUp = scrollDown = SDL_FALSE;

    if (ev->motion.x == ovl->w - 1)
        scrollRight = SDL_TRUE;
    else if (ev->motion.x == 0)
        scrollLeft = SDL_TRUE;

    if (ev->motion.y == ovl->h - 1)
        scrollDown = SDL_TRUE;
    else if (ev->motion.y == 0)
        scrollUp = SDL_TRUE;

    if (scrollLeft || scrollRight || scrollUp || scrollDown) {
        if (timerSet)
            return SDL_TRUE;

        return DoBumpScroll();
    }

    if (timerSet) {
        SDL_RemoveTimer(timer);
        timerSet = SDL_FALSE;
    }

    return SDL_FALSE;
}

static SDL_bool
DoBumpScroll()
{
    int oldx = desktopX, oldy = desktopY;

    if (scrollRight) {
        if (desktopX < si.framebufferWidth - ovl->w) {
            desktopX += appData.bumpScrollPixels;
            if (desktopX > si.framebufferWidth - ovl->w)
                desktopX = si.framebufferWidth - ovl->w;
        }
    } else if (scrollLeft) {
        if (desktopX > 0) {
            desktopX -= appData.bumpScrollPixels;
            if (desktopX < 0)
                desktopX = 0;
        }
    }

    if (scrollDown) {
        if (desktopY < si.framebufferHeight - ovl->h) {
            desktopY += appData.bumpScrollPixels;
            if (desktopY > si.framebufferHeight - ovl->h)
                desktopY = si.framebufferHeight - ovl->h;
        }
    } else if (scrollUp) {
        if (desktopY > 0) {
            desktopY -= appData.bumpScrollPixels;
            if (desktopY < 0)
                desktopY = 0;
        }
    }

    if (oldx != desktopX || oldy != desktopY) {
#if 0
        // TODO: Should copy the surface to simulate scrolling
        XawViewportSetCoordinates(viewport, desktopX, desktopY);
#endif
        timer = SDL_AddTimer(appData.bumpScrollTime, BumpScrollTimerCallback, NULL);
        timerSet = SDL_TRUE;
        return SDL_TRUE;
    }

    timerSet = SDL_FALSE;
    return SDL_FALSE;
}

Uint32
BumpScrollTimerCallback(Uint32 interval, void* param)
{
    DoBumpScroll();

    return interval;
}
