/*
 *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * misc.c - miscellaneous functions.
 */

#include <signal.h>
#include "imconfig.h"
#include "vnes.h"
#include <vncviewer.h>
#include "libkeyinput.h"

#ifdef USE_ESCVP21
#include "SDL_e21.h"
#endif

static void ShutdownVNC(int code);
static void CleanupSignalHandler(int sig);

/*
 * ToplevelInitBeforeRealization sets the title, geometry and other resources
 * on the toplevel window.
 */

void
ToplevelInitBeforeRealization()
{
    struct sigaction act;

    /* Set the window caption and icon name */
    SDL_WM_SetCaption("sdlvnc", "sdlvnc");

    /* Test if the keyboard is grabbed.  If so, it's probably because the
       XDM login window is up, so try iconifying it to release the grab */

    if (SDL_WM_GrabInput(SDL_GRAB_QUERY) == SDL_GRAB_ON) {
        SDL_WM_GrabInput(SDL_GRAB_OFF);
    }

    /* Set handlers for signals and X errors to perform cleanup */
    memset(&act, 0, sizeof(act));
    act.sa_handler = CleanupSignalHandler;
    sigemptyset(&act.sa_mask);
    sigaddset(&act.sa_mask, SIGHUP);
    sigaddset(&act.sa_mask, SIGTERM);
    sigaddset(&act.sa_mask, SIGPIPE);
    sigaction(SIGHUP, &act, NULL);
    sigaction(SIGTERM, &act, NULL);
    sigaction(SIGPIPE, &act, NULL);
}


/*
 * ToplevelInitAfterRealization initialises things which require the X windows
 * to exist.  It goes into full-screen mode if appropriate, and tells the
 * window manager we accept the "delete window" message.
 */

void
ToplevelInitAfterRealization()
{
    if ( !appData.fullScreen ) {
        FullScreenOff();
    }
}


/*
 * Cleanup - perform any cleanup operations prior to exiting.
 */

void
Cleanup(int code)
{
    sigset_t sset, pending;

    DebugPrint("Cleanup called\n");

	// block the signal handlers temporarily while shutting down SDL
	// and ACL as they can react badly when interrupted
    sigemptyset(&sset);
    sigaddset(&sset, SIGHUP);
    sigaddset(&sset, SIGTERM);
    sigaddset(&sset, SIGPIPE);
    sigprocmask(SIG_BLOCK, &sset, NULL);
    DebugPrint("SIGNALS BLOCKED\n");

	// clean and terminate all resources
    ShutdownVNC(code);

	// save the pending signals state.  We use this to determine if a
	// signal was recieved while we hand all signals blocked.  If
	// there is a pending signal, then we assume that we were shut
	// down from the PC side, even if there was a ERROR_SOCKET.
	sigpending(&pending);

	// if the isFinished flag is set, modify the exit code to success
	if (appData.isFinished || sigismember(&pending, SIGTERM) ) {
		code = EXIT_SUCCESS;
	}

    DebugPrint("VNeS exiting: %d\n", code);
    exit(code);
}

/*
 * ShutdownVNC - Do that actual shutdown of all sockets etc.  This
 * function must be called with the appropriate signals blocked.
 */
static void
ShutdownVNC(int code)
{
    Uint32 black;

    DebugPrint("ShutdownVNC called\n");

#ifdef BENCHMARK
    DebugPrint( "Encrypted Bytes: %d  Encrypt Time: %d ms\n",
            encryptBytes, encryptTime);
    DebugPrint( "Decrypted Bytes: %d  Decrypt Time: %d ms\n",
            decryptBytes, decryptTime);
#endif /* BENCHMARK */

    DebugPrint("Clearing screen to black\n");
    if ( dpy ) {
        black = SDL_MapRGBA(dpy->format, 0, 0, 0, SDL_ALPHA_OPAQUE);
        SDL_FillRect(dpy, NULL, black);
        SDL_UpdateRect(dpy, 0, 0, 0, 0);
    }

    DebugPrint("Clearing screen to black\n");
    if ( ovl ) {
        black = SDL_MapRGB(ovl->format, 0,0,0);
		SDL_FillRect(ovl, NULL, black);
		UpdateOverlayRect(ovl, 0, 0, 0, 0);
    }

	CleanupVideoSurfaces();

	/* Free up resourced used by the UI */
	DestroyUI();

#ifdef USE_ESCVP21
    DebugPrint("Calling SDL_E21Quit\n");
    SDL_E21Quit();  // libSDLescvp21.so quit
#endif //USE_ESCVP21

    DebugPrint("Calling DisconnectDomainSocket\n");
    DisconnectDomainSocket();

#ifdef USE_ESCVP21
    DebugPrint("Calling LKI_CloseSocket\n");
    LKI_CloseSocket();
#endif

    DebugPrint("Calling SDL_Quit\n");
    SDL_Quit();
}

static void
CleanupSignalHandler(int sig)
{
    DebugPrint("CleanupSignalHandler called: %d\n", sig);

    if ( rfbsockTCP > 0 ) {
        DebugPrint("Closing TCP socket\n");
        close(rfbsockTCP);
        rfbsockTCP = -1;
    }

    appData.isFinished = SDL_TRUE;
}
