/*
 *  Copyright (C) 2000, 2001 Const Kaplinsky.  All Rights Reserved.
 *  Copyright (C) 2000 Tridia Corporation.  All Rights Reserved.
 *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * rfbproto.c - functions to deal with client side of RFB protocol.
 */

#include <unistd.h>
#include <errno.h>
#include <pwd.h>
#include <vncviewer.h>
#include <vncauth.h>
#include <zlib.h>
#include <jpeglib.h>
#include <setjmp.h>
#include <minilzo.h>
#include "config.h"
#include "vnes.h"

static SDL_bool HandleRRE8(int rx, int ry, int rw, int rh);
static SDL_bool HandleRRE16(int rx, int ry, int rw, int rh);
static SDL_bool HandleRRE32(int rx, int ry, int rw, int rh);
static SDL_bool HandleCoRRE8(int rx, int ry, int rw, int rh);
static SDL_bool HandleCoRRE16(int rx, int ry, int rw, int rh);
static SDL_bool HandleCoRRE32(int rx, int ry, int rw, int rh);
static SDL_bool HandleHextile8(int rx, int ry, int rw, int rh);
static SDL_bool HandleHextile16(int rx, int ry, int rw, int rh);
static SDL_bool HandleHextile32(int rx, int ry, int rw, int rh);
static SDL_bool HandleZlib8(int rx, int ry, int rw, int rh);
static SDL_bool HandleZlib16(int rx, int ry, int rw, int rh);
static SDL_bool HandleZlib32(int rx, int ry, int rw, int rh);
static SDL_bool HandleTight8(int rx, int ry, int rw, int rh);
static SDL_bool HandleTight16(int rx, int ry, int rw, int rh);
static SDL_bool HandleTight32(int rx, int ry, int rw, int rh);

static long ReadCompactLen (void);

static void JpegInitSource(j_decompress_ptr cinfo);
static boolean JpegFillInputBuffer(j_decompress_ptr cinfo);
static void JpegSkipInputData(j_decompress_ptr cinfo, long num_bytes);
static void JpegTermSource(j_decompress_ptr cinfo);
static void JpegSetSrcManager(j_decompress_ptr cinfo, Uint8 *compressedData,
                              int compressedLen);

struct my_error_mgr {
	struct jpeg_error_mgr pub;	/* "public" fields */
	jmp_buf setjmp_buffer;		/* for return to caller */
};

static void my_error_exit(j_common_ptr cinfo)
{
	/* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
	struct my_error_mgr* myerr = (struct my_error_mgr*) cinfo->err;

	DebugPrint("JPEG decode error\n");

	/* Return control to the setjmp point */
	longjmp(myerr->setjmp_buffer, 1);
}


int rfbsockTCP;

char *desktopName;
rfbPixelFormat myFormat;
rfbServerInitMsg si;
char *serverCutText = NULL;
SDL_bool newServerCutText = SDL_FALSE;

int endianTest = 1;

/* Note that the CoRRE encoding uses this buffer and assumes it is big
   enough to hold 255 * 255 * 32 bits -> 260100 bytes.  640*480 =
   307200 bytes.  Hextile also assumes it is big enough to hold 16 *
   16 * 32 bits.  Tight encoding assumes BUFFER_SIZE is at least 16384
   bytes.
*/

#define BUFFER_SIZE (512*1024)
static char buffer[BUFFER_SIZE];


/* The zlib encoding requires expansion/decompression/deflation of the
   compressed data in the "buffer" above into another, result buffer.
   However, the size of the result buffer can be determined precisely
   based on the bitsPerPixel, height and width of the rectangle.  We
   allocate this buffer one time to be the full size of the buffer. */

static int raw_buffer_size = -1;
static char *raw_buffer;

static z_stream decompStream;
static SDL_bool decompStreamInited = SDL_FALSE;


/*
 * Variables for the ``tight'' encoding implementation.
 */

/* Separate buffer for compressed data. */
#define ZLIB_BUFFER_SIZE 512
static char zlib_buffer[ZLIB_BUFFER_SIZE];

/* Four independent compression streams for zlib library. */
static z_stream zlibStream[4];
static SDL_bool zlibStreamActive[4] = {
    SDL_FALSE, SDL_FALSE, SDL_FALSE, SDL_FALSE
};

/* Filter stuff. Should be initialized by filter initialization code. */
static SDL_bool cutZeros;
static int rectWidth, rectColors;
static char tightPalette[256*4];
static Uint8 tightPrevRow[2048*3*sizeof(Uint16)];

/* JPEG decoder state. */
static SDL_bool jpegError;

/* LZO requires a large buffer to read in a whole block at once. 
   The buffer will grow as required by LZO decompress algorithm.
*/
static int   lzo_buffer_size = 0;
static char* lzo_buffer;



static inline Sint32 MAX(const Sint32 a, const Sint32 b) {
    return ((a > b) ? a : b);
}
static inline Sint32 MIN(const Sint32 a, const Sint32 b) {
    return ((a < b) ? a : b);
}

/*
 * ConnectToRFBServer.
 */

SDL_bool
ConnectToRFBServer(const char *hostname, int port)
{
    unsigned int host;

    if (!StringToIPAddr(hostname, &host)) {
        DebugPrint("Couldn't convert '%s' to host address\n", hostname);
        return SDL_FALSE;
    }

    rfbsockTCP = ConnectToTcpAddr(host, port);
    if (rfbsockTCP < 0) {
        DebugPrint("Unable to connect to VNC server\n");
        return SDL_FALSE;
    }

    return SetNonBlocking(rfbsockTCP);
}


/*
 * InitialiseRFBConnection.
 */

SDL_bool
InitialiseRFBConnection()
{
    rfbProtocolVersionMsg pv;
    int major, minor;

#if CAN_SEND
    Uint32 authScheme, reasonLen, authResult;
    char *reason;
    Uint8 challenge[CHALLENGESIZE];
    char *passwd;
    int i;
    rfbClientInitMsg ci;
#endif

#if CAN_SEND
    if (!ReadFromRFBServer(pv, sz_rfbProtocolVersionMsg)) return SDL_FALSE;
#else
    strncpy(pv, "RFB 003.003\n", sz_rfbProtocolVersionMsg);
#endif

    errorMessageOnReadFailure = SDL_TRUE;
    pv[sz_rfbProtocolVersionMsg] = 0;

    if (sscanf(pv,rfbProtocolVersionFormat,&major,&minor) != 2) {
        DebugPrint("Not a valid VNC server\n");
        return SDL_FALSE;
    }

    DebugPrint("VNC server supports protocol version %d.%d (viewer %d.%d)\n",
            major, minor, rfbProtocolMajorVersion, rfbProtocolMinorVersion);

    major = rfbProtocolMajorVersion;
    minor = rfbProtocolMinorVersion;

    sprintf(pv,rfbProtocolVersionFormat,major,minor);

#if CAN_SEND
    if (!WriteToRFBServer( pv, sz_rfbProtocolVersionMsg)) return SDL_FALSE;
    if (!ReadFromRFBServer((char *)&authScheme, 4)) return SDL_FALSE;
#endif

#if CAN_SEND
    ci.shared = (appData.shareDesktop ? 1 : 0);

    if (!WriteToRFBServer( (char *)&ci, sz_rfbClientInitMsg)) return SDL_FALSE;
    if (!ReadFromRFBServer((char *)&si, sz_rfbServerInitMsg)) return SDL_FALSE;
    si.framebufferWidth = Swap16IfLE(si.framebufferWidth);
    si.framebufferHeight = Swap16IfLE(si.framebufferHeight);
    si.format.redMax = Swap16IfLE(si.format.redMax);
    si.format.greenMax = Swap16IfLE(si.format.greenMax);
    si.format.blueMax = Swap16IfLE(si.format.blueMax);
    si.nameLength = Swap32IfLE(si.nameLength);
#else
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    si.format.bigEndian = SDL_TRUE;
#else
    si.format.bigEndian = SDL_FALSE;
#endif
    si.nameLength = 5;
#endif

    desktopName = malloc(si.nameLength + 1);
    if (!desktopName) {
        DebugPrint( "Error allocating memory for desktop name, %lu bytes\n",
                (unsigned long)si.nameLength);
        return SDL_FALSE;
    }

#if CAN_SEND
    if (!ReadFromRFBServer(desktopName, si.nameLength)) return SDL_FALSE;
#else
    strncpy(desktopName, "VNes", si.nameLength);
#endif
    desktopName[si.nameLength] = 0;

    DebugPrint("Desktop name \"%s\"\n",desktopName);

    DebugPrint("Connected to VNC server, using protocol version %d.%d\n",
            rfbProtocolMajorVersion, rfbProtocolMinorVersion);

    DebugPrint("VNC server default format:\n");
    PrintPixelFormat(&si.format);

    return SDL_TRUE;
}


/*
 * SetFormatAndEncodings.
 */

SDL_bool
SetFormatAndEncodings()
{
    rfbSetPixelFormatMsg spf;
    char buf[sz_rfbSetEncodingsMsg + MAX_ENCODINGS * 4];
    rfbSetEncodingsMsg *se = (rfbSetEncodingsMsg *)buf;
    Uint32 *encs = (Uint32 *)(&buf[sz_rfbSetEncodingsMsg]);
    int len = 0;
    SDL_bool requestCompressLevel = SDL_FALSE;
    SDL_bool requestQualityLevel = SDL_FALSE;
    SDL_bool requestLastRectEncoding = SDL_FALSE;

    spf.type = rfbSetPixelFormat;
    spf.format = myFormat;
    spf.format.redMax = Swap16IfLE(spf.format.redMax);
    spf.format.greenMax = Swap16IfLE(spf.format.greenMax);
    spf.format.blueMax = Swap16IfLE(spf.format.blueMax);

#if CAN_SEND
    if (!WriteToRFBServer( (char *)&spf, sz_rfbSetPixelFormatMsg))
        return SDL_FALSE;
#endif // CAN_SEND

    se->type = rfbSetEncodings;
    se->nEncodings = 0;

    if (appData.encodingsString) {
        char *encStr = appData.encodingsString;
        int encStrLen;
        do {
            char *nextEncStr = strchr(encStr, ' ');
            if (nextEncStr) {
                encStrLen = nextEncStr - encStr;
                nextEncStr++;
            } else {
                encStrLen = strlen(encStr);
            }

            if (strncasecmp(encStr,"raw",encStrLen) == 0) {
                encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);
            } else if (strncasecmp(encStr,"copyrect",encStrLen) == 0) {
                encs[se->nEncodings++] = Swap32IfLE(rfbEncodingCopyRect);
            } else if (strncasecmp(encStr,"tight",encStrLen) == 0) {
                encs[se->nEncodings++] = Swap32IfLE(rfbEncodingTight);
                requestLastRectEncoding = SDL_TRUE;
                if (appData.compressLevel >= 0 && appData.compressLevel <= 9)
                    requestCompressLevel = SDL_TRUE;
                if (appData.qualityLevel >= 0 && appData.qualityLevel <= 9)
                    requestQualityLevel = SDL_TRUE;
            } else if (strncasecmp(encStr,"hextile",encStrLen) == 0) {
                encs[se->nEncodings++] = Swap32IfLE(rfbEncodingHextile);
            } else if (strncasecmp(encStr,"zlib",encStrLen) == 0) {
                encs[se->nEncodings++] = Swap32IfLE(rfbEncodingZlib);
                if (appData.compressLevel >= 0 && appData.compressLevel <= 9)
                    requestCompressLevel = SDL_TRUE;
            } else if (strncasecmp(encStr,"corre",encStrLen) == 0) {
                encs[se->nEncodings++] = Swap32IfLE(rfbEncodingCoRRE);
            } else if (strncasecmp(encStr,"rre",encStrLen) == 0) {
                encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRRE);
            } else {
                DebugPrint("Unknown encoding '%.*s'\n",encStrLen,encStr);
            }

            encStr = nextEncStr;
        } while (encStr && se->nEncodings < MAX_ENCODINGS);

        if (se->nEncodings < MAX_ENCODINGS && requestCompressLevel) {
            encs[se->nEncodings++] = Swap32IfLE(appData.compressLevel +
                                                rfbEncodingCompressLevel0);
        }

        if (se->nEncodings < MAX_ENCODINGS && requestQualityLevel) {
            encs[se->nEncodings++] = Swap32IfLE(appData.qualityLevel +
                                                rfbEncodingQualityLevel0);
        }

        if (appData.useRemoteCursor) {
            if (se->nEncodings < MAX_ENCODINGS)
                encs[se->nEncodings++] = Swap32IfLE(rfbEncodingXCursor);
            if (se->nEncodings < MAX_ENCODINGS)
                encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRichCursor);
        }

        if (se->nEncodings < MAX_ENCODINGS && requestLastRectEncoding) {
            encs[se->nEncodings++] = Swap32IfLE(rfbEncodingLastRect);
        }
    } else {
        if (SameMachine(rfbsockTCP)) {
            DebugPrint("Same machine: preferring raw encoding\n");
            encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRaw);
        }

        encs[se->nEncodings++] = Swap32IfLE(rfbEncodingCopyRect);
        encs[se->nEncodings++] = Swap32IfLE(rfbEncodingTight);
        encs[se->nEncodings++] = Swap32IfLE(rfbEncodingHextile);
        encs[se->nEncodings++] = Swap32IfLE(rfbEncodingZlib);
        encs[se->nEncodings++] = Swap32IfLE(rfbEncodingCoRRE);
        encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRRE);

        if (appData.compressLevel >= 0 && appData.compressLevel <= 9) {
            encs[se->nEncodings++] = Swap32IfLE(appData.compressLevel +
                                                rfbEncodingCompressLevel0);
        } else {
            encs[se->nEncodings++] = Swap32IfLE(rfbEncodingCompressLevel1);
        }

        if (appData.qualityLevel >= 0 && appData.qualityLevel <= 9) {
            encs[se->nEncodings++] = Swap32IfLE(appData.qualityLevel +
                                                rfbEncodingQualityLevel0);
        }

        if (appData.useRemoteCursor) {
            encs[se->nEncodings++] = Swap32IfLE(rfbEncodingXCursor);
            encs[se->nEncodings++] = Swap32IfLE(rfbEncodingRichCursor);
        }

        encs[se->nEncodings++] = Swap32IfLE(rfbEncodingLastRect);
    }

    len = sz_rfbSetEncodingsMsg + se->nEncodings * 4;

    se->nEncodings = Swap16IfLE(se->nEncodings);

#if CAN_SEND
    if (!WriteToRFBServer( buf, len)) return SDL_FALSE;
#endif // CAN_SEND

    return SDL_TRUE;
}


/*
 * SendIncrementalFramebufferUpdateRequest.
 */

SDL_bool
SendIncrementalFramebufferUpdateRequest()
{
    return SendFramebufferUpdateRequest(0, 0, si.framebufferWidth,
                                        si.framebufferHeight, SDL_TRUE);
}


/*
 * SendFramebufferUpdateRequest.
 */

SDL_bool
SendFramebufferUpdateRequest(int x, int y, int w, int h, SDL_bool incremental)
{
#if CAN_SEND   // Don't send any updates for VNeS
    rfbFramebufferUpdateRequestMsg fur;

    fur.type = rfbFramebufferUpdateRequest;
    fur.incremental = incremental ? 1 : 0;
    fur.x = Swap16IfLE(x);
    fur.y = Swap16IfLE(y);
    fur.w = Swap16IfLE(w);
    fur.h = Swap16IfLE(h);

    if (!WriteToRFBServer( (char *)&fur, sz_rfbFramebufferUpdateRequestMsg))
        return SDL_FALSE;
#else

    return SDL_TRUE;

#endif
}


/*
 * SendPointerEvent.
 */

SDL_bool
SendPointerEvent(int x, int y, int buttonMask)
{
#if CAN_SEND
    rfbPointerEventMsg pe;

    pe.type = rfbPointerEvent;
    pe.buttonMask = buttonMask;
    if (x < 0) x = 0;
    if (y < 0) y = 0;
    SoftCursorMove(x, y);
    pe.x = Swap16IfLE(x);
    pe.y = Swap16IfLE(y);
    return WriteToRFBServer( (char *)&pe, sz_rfbPointerEventMsg);
#else

    return SDL_TRUE;

#endif // CAN_SEND
}


/*
 * SendKeyEvent.
 */

SDL_bool
SendKeyEvent(Uint32 key, SDL_bool down)
{
#if CAN_SEND
    rfbKeyEventMsg ke;

    ke.type = rfbKeyEvent;
    ke.down = down ? 1 : 0;
    ke.key = Swap32IfLE(key);
    return WriteToRFBServer( (char *)&ke, sz_rfbKeyEventMsg);

#else
    /* Pass key events to the controlling daemon */
    if ((KEY_EVENT_TYPE == SDL_KEYDOWN && down) ||
        (KEY_EVENT_TYPE == SDL_KEYUP  && !down)) {

        vnes_msg_t vnesmsg;
        memset(&vnesmsg, 0, sizeof(vnesmsg));        
        vnesmsg.type = VNES_MSG_KEY_PRESSED;
        vnesmsg.msgdata.keypress.keyid = key;

        SendDomainMessage(&vnesmsg);
    }
    return SDL_TRUE;
#endif
}


/*
 * SendClientCutText.
 */

SDL_bool
SendClientCutText(char *str, int len)
{
#if CAN_SEND
    rfbClientCutTextMsg cct;

    if (serverCutText)
        free(serverCutText);
    serverCutText = NULL;

    cct.type = rfbClientCutText;
    cct.length = Swap32IfLE(len);
    return  (WriteToRFBServer( (char *)&cct, sz_rfbClientCutTextMsg) &&
             WriteToRFBServer( str, len));
#else
    return SDL_TRUE;
#endif
}



/*
 * HandleRFBServerMessage.
 */

SDL_bool
HandleRFBServerMessage()
{
    int ret;
    rfbServerToClientMsg msg;

    /* reset the zoom rectangle flag */
    appData.useZoomRect = SDL_FALSE;

    ret = ReadVnesHeader();
    if (ret <= 0) {
        if ( ret < 0 ) {
            return SDL_FALSE;
        } else {
            return SDL_TRUE;
        }
    }

    if (!ReadFromRFBServer((char *)&msg, 1))
        return SDL_FALSE;

    switch (msg.type) {

    case rfbSetColourMapEntries:
        {
            int i;
            Uint16 rgb[3];
            SDL_Color color;

            if (!ReadFromRFBServer(((char *)&msg) + 1,
                                   sz_rfbSetColourMapEntriesMsg - 1))
                return SDL_FALSE;

            msg.scme.firstColour = Swap16IfLE(msg.scme.firstColour);
            msg.scme.nColours = Swap16IfLE(msg.scme.nColours);

            for (i = 0; i < msg.scme.nColours; i++) {
                if (!ReadFromRFBServer((char *)rgb, 6))
                    return SDL_FALSE;

                color.r = Swap16IfLE(rgb[0]) >> 8;
                color.g = Swap16IfLE(rgb[1]) >> 8;
                color.b = Swap16IfLE(rgb[2]) >> 8;
                SDL_SetColors(ovl, &color, msg.scme.firstColour+i, 1);
            }

            break;
        }

    case rfbFramebufferUpdate:
        {
            rfbFramebufferUpdateRectHeader rect;
            int linesToRead;
            int bytesPerLine;
            int i;
            SDL_Rect updateRect = {-1, -1, 0, 0};

            if (!ReadFromRFBServer(((char *)&msg.fu) + 1,
                                   sz_rfbFramebufferUpdateMsg - 1))
                return SDL_FALSE;

            msg.fu.nRects = Swap16IfLE(msg.fu.nRects);

            for (i = 0; i < msg.fu.nRects; i++) {
                if (!ReadFromRFBServer((char *)&rect, sz_rfbFramebufferUpdateRectHeader))
                    return SDL_FALSE;

                rect.encoding = Swap32IfLE(rect.encoding);
                if (rect.encoding == rfbEncodingLastRect)
                    break;

                rect.r.x = Swap16IfLE(rect.r.x);
                rect.r.y = Swap16IfLE(rect.r.y);
                rect.r.w = Swap16IfLE(rect.r.w);
                rect.r.h = Swap16IfLE(rect.r.h);

                // Merge the update area rectangles
                if ( updateRect.x == -1 ) {
                    updateRect.x = rect.r.x;
                    updateRect.y = rect.r.y;
                    updateRect.w = rect.r.w;
                    updateRect.h = rect.r.h;
                } else {
                    if ( rect.r.x < updateRect.x ) {
                        updateRect.w += updateRect.x - rect.r.x;
                        updateRect.x = rect.r.x;
                    }
                    if ( rect.r.y < updateRect.y ) {
                        updateRect.h += updateRect.y - rect.r.y;
                        updateRect.y = rect.r.y;
                    }
                    if ((rect.r.w + rect.r.x) > (updateRect.w + updateRect.x))
                        updateRect.w = rect.r.w + rect.r.x - updateRect.x;
                    if ((rect.r.h + rect.r.y) > (updateRect.h + updateRect.y))
                        updateRect.h = rect.r.h + rect.r.y - updateRect.y;
                }

                if (rect.encoding == rfbEncodingXCursor) {
                    if (!HandleXCursor(rect.r.x, rect.r.y, rect.r.w, rect.r.h)) {
                        return SDL_FALSE;
                    }
                    continue;
                }
                if (rect.encoding == rfbEncodingRichCursor) {
                    if (!HandleRichCursor(rect.r.x, rect.r.y, rect.r.w, rect.r.h)) {
                        return SDL_FALSE;
                    }
                    continue;
                }

                if ((rect.r.x + rect.r.w > si.framebufferWidth) ||
                    (rect.r.y + rect.r.h > si.framebufferHeight))
                {
                    DebugPrint("Rect too large: %dx%d at (%d, %d)\n",
                            rect.r.w, rect.r.h, rect.r.x, rect.r.y);
                    return SDL_FALSE;
                }

                if ((rect.r.h * rect.r.w) == 0) {
                    DebugPrint("WARNING: Zero area rect (%d x %d)\n",
                            rect.r.w, rect.r.h);
                }

                /* If RichCursor encoding is used, we should prevent collisions
                   between framebuffer updates and cursor drawing operations. */
                SoftCursorLockArea(rect.r.x, rect.r.y, rect.r.w, rect.r.h);

                switch (rect.encoding) {

                case rfbEncodingRaw:

                    bytesPerLine = rect.r.w * myFormat.bitsPerPixel / 8;
                    linesToRead = BUFFER_SIZE / bytesPerLine;

                    while (rect.r.h > 0) {
                        if (linesToRead > rect.r.h)
                            linesToRead = rect.r.h;

                        if (!ReadFromRFBServer(buffer,bytesPerLine * linesToRead))
                            return SDL_FALSE;

                        CopyDataToScreen(buffer, rect.r.x, rect.r.y, rect.r.w,
                                         linesToRead);

                        rect.r.h -= linesToRead;
                        rect.r.y += linesToRead;

                    }
                    break;

                case rfbEncodingCopyRect:
                    {
                        SDL_Rect sRect = {0, 0, rect.r.w, rect.r.h};
                        SDL_Rect dRect = {rect.r.x, rect.r.y, rect.r.w, rect.r.h};
                        rfbCopyRect cr;

                        if (!ReadFromRFBServer((char *)&cr, sz_rfbCopyRect))
                            return SDL_FALSE;

                        sRect.x = Swap16IfLE(cr.srcX);
                        sRect.y = Swap16IfLE(cr.srcY);

                        /* If RichCursor encoding is used, we should extend our
                           "cursor lock area" (previously set to destination
                           rectangle) to the source rectangle as well. */
                        SoftCursorLockArea(cr.srcX, cr.srcY, rect.r.w, rect.r.h);

                        if (appData.copyRectDelay != 0) {
                            usleep(appData.copyRectDelay * 1000);
                        }

                        SDL_BlitSurface(ovl, &sRect, ovl, &dRect);
                        break;
                    }

                case rfbEncodingRRE:
                    {
                        switch (myFormat.bitsPerPixel) {
                        case 8:
                            if (!HandleRRE8(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        case 16:
                            if (!HandleRRE16(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        case 32:
                            if (!HandleRRE32(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        }
                        break;
                    }

                case rfbEncodingCoRRE:
                    {
                        switch (myFormat.bitsPerPixel) {
                        case 8:
                            if (!HandleCoRRE8(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        case 16:
                            if (!HandleCoRRE16(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        case 32:
                            if (!HandleCoRRE32(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        }
                        break;
                    }

                case rfbEncodingHextile:
                    {
                        switch (myFormat.bitsPerPixel) {
                        case 8:
                            if (!HandleHextile8(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        case 16:
                            if (!HandleHextile16(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        case 32:
                            if (!HandleHextile32(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        }
                        break;
                    }

                case rfbEncodingZlib:
                    {
                        switch (myFormat.bitsPerPixel) {
                        case 8:
                            if (!HandleZlib8(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        case 16:
                            if (!HandleZlib16(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        case 32:
                            if (!HandleZlib32(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        }
                        break;
                    }

                case rfbEncodingTight:
                    {
                        switch (myFormat.bitsPerPixel) {
                        case 8:
                            if (!HandleTight8(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        case 16:
                            if (!HandleTight16(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        case 32:
                            if (!HandleTight32(rect.r.x,rect.r.y,rect.r.w,rect.r.h))
                                return SDL_FALSE;
                            break;
                        }
                        break;
                    }

                default:
                    DebugPrint("Unknown rect encoding %d\n",
                            (int)rect.encoding);
                    return SDL_FALSE;
                }

                /* Now we may discard "soft cursor locks". */
                SoftCursorUnlockScreen();
            }

            // Update the screen (only the merged update areas however)
            UpdateOverlayRect(ovl, 
							  updateRect.x, updateRect.y,
							  updateRect.w, updateRect.h);

#ifdef mips
            // Zoom in on the update area if the zoom rect flag
            // command was passed
            if ( appData.useZoomRect ) {
#ifdef DEBUG
				fprintf(stderr, "Update rect: %4dx%4d+%4d+%4d\tZoom rect: %4dx%4d+%4d+%4d\n",
						updateRect.w, updateRect.h,
						updateRect.x, updateRect.y,
						appData.zoomRect.w, appData.zoomRect.h,
						appData.zoomRect.x, appData.zoomRect.y );
#endif
                ZoomOverlay(&appData.zoomRect);
            } else {
#ifdef DEBUG
				fprintf(stderr, "Update rect: %4dx%4d+%4d+%4d\n",
						updateRect.w, updateRect.h,
						updateRect.x, updateRect.y );
#endif
				/* zoom back to fullscreen */
                ZoomOverlay(NULL);
            }
#endif

            if (!SendIncrementalFramebufferUpdateRequest())
                return SDL_FALSE;

            break;
        }

    case rfbBell:
        {
            /* TODO: Implement bell somehow in SDL */
            break;
        }

    case rfbServerCutText:
        {
            if (!ReadFromRFBServer(((char *)&msg) + 1,
                                   sz_rfbServerCutTextMsg - 1))
                return SDL_FALSE;

            msg.sct.length = Swap32IfLE(msg.sct.length);

            if (serverCutText)
                free(serverCutText);

            serverCutText = malloc(msg.sct.length+1);

            if (!ReadFromRFBServer(serverCutText, msg.sct.length))
                return SDL_FALSE;

            serverCutText[msg.sct.length] = 0;

            newServerCutText = SDL_TRUE;

            break;
        }

    default:
        DebugPrint("Unknown message type %d from VNC server\n",msg.type);
        return SDL_FALSE;
    }

    return SDL_TRUE;
}


#define GET_PIXEL8(pix, ptr) ((pix) = *(ptr)++)

#define GET_PIXEL16(pix, ptr) (((Uint8*)&(pix))[0] = *(ptr)++, \
			       ((Uint8*)&(pix))[1] = *(ptr)++)

#define GET_PIXEL32(pix, ptr) (((Uint8*)&(pix))[0] = *(ptr)++, \
			       ((Uint8*)&(pix))[1] = *(ptr)++, \
			       ((Uint8*)&(pix))[2] = *(ptr)++, \
			       ((Uint8*)&(pix))[3] = *(ptr)++)

/* CONCAT2 concatenates its two arguments.  CONCAT2E does the same but also
   expands its arguments if they are macros */

#define CONCAT2(a,b) a##b
#define CONCAT2E(a,b) CONCAT2(a,b)

#define BPP 8
#include "rre.c"
#include "corre.c"
#include "hextile.c"
#include "zlib.c"
#include "tight.c"
#undef BPP
#define BPP 16
#include "rre.c"
#include "corre.c"
#include "hextile.c"
#include "zlib.c"
#include "tight.c"
#undef BPP
#define BPP 32
#include "rre.c"
#include "corre.c"
#include "hextile.c"
#include "zlib.c"
#include "tight.c"
#undef BPP


/*
 * PrintPixelFormat.
 */

void
PrintPixelFormat(format)
     rfbPixelFormat *format;
{
    if (format->bitsPerPixel == 1) {
        DebugPrint("  Single bit per pixel.\n");
        DebugPrint(
                "  %s significant bit in each byte is leftmost on the screen.\n",
                (format->bigEndian ? "Most" : "Least"));
    } else {
        DebugPrint("  %d bits per pixel.\n",format->bitsPerPixel);
        if (format->bitsPerPixel != 8) {
            DebugPrint("  %s significant byte first in each pixel.\n",
                    (format->bigEndian ? "Most" : "Least"));
        }
        if (format->trueColour) {
            DebugPrint("  SDL_TRUE colour: max red %d green %d blue %d\n",
                    format->redMax, format->greenMax, format->blueMax);
            DebugPrint(", shift red %d green %d blue %d\n",
                    format->redShift, format->greenShift, format->blueShift);
        } else {
            DebugPrint("  Colour map (not true colour).\n");
        }
    }
}

static long
ReadCompactLen (void)
{
    long len;
    Uint8 b;

    if (!ReadFromRFBServer((char *)&b, 1))
        return -1;
    len = (int)b & 0x7F;
    if (b & 0x80) {
        if (!ReadFromRFBServer((char *)&b, 1))
            return -1;
        len |= ((int)b & 0x7F) << 7;
        if (b & 0x80) {
            if (!ReadFromRFBServer((char *)&b, 1))
                return -1;
            len |= ((int)b & 0xFF) << 14;
        }
    }
    return len;
}

/*
 * JPEG source manager functions for JPEG decompression in Tight decoder.
 */

static struct jpeg_source_mgr jpegSrcManager;
static JOCTET *jpegBufferPtr;
static size_t jpegBufferLen;

static void
JpegInitSource(j_decompress_ptr cinfo)
{
    jpegError = SDL_FALSE;
}

static boolean
JpegFillInputBuffer(j_decompress_ptr cinfo)
{
    jpegError = SDL_TRUE;
    jpegSrcManager.bytes_in_buffer = jpegBufferLen;
    jpegSrcManager.next_input_byte = (JOCTET *)jpegBufferPtr;

    return TRUE;
}

static void
JpegSkipInputData(j_decompress_ptr cinfo, long num_bytes)
{
    if (num_bytes < 0 || num_bytes > jpegSrcManager.bytes_in_buffer) {
        jpegError = SDL_TRUE;
        jpegSrcManager.bytes_in_buffer = jpegBufferLen;
        jpegSrcManager.next_input_byte = (JOCTET *)jpegBufferPtr;
    } else {
        jpegSrcManager.next_input_byte += (size_t) num_bytes;
        jpegSrcManager.bytes_in_buffer -= (size_t) num_bytes;
    }
}

static void
JpegTermSource(j_decompress_ptr cinfo)
{
    /* No work necessary here. */
}

static void
JpegSetSrcManager(j_decompress_ptr cinfo, Uint8 *compressedData,
                  int compressedLen)
{
    jpegBufferPtr = (JOCTET *)compressedData;
    jpegBufferLen = (size_t)compressedLen;

    jpegSrcManager.init_source = JpegInitSource;
    jpegSrcManager.fill_input_buffer = JpegFillInputBuffer;
    jpegSrcManager.skip_input_data = JpegSkipInputData;
    jpegSrcManager.resync_to_restart = jpeg_resync_to_restart;
    jpegSrcManager.term_source = JpegTermSource;
    jpegSrcManager.next_input_byte = jpegBufferPtr;
    jpegSrcManager.bytes_in_buffer = jpegBufferLen;

    cinfo->src = &jpegSrcManager;
}
