/*
 *  Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 *
 *  This is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this software; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
 *  USA.
 */

/*
 * sockets.c - functions to deal with sockets.
 */

#include "debug.h"
#include <unistd.h>
#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/time.h>
#include "VNeSCommon.h"
#include "vnes.h"
#include "config.h"
#include "d3des.h"
#include "aes.h"
#include "vncviewer.h"

#ifdef DEBUG
static void PrintInHex(char *buf, int len);
#endif

static SDL_bool WriteExact(int sock, char *buf, int n);
static void DecryptDES(char* buf, unsigned int len);
static void EncryptDES(char* buf, unsigned int len);
static void DecryptAES(char* buf, unsigned int len);
static void EncryptAES(char* buf, unsigned int len);

SDL_bool errorMessageOnReadFailure = SDL_TRUE;

// For encryption (AES requires 16 bytes per block, DES requires 8)
#define DESBLOCK_SIZE	8
#define AESBLOCK_SIZE	16

typedef union {
	int size;
	char pad[DESBLOCK_SIZE];
} des_header_t;

typedef union {
	int size;
	char pad[AESBLOCK_SIZE];
} aes_header_t;


// Buffer for decrypted data that can be passed to back
// to the application
#define ENC_BUF_SIZE (64 * 1024)
static unsigned int g_TotalMessage = 0;             // Number of bytes in the current message
static char         g_decryptedSize = 0;            // Number of bytes in decryption buffer
static char         g_decryptedBuf[ENC_BUF_SIZE];   // Buffer for decrypted characters
static char*        g_decryptedPtr = g_decryptedBuf;// Pointer to decryption buffer


#define SOCK_BUF_SIZE (128 * 1024)                  // Size of the socket buffer
static unsigned int g_socketBufSize = 0;            // Bytes of data in the buffer
static char*        g_socketBufPtr;                 // Pointer to first data byte
static char         g_socketBuffer[SOCK_BUF_SIZE];  // Buffer of socket data

#ifdef BENCHMARK
unsigned int decryptTime;
unsigned int decryptBytes;
unsigned int encryptTime;
unsigned int encryptBytes;
#endif

static void
dataEncrypt( char* buf, unsigned int len )
{
#ifdef BENCHMARK
	struct timeval start, end;

	encryptBytes += len;
	gettimeofday(&start, NULL);
#endif

	switch (appData.useEncryptionType) {
	case VNES_ENC_DES:
		EncryptDES(buf, len);
		break;
	case VNES_ENC_AES:
		EncryptAES(buf, len);
		break;
	default:
		break;
	}

#ifdef BENCHMARK
	gettimeofday(&end, NULL);
	encryptTime += ( end.tv_sec*1000 + end.tv_usec/1000 ) -
		( start.tv_sec*1000 + start.tv_usec/1000 );
#endif
}

static void
dataDecrypt( char* buf, unsigned int len )
{
#ifdef BENCHMARK
	struct timeval start, end;

	decryptBytes += len;
	gettimeofday(&start, NULL);
#endif

	switch (appData.useEncryptionType) {
	case VNES_ENC_DES:
		DecryptDES(buf, len);
		break;
	case VNES_ENC_AES:
		DecryptAES(buf, len);
		break;
	default:
		break;
	}

#ifdef BENCHMARK
	gettimeofday(&end, NULL);
	decryptTime += ( end.tv_sec*1000 + end.tv_usec/1000 ) -
		( start.tv_sec*1000 + start.tv_usec/1000 );
#endif
}

static SDL_bool rfbsockReady = SDL_FALSE;
static void
ProcessEvents()
{
    SDL_Event event;
    struct timeval tm;
    int fd;
    int ret;
    fd_set   fds;

    fd = rfbsockTCP;

	rfbsockReady = SDL_FALSE;
    while (! appData.isFinished && rfbsockReady == SDL_FALSE) {
        if (SDL_PollEvent(&event)) {
            if (event.type == SDL_VIDEOEXPOSE ||
                event.type == SDL_ACTIVEEVENT)
            {
                HandleBasicDesktopEvent(&event);
            } else {
                /* Send the event to the dialog */
                PollDialog(&event);
                SendRFBEvent(&event, NULL, NULL);

                if ( event.type == KEY_EVENT_TYPE ) {
                    SDL_Event repeat;
                    Uint32 starttime = SDL_GetTicks();

                    // Throw out any key events for the next 10 ms
                    while(SDL_PeepEvents(&repeat, 1, SDL_GETEVENT,
                                         (SDL_KEYDOWNMASK|SDL_KEYUPMASK)) > 0)
                    {
                        if ((SDL_GetTicks() - starttime) > 10) {
                            // break after 10 millisec of events to
                            // avoid network starvation via key input
                            break;
                        }
                    }
                }
            }
        } else {
			/* Poll the dialog to update its status */
			PollDialog(NULL);
		}

        // if there is any data left in the read buffer, process it as
        // another rfbUpdate message
        if ( g_socketBufSize > 0 ) {
            rfbsockReady = SDL_TRUE;
            return;
        }

        FD_ZERO(&fds);
        FD_SET(fd,&fds);
        tm.tv_usec = 50000; /* 50 millisec timeout */
        tm.tv_sec = 0;

        ret = select(fd+1, &fds, NULL, NULL, &tm);
        if (ret > 0) {
            rfbsockReady = SDL_TRUE;
        } else if ( ret < 0 ) {
            DebugPrint("Socket error on select\n");
            break;
        }
    }
}


static int
ReadExact( int sock, char* out, unsigned int len )
{
    int totalbytes = 0;
    int i;

    if ( len == 0 ) {
        DebugPrint( "ReadExact: Read of length zero requested\n");
        return 0;
    }

    if ( len <= g_socketBufSize ) {
        memcpy(out, g_socketBufPtr, len);
        g_socketBufSize -= len;
        g_socketBufPtr += len;

#ifdef DEBUG
        DebugPrint( "ReadExact: Buffered TCP (%d / %d)\n",
                    len, g_socketBufSize );
#endif

        return len;

    } else {
        if ( g_socketBufSize > 0 ) {
            memcpy(out, g_socketBufPtr, g_socketBufSize);
            out += g_socketBufSize;
            totalbytes += g_socketBufSize;
            len -= g_socketBufSize;

#ifdef DEBUG
            DebugPrint( "ReadExact: Emptied buffer of %d bytes\n",
                        g_socketBufSize );
#endif
        }
    }

    g_socketBufSize = 0;
    g_socketBufPtr  = g_socketBuffer;

    if ( len < SOCK_BUF_SIZE ) {
        // Starting TCP buffered-read loop
        while (g_socketBufSize < len) {
            i = read(sock,
                     g_socketBuffer + g_socketBufSize,
                     SOCK_BUF_SIZE - g_socketBufSize);
            if (i <= 0) {
                if (i < 0) {
                    if (errno == EWOULDBLOCK || errno == EAGAIN) {
                        ProcessEvents();
                        i = 0;
                    } else {
                        DebugPrint("%s: %s", programName, sys_errlist[errno]);
                        Cleanup(ERROR_SOCKET);
                    }
                } else {
                    if (errorMessageOnReadFailure) {
                        DebugPrint("%s: VNC server closed connection\n",programName);
                    }
                    Cleanup(ERROR_SOCKET);
                }
            } else {
                g_socketBufSize += i;
            }
        }

        // Copy the first part of the recieved data and update the
        // pointer/size of the remaining buffer data
        memcpy(out, g_socketBuffer, len);
        g_socketBufSize -= len;
        g_socketBufPtr += len;
        totalbytes += len;

#ifdef DEBUG
        DebugPrint( "ReadExact: Buffered TCP Read (%d / %d) bytes\n",
                    totalbytes, g_socketBufSize);
#endif

    } else {
        unsigned int count = 0;

        // Starting TCP direct-read loop
        while (count < len) {
            i = read(sock, out + count, len - count);
            if (i <= 0) {
                if (i < 0) {
                    if (errno == EWOULDBLOCK || errno == EAGAIN) {
                        ProcessEvents();
                        i = 0;
                    } else {
                        DebugPrint("%s: %s", programName, sys_errlist[errno]);
                        Cleanup(ERROR_SOCKET);
                    }
                } else {
                    if (errorMessageOnReadFailure) {
                        DebugPrint("%s: VNC server closed connection\n",programName);
                    }
                    Cleanup(ERROR_SOCKET);
                }
            } else {
                count += i;
            }
        }

        totalbytes += count;

#ifdef DEBUG
        DebugPrint( "ReadExact: Direct TCP Read %d bytes\n", totalbytes);
#endif
    }


	return totalbytes;
}

SDL_bool
ReadDecryptHeader( int sock )
{
	int count;

	assert( g_TotalMessage == 0 );

	// We are expecting the 8 byte DES encrypted size header
	if (appData.useEncryptionType == VNES_ENC_DES) {
		des_header_t deshdr;

		count = ReadExact( sock, (char*)&deshdr, DESBLOCK_SIZE );
		if ( count == -1 ) {
			return SDL_FALSE;
		}

		DebugPrint("DES_TEST count: %d\n", count);
		DebugPrint("DES_TEST input des_hdr: %hhx %hhx %hhx %hhx %hhx %hhx %hhx %hhx\n", 
				   deshdr.pad[0], deshdr.pad[1], deshdr.pad[2], deshdr.pad[3],
				   deshdr.pad[4], deshdr.pad[5], deshdr.pad[6], deshdr.pad[7]);

		// Decrypt data
		dataDecrypt( (char*)&deshdr, count );
		g_TotalMessage = Swap32IfLE(deshdr.size);

		DebugPrint("DES_TEST output des_hdr: %hhx %hhx %hhx %hhx %hhx %hhx %hhx %hhx\n", 
				   deshdr.pad[0], deshdr.pad[1], deshdr.pad[2], deshdr.pad[3],
				   deshdr.pad[4], deshdr.pad[5], deshdr.pad[6], deshdr.pad[7]);

		DebugPrint( "Decoded DES data size: %u\n", g_TotalMessage );
	} else {
		aes_header_t aeshdr;

		count = ReadExact( sock, (char*)&aeshdr, AESBLOCK_SIZE );	
		if ( count == -1 ) {
			return SDL_FALSE;
		}

		DebugPrint("AES_TEST count: %d\n", count);
		DebugPrint("AES_TEST input aes_hdr: %hhx %hhx %hhx %hhx %hhx %hhx %hhx %hhx\n", 
				   aeshdr.pad[0], aeshdr.pad[1], aeshdr.pad[2], aeshdr.pad[3],
				   aeshdr.pad[4], aeshdr.pad[5], aeshdr.pad[6], aeshdr.pad[7]);

		// Decrypt data
		dataDecrypt( (char*)&aeshdr, count );
		g_TotalMessage = Swap32IfLE(aeshdr.size);

		DebugPrint("AES_TEST output aes_hdr: %hhx %hhx %hhx %hhx %hhx %hhx %hhx %hhx\n", 
				   aeshdr.pad[0], aeshdr.pad[1], aeshdr.pad[2], aeshdr.pad[3],
				   aeshdr.pad[4], aeshdr.pad[5], aeshdr.pad[6], aeshdr.pad[7]);

		DebugPrint( "Decoded AES header size: %u\n", g_TotalMessage );
	}

	assert( g_TotalMessage > 0 );
	return SDL_TRUE;
}

SDL_bool
WriteEncryptHeader( int sock, unsigned int len )
{
	assert( len > 0 );

	// We are expecting the 8 byte DES encrypted size header
	if (appData.useEncryptionType == VNES_ENC_DES) {
		des_header_t deshdr;

		// Encrypt data
		deshdr.size = Swap32IfLE(len);
		dataEncrypt( (char*)&deshdr, sizeof(deshdr) );

#ifdef DEBUG
		DebugPrint( "Sent DES header size: %d\n", len );
#endif

		// We are expecting the 8 byte DES encrypted size header
		return WriteExact(sock, (char*)&deshdr, sizeof(deshdr));
	} else {
		aes_header_t aeshdr;

		// Encrypt data
		aeshdr.size = Swap32IfLE(len);
		dataEncrypt( (char*)&aeshdr, sizeof(aeshdr) );

#ifdef DEBUG
		DebugPrint( "Sent AES header size: %d\n", len );
#endif

		// We are expecting the 16 byte AES encrypted size header
		return WriteExact(sock, (char*)&aeshdr, sizeof(aeshdr));
	}
}

int
ReadDecryptBody( int sock, char *out, unsigned int n )
{
	int total = 0;
	int count;
	int size;

	assert( g_TotalMessage > 0 );
	assert( n <= g_TotalMessage );
	assert( g_decryptedSize == 0 );

	// Round n down to nearest multiple of appData.encBlockSize
	size = (n / appData.encBlockSize) * appData.encBlockSize;
	if ( size > 0 ) {
		count = ReadExact( sock, out, size );
		if ( count == -1 ) {
			return -1;
		}

		// Decrypt data
		dataDecrypt( out, count );
		out += count;
		total += count;

		assert( g_TotalMessage >= count );
		g_TotalMessage -= count;
	}

	if ( (n % appData.encBlockSize) > 0 ) {
		int remainder = n % appData.encBlockSize;

		// Read a partial block of data
		count = ReadExact( sock, g_decryptedBuf, appData.encBlockSize );
		if ( count == -1 ) {
			return -1;
		}

		// Decrypt data
		dataDecrypt( g_decryptedBuf, count );
		memcpy(out, g_decryptedBuf, remainder);
		total += remainder;

		assert( g_TotalMessage >= remainder );
		g_TotalMessage -= remainder;

		if ( g_TotalMessage > 0 ) {
			g_decryptedPtr = g_decryptedBuf + remainder;
			g_decryptedSize = appData.encBlockSize - remainder;
			if ( g_decryptedSize > g_TotalMessage ) {
				g_decryptedSize = g_TotalMessage;
			}
			g_TotalMessage -= g_decryptedSize;
		}
	}

	return total;
}

SDL_bool
DecryptFromRFBServer( int sock, char *out, unsigned int n )
{
	int ret;

	// The are enough decrypted bytes already
	if ( n <= g_decryptedSize ) {
		memcpy( out, g_decryptedPtr, n );
		g_decryptedSize -= n;
		g_decryptedPtr += n;
		return SDL_TRUE;
	}

	if ( g_decryptedSize > 0 ) {
		// Empty any remaining partially-decrypted blocks
		memcpy( out, g_decryptedPtr, g_decryptedSize );
		n -= g_decryptedSize;
		out += g_decryptedSize;
		g_decryptedPtr = g_decryptedBuf;
		g_decryptedSize = 0;
	}

	// Update the g_TotalMessage variable if necessary
	if ( g_TotalMessage == 0 ) {
		if ( ReadDecryptHeader(sock) == SDL_FALSE ) {
			return SDL_FALSE;
		}
	}

    // Repeat until there is sufficient data available
    // in a single packet to fulfill 'n' bytes
	while (n > g_TotalMessage) {
		// Read a message body
		ret = ReadDecryptBody( sock, out, g_TotalMessage );
		if ( ret == -1 ) {
			return SDL_FALSE;
		}
		n -= ret;
		out += ret;

		// Update the g_TotalMessage variable if necessary
		if ( g_TotalMessage == 0 ) {
			if ( ReadDecryptHeader(sock) == SDL_FALSE ) {
				return SDL_FALSE;
			}
		}
	}

    // Read remaining required data
	if ( n > 0 ) {
		// Read a message body
		return (ReadDecryptBody( sock, out, n ) != -1);
	}

	return SDL_TRUE;
}


/*
 * ReadFromRFBServer is called whenever we want to read some data from the RFB
 * server.  It is non-trivial for two reasons:
 *
 * 1. For efficiency it performs some intelligent buffering, avoiding invoking
 *    the read() system call too often.  For small chunks of data, it simply
 *    copies the data out of an internal buffer.  For large amounts of data it
 *    reads directly into the buffer provided by the caller.
 *
 * 2. Whenever read() would block, it invokes the Xt event dispatching
 *    mechanism to process X events.  In fact, this is the only place these
 *    events are processed, as there is no XtAppMainLoop in the program.
 */
SDL_bool
ReadFromRFBServer( char *out, unsigned int n)
{
	int sock = rfbsockTCP;

	if (appData.useEncryptionType) {
		return DecryptFromRFBServer( sock, out, n );
	} else {
		return ( ReadExact( sock, out, n ) == n );
	}
}


/*
 * Write an exact number of bytes, and don't return until you've sent them.
 */

SDL_bool
WriteExact(int sock, char *buf, int n)
{
	fd_set fds;
	int i = 0;
	int j;

	while (i < n) {
		j = write(sock, buf + i, (n - i));
		if (j <= 0) {
			if (j < 0) {
				if (errno == EWOULDBLOCK || errno == EAGAIN) {
					FD_ZERO(&fds);
					FD_SET(rfbsockTCP,&fds);

					if (select(rfbsockTCP+1, NULL, &fds, NULL, NULL) <= 0) {
                        DebugPrint("%s: %s", programName, sys_errlist[errno]);
						return SDL_FALSE;
					}
					j = 0;
				} else {
					DebugPrint("%s: %s", programName, sys_errlist[errno]);
					return SDL_FALSE;
				}
			} else {
				DebugPrint("%s: write failed\n", programName);
				return SDL_FALSE;
			}
		}
		i += j;
	}

#ifdef DEBUG
	DebugPrint("WriteExact: Sent %d bytes\n", i );
#endif

	return SDL_TRUE;
}

SDL_bool
EncryptToRFBServer( int sock, char* out, unsigned int n )
{
	int size;

	assert( n > 0 );

	if ( WriteEncryptHeader( sock, n ) == SDL_FALSE ) {
		return SDL_FALSE;
	}

	size = (n / appData.encBlockSize) * appData.encBlockSize;
	if ( size > 0 ) {
		dataEncrypt( out, size );
		if ( WriteExact( sock, out, size ) == SDL_FALSE ) {
			return SDL_FALSE;
		}
		out += size;
		n -= size;
	}

	if ( (n % appData.encBlockSize) > 0 ) {
		char encrypt[AESBLOCK_SIZE] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};

		memcpy( encrypt, out, (n % appData.encBlockSize) );
		dataEncrypt( encrypt, appData.encBlockSize );

		if ( WriteExact( sock, encrypt, appData.encBlockSize ) == SDL_FALSE ) {
			return SDL_FALSE;
		}
	}

	return SDL_TRUE;
}

SDL_bool
WriteToRFBServer( char *out, unsigned int n)
{
#if CAN_SEND
	if ( g_UseEncryption ) {
		return EncryptToRFBServer( rfbsockTCP, out, n );
	} else {
		return WriteExact( rfbsockTCP, out, n );
	}
#else // CAN_SEND
    return SDL_TRUE;
#endif // CAN_SEND
}


/*
 * ConnectToTcpAddr connects to the given TCP port.
 */

int
ConnectToTcpAddr(unsigned int host, int port)
{
    int sock;
    struct sockaddr_in addr;
    int one = 1;

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = host;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
		DebugPrint("%s: %s", programName, sys_errlist[errno]);
        return -1;
    }

    if (connect(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		DebugPrint("%s: %s", programName, sys_errlist[errno]);
        close(sock);
        return -1;
    }

    if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY,
                   (char *)&one, sizeof(one)) < 0) {
		DebugPrint("%s: %s", programName, sys_errlist[errno]);
        close(sock);
        return -1;
    }

    return sock;
}


/*
 * FindFreeTcpPort tries to find unused TCP port in the range
 * (TUNNEL_PORT_OFFSET, TUNNEL_PORT_OFFSET + 99]. Returns 0 on failure.
 */

int
FindFreeTcpPort(void)
{
    int sock, port;
    struct sockaddr_in addr;

    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = INADDR_ANY;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
		DebugPrint("%s: %s", programName, sys_errlist[errno]);
        return 0;
    }

    for (port = TUNNEL_PORT_OFFSET + 99; port > TUNNEL_PORT_OFFSET; port--) {
        addr.sin_port = htons((unsigned short)port);
        if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) == 0) {
            close(sock);
            return port;
        }
    }

    close(sock);
    return 0;
}


/*
 * ListenAtTcpPort starts listening at the given TCP port.
 */

int
ListenAtTcpPort(int port)
{
    int sock;
    struct sockaddr_in addr;
    int one = 1;

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = INADDR_ANY;

    sock = socket(AF_INET, SOCK_STREAM, 0);
    if (sock < 0) {
		DebugPrint("%s: %s", programName, sys_errlist[errno]);
        return -1;
    }

    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
                   (const char *)&one, sizeof(one)) < 0) {
		DebugPrint("%s: %s", programName, sys_errlist[errno]);
        close(sock);
        return -1;
    }

    if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
		DebugPrint("%s: %s", programName, sys_errlist[errno]);
        close(sock);
        return -1;
    }

    if (listen(sock, 5) < 0) {
		DebugPrint("%s: %s", programName, sys_errlist[errno]);
        close(sock);
        return -1;
    }

    return sock;
}



/*
 * AcceptTcpConnection accepts a TCP connection.
 */

int
AcceptTcpConnection(int listenSock)
{
    int sock;
    struct sockaddr_in addr;
    int addrlen = sizeof(addr);
    int one = 1;

    sock = accept(listenSock, (struct sockaddr *) &addr, &addrlen);
    if (sock < 0) {
		DebugPrint("%s: %s", programName, sys_errlist[errno]);
        return -1;
    }

    if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY,
                   (char *)&one, sizeof(one)) < 0) {
		DebugPrint("%s: %s", programName, sys_errlist[errno]);
        close(sock);
        return -1;
    }

    return sock;
}


/*
 * SetNonBlocking sets a socket into non-blocking mode.
 */

SDL_bool
SetNonBlocking(int sock)
{
    if (fcntl(sock, F_SETFL, O_NONBLOCK) < 0) {
		DebugPrint("%s: %s", programName, sys_errlist[errno]);
        return SDL_FALSE;
    }
    return SDL_TRUE;
}


/*
 * StringToIPAddr - convert a host string to an IP address.
 */

SDL_bool
StringToIPAddr(const char *str, unsigned int *addr)
{
    struct hostent *hp;

    if (strcmp(str,"") == 0) {
        *addr = 0; /* local */
        return SDL_TRUE;
    }

    *addr = inet_addr(str);

    if (*addr != -1)
        return SDL_TRUE;

    hp = gethostbyname(str);

    if (hp) {
        *addr = *(unsigned int *)hp->h_addr;
        return SDL_TRUE;
    }

    return SDL_FALSE;
}


/*
 * Test if the other end of a socket is on the same machine.
 */

SDL_bool
SameMachine(int sock)
{
    struct sockaddr_in peeraddr, myaddr;
    int addrlen = sizeof(struct sockaddr_in);

    getpeername(sock, (struct sockaddr *)&peeraddr, &addrlen);
    getsockname(sock, (struct sockaddr *)&myaddr, &addrlen);

    return (peeraddr.sin_addr.s_addr == myaddr.sin_addr.s_addr);
}


#ifdef DEBUG
/*
 * Print out the contents of a packet for debugging.
 */

void
PrintInHex(char *buf, int len)
{
    int i, j;
    char c, str[17];

    str[16] = 0;

    fprintf(stderr,"ReadExact: ");

    for (i = 0; i < len; i++)
    {
        if ((i % 16 == 0) && (i != 0)) {
            fprintf(stderr,"           ");
        }
        c = buf[i];
        str[i % 16] = (((c > 31) && (c < 127)) ? c : '.');
        fprintf(stderr,"%02x ",(unsigned char)c);
        if ((i % 4) == 3)
            fprintf(stderr," ");
        if ((i % 16) == 15)
        {
            fprintf(stderr,"%s\n",str);
        }
    }
    if ((i % 16) != 0)
    {
        for (j = i % 16; j < 16; j++)
        {
            fprintf(stderr,"   ");
            if ((j % 4) == 3) fprintf(stderr," ");
        }
        str[i % 16] = 0;
        fprintf(stderr,"%s\n",str);
    }

    fflush(stderr);
}
#endif


/**
 * Scan the byte stream continuously for a VNeS magic number.  When a
 * match is found then further processing of the message can continue
 * elsewhere.  Bytes are read one at a time and compared against the
 * magic number.  In case of no match, the process starts over, and is
 * repeated until the whole VNES magic number is found.
 *
 * @param header VNeS message header in which to fill in the magic number
 *
 * @return 0 if successful, or -1 on a fatal error
 */
static int
ScanForVnesMagic(VNESHEADER* header)
{
    int i = 0;

#ifdef DEBUG
    DebugPrint( "Scanning for VNeS magic number\n");
#endif

    for (;;) {
		if ( g_socketBufSize > 0 ) {
			header->dwMagicNum[i] = *g_socketBufPtr;
			g_socketBufSize -= 1;
			g_socketBufPtr += 1;
		} else {
			int r = ReadExact(rfbsockTCP, &header->dwMagicNum[i], sizeof(char));
			if ( r != sizeof(char) ) {
				DebugPrint( "magic read failed\n");
				return -1;
			}
		}

        // mismatch magic char, so restart
        if ( header->dwMagicNum[i] != VNES_MAGICNUMBER[i] ) {
            // skip the first character check if the current character
            // is a match for it already
            if ( header->dwMagicNum[i] == VNES_MAGICNUMBER[0] ) {
                header->dwMagicNum[0] = VNES_MAGICNUMBER[0];
#ifdef DEBUG
				DebugPrint( "restarting match on byte %d\n", i);
#endif
                i = 1;
            } else {
#ifdef DEBUG
				DebugPrint( "magic mismatch on byte %d\n", i);
#endif
				i = 0;
			}
        } else {
            i++;
        }

        if ( i == sizeof(VNES_MAGICNUMBER) ) {
            break;
        }
    }

#ifdef DEBUG
    DebugPrint( "Found VNeS magic number\n");
#endif

    return 0;
}


/*!  Read the VNeS header and verify that the fields match our
  settings.

  \return -1 on read error, 0 on mismatch or 1 on okay packet.
*/
int
ReadVnesHeader()
{
    int rd;
    VNESHEADER msgHeader;

    // DebugPrint( "Reading VNeS header...\n");

    // Want to make sure we process events at least once per update
    ProcessEvents();

    // Find a matching magic number
    if ( ScanForVnesMagic(&msgHeader) != 0 ) {
        return -1;
    }

    // Read in the remainder of the message header, skipping the
    // already read magic number
    rd = ReadExact(rfbsockTCP, ((char*)&msgHeader)+4, sizeof(VNESHEADER)-4);
    if ( rd != (sizeof(VNESHEADER) - 4) ) {
        DebugPrint( "VNeS header read failed.\n");
        return -1;
    }

    // Verify the magic number is intact
    if (memcmp(msgHeader.dwMagicNum, VNES_MAGICNUMBER, 4)) {
        DebugPrint( "VNeS magic number mismatch\n");
        return 0;
    }

    // Verify the major version is okay
    if (memcmp(msgHeader.MajorVersion, VNES_MAJORVERSION, 2)) {
        DebugPrint( "VNeS major version mismatch\n");
        return 0;
    }

    // Verify the server address matches ours
    if (memcmp(msgHeader.dwServerIP, appData.serverIP, 4)) {
        DebugPrint( "VNeS server mismatch\n");
        return 0;
    }

    // Check the command parameter
    msgHeader.dwCommand = Swap32IfLE(msgHeader.dwCommand);
    if ( msgHeader.dwCommand == VNES_ZOOM_RECT ) {
        fprintf(stderr, "VNES Zoom command recieved, but ignored\n");
    }

    return 1;
}

void DecryptDES(char* buf, unsigned int len)
{
	assert( (len % DESBLOCK_SIZE) == 0 );

	deskey( appData.encryptString, DE1 );

	// (D)encrypt in-place
	while ( len > 0 ) {
		des( buf, buf );
		buf += 8;
		len -= 8;
	}
}

void DecryptAES(char* buf, unsigned  int len)
{
	aes_context ctx;
	char deckey[16];

	assert( (len % AESBLOCK_SIZE) == 0 );

	// the key is only 8 bytes long, but AES needs a 16 byte key, so
	// just copy the key twice
	memcpy(deckey  , appData.encryptString, 8);
	memcpy(deckey+8, appData.encryptString, 8);

	aes_set_key(&ctx, deckey, 128);

	// (D)encrypt in-place
	while ( len > 0 ) {
		aes_decrypt( &ctx, buf, buf );
		buf += 16;
		len -= 16;
	}
}

void EncryptDES(char* buf, unsigned int len)
{
	assert( (len % DESBLOCK_SIZE) == 0 );

	deskey( appData.encryptString, EN0 );

	// (D)encrypt in-place
	while ( len > 0 ) {
		des( buf, buf );
		buf += 8;
		len -= 8;
	}
}

void EncryptAES(char* buf, unsigned int len)
{
	aes_context ctx;
	char deckey[16];

	assert( (len % AESBLOCK_SIZE) == 0 );

	// the key is only 8 bytes long, but AES needs a 16 byte key, so
	// just copy the key twice
	memcpy(deckey  , appData.encryptString, 8);
	memcpy(deckey+8, appData.encryptString, 8);

	aes_set_key(&ctx, deckey, 128);

	// (D)encrypt in-place
	while ( len > 0 ) {
		aes_encrypt( &ctx, buf, buf );
		buf += 16;
		len -= 16;
	}
}
