/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2004. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <cassert>
#include "imconfig.h"

#include "vncviewer.h"
#include "config.h"
#include "debug.h"
#include "helper.h"
#include "Dialog.h"
#include "InfoBar.h"

#ifdef USE_ESCVP21
#include "libkeyinput.h"
#include "SDL_e21.h"
#endif

namespace {
	const Uint32 DIALOG_TIMEOUT = 7000;      
	const SDL_Rect FULL_DIALOG_RECT = { 0, 472, 1024, 296 };
}

static Dialog* sg_dialog = 0;
static Uint32  sg_timeout;
static bool    sg_isDialogHidden = true;
static SDL_Surface* sg_shadowSurface;
static SDL_Rect sg_videoArea;

static void OpenDialog();
static void DrawDialog();
static void EventUser(const SDL_Event* event);
static void UpdateAlpha(SDL_Surface* surf, const SDL_Rect& rect, unsigned char val);
static void UpdateBlackBorders();
static void UpdateDisplayRect(SDL_Rect* rect);


void InitUI() 
{
    Uint8 lang;
#ifdef USE_ESCVP21
    if (LKI_OpenSocket(NULL) != LKI_SUCCESS) {
        DebugPrint("***** Failure to initialize LKI\n");
        Cleanup(ERROR_UI);
    }
    lang = get_lang();
#else
    lang = 0x01;
#endif  

    if (UIRsrcMgr::Instance(lang)->Init(UIRsrcMgr::ModVNC)) {
        DebugPrint("***** Failure to initialize UI resource manager\n");
        Cleanup(ERROR_UI);
    }

	sg_shadowSurface = SDL_CreateRGBSurface(SDL_SWSURFACE, 
											dpy->w, dpy->h,
											32, 
											dpy->format->Rmask,
											dpy->format->Gmask,
											dpy->format->Bmask,
											dpy->format->Amask);
	if ( sg_shadowSurface == NULL ) {
        DebugPrint("***** Failure to create dialog shadow surface\n");
        Cleanup(ERROR_UI);		
	}

	
	SDL_SetAlpha(sg_shadowSurface, 0, SDL_ALPHA_OPAQUE);

	
	SDL_FillRect(sg_shadowSurface, NULL,
				 SDL_MapRGBA(sg_shadowSurface->format, 0, 0, 0, SDL_ALPHA_OPAQUE));

	
	
	if ( sg_dialog == NULL ) {
		sg_dialog = new Dialog(sg_shadowSurface, VNES_DIALOG_DEFAULT_RECT);
		sg_dialog->Init();
	}
}


void DestroyUI()
{
    delete sg_dialog;
    sg_dialog = 0;

	if ( sg_shadowSurface ) {
		SDL_FreeSurface(sg_shadowSurface);
		sg_shadowSurface = NULL;
	}
}


void OpenDialog()
{
    DebugPrint("%s: Opening Dialog\n", __FUNCTION__);

	sg_isDialogHidden = false;

	sg_dialog->SelectButton(DlgBtn::BtnContinue);
}


void DrawDialog()
{
    if ( sg_isDialogHidden ) {
        return;
    }

	sg_dialog->Draw();
	sg_dialog->Update();

	SDL_Rect update = FULL_DIALOG_RECT;
	UpdateAlpha(sg_shadowSurface, update, (255 * 9 / 10));
	UpdateBlackBorders();
	UpdateDisplayRect(NULL);
}


void PollDialog(const SDL_Event* event)
{
    
    
    
    if ( event ) {
        if ( event->type == KEY_EVENT_TYPE ) {
            DebugPrint("%s: Got usuable key event\n", __FUNCTION__);
            sg_timeout = SDL_GetTicks();

            if ( sg_dialog == 0 || sg_isDialogHidden ) {
                if ( event->key.keysym.sym == SDLK_ESCAPE ) {        
                    OpenDialog();
					DrawDialog();
                }
                return;
            }
        } else if ( event->type == SDL_USEREVENT ) {
            DebugPrint("%s: Got user event\n", __FUNCTION__);
            EventUser(event);
        }
    }

    if ( !sg_isDialogHidden ) {
        if ((SDL_GetTicks() - sg_timeout) > DIALOG_TIMEOUT) {
            DebugPrint("%s: Timeout!\n", __FUNCTION__);
            CloseDialog();
            return;
        }
    }

    
    if ( !sg_isDialogHidden && sg_dialog && event ) {
        DebugPrint("%s: Passing event type '%d' to dialog\n", 
                   __FUNCTION__, (int)event->type);
        if (sg_dialog->Event(*event) == ProcKeepGoing) {
			SDL_Rect update = FULL_DIALOG_RECT;
			UpdateAlpha(sg_shadowSurface, update, (255 * 9 / 10));
			UpdateBlackBorders();
			UpdateDisplayRect(NULL);
		}
    }
}


void CloseDialog()
{
    DebugPrint("%s: Closing Dialog\n", __FUNCTION__);
    sg_timeout = 0;
	sg_isDialogHidden = true;

	SDL_Rect update = FULL_DIALOG_RECT;
	SDL_FillRect(sg_shadowSurface, &update, 
				 SDL_MapRGBA(sg_shadowSurface->format, 0, 0, 0, SDL_ALPHA_TRANSPARENT));

	UpdateBlackBorders();
	UpdateDisplayRect(&update);
}


static void
EventUser(const SDL_Event* event)
{
#ifdef USE_ESCVP21
    assert( event != 0 );

    DebugPrint("EventUser\n");
    if (event->user.code != SDL_E21_MAGIC) {
        return;
    }

    if (!event->user.data1) {
        return;
    }

    SDL_E21Event *e21ev = static_cast<SDL_E21Event *>(event->user.data1);
    switch(e21ev->code) {
    case EV_LANG:
        UIRsrcMgr::Instance()->SetLang(e21ev->param[0]);

        if ( sg_dialog ) {
            sg_dialog->UpdateResources();
            sg_dialog->Redraw();
            sg_dialog->Update();
        }

        LKI_Send(cmdIMSTAT, "21");
        break;
    default:
        break;
    }

    free(event->user.data1);
    const_cast<SDL_Event*>(event)->user.data1 = NULL;
#endif
}



void UpdateAlpha(SDL_Surface* surf, const SDL_Rect& rect, unsigned char val)
{
    assert(surf && surf->pixels != NULL);

	Uint16 width;
	Uint8* row;

	DebugPrint("%s: alpha = %d\trect = %d x %d + %d + %d\n", __FUNCTION__,
			   val, rect.w, rect.h, rect.x, rect.y);

    Uint8* dst_row = (Uint8*)surf->pixels + rect.y * surf->pitch +
		rect.x * surf->format->BytesPerPixel;
    Uint32 pitch   = surf->pitch;
    Uint16 height  = rect.h;

	if (surf->format->BytesPerPixel == 4) {
		while (height--) {
			width  = rect.w;
			row = dst_row;
			while (width--) {
				row[3] = val;	
				row += 4;
			}
			dst_row += pitch;
		}
	} else if ( surf->format->BytesPerPixel == 2 && val > 0 ) {
		while (height--) {
			width = rect.w;
			row = dst_row;
			while (width--) {
				row[1] |= 0x80;	
				row += 2;
			}
			dst_row += pitch;
		}

	} else if ( surf->format->BytesPerPixel == 2 ) {
		while (height--) {
			width  = rect.w;
			row = dst_row;
			while (width--) {
				row[1] &= 0x7F; 
				row += 2;
			}
			dst_row += pitch;
		}
    }
}



void SetVideoArea(SDL_Rect* area)
{
	assert(sg_shadowSurface != NULL);

	if ( (area == NULL) || (area->h == 0) || (area->w == 0) ) {
		sg_videoArea.x = 0;
		sg_videoArea.y = 0;
		sg_videoArea.w = sg_shadowSurface->w;
		sg_videoArea.h = sg_shadowSurface->h;
	} else {
		sg_videoArea = *area;
	}

	
	
	sg_videoArea.h--;

	DebugPrint("SetVideoArea: %d x %d + %d + %d\n", 
			   sg_videoArea.w, sg_videoArea.h, 
			   sg_videoArea.x, sg_videoArea.y);

	SDL_Rect varea = sg_videoArea;
	SDL_FillRect(sg_shadowSurface, &varea,
				 SDL_MapRGBA(sg_shadowSurface->format, 0, 0, 0, 0));

    if ( sg_isDialogHidden ) {
		UpdateBlackBorders();
		UpdateDisplayRect(NULL);
	} else {
		DrawDialog();
	}
}



void
UpdateBlackBorders()
{
	SDL_Rect tmpFill;

	assert(sg_shadowSurface != NULL);

	
	if ( sg_videoArea.y > 0 ) {
		DebugPrint("%s: drawing top border\n", __FUNCTION__);
		tmpFill.x = 0;
		tmpFill.y = 0;
		tmpFill.w = sg_shadowSurface->w;
		tmpFill.h = sg_videoArea.y;
		UpdateAlpha(sg_shadowSurface, tmpFill, SDL_ALPHA_OPAQUE);
	}

	
	if ( sg_videoArea.x > 0 ) {
		DebugPrint("%s: drawing left border\n", __FUNCTION__);
		tmpFill.x = 0;
		tmpFill.y = 0;
		tmpFill.w = sg_videoArea.x;
		tmpFill.h = sg_shadowSurface->h;
		UpdateAlpha(sg_shadowSurface, tmpFill, SDL_ALPHA_OPAQUE);
	}

	
	if ( (sg_videoArea.x + sg_videoArea.w) < sg_shadowSurface->w ) {
		DebugPrint("%s: drawing right border\n", __FUNCTION__);
		tmpFill.x = sg_videoArea.x + sg_videoArea.w;
		tmpFill.y = 0;
		tmpFill.w = sg_shadowSurface->w - (sg_videoArea.x + sg_videoArea.w);
		tmpFill.h = sg_shadowSurface->h;
		UpdateAlpha(sg_shadowSurface, tmpFill, SDL_ALPHA_OPAQUE);
	}

	
	if ( (sg_videoArea.y + sg_videoArea.h) < sg_shadowSurface->h ) {
		DebugPrint("%s: drawing bottom border\n", __FUNCTION__);
		tmpFill.x = 0;
		tmpFill.y = sg_videoArea.y + sg_videoArea.h;
		tmpFill.w = sg_shadowSurface->w;
		tmpFill.h = sg_shadowSurface->h - (sg_videoArea.y + sg_videoArea.h);
		UpdateAlpha(sg_shadowSurface, tmpFill, SDL_ALPHA_OPAQUE);
	}
}



void
UpdateDisplayRect(SDL_Rect* rect)
{
	SDL_BlitSurface(sg_shadowSurface, rect, dpy, rect);
}
