/*
    Copyright (C) SEIKO EPSON CORPORATION 2002-2004. All Rights Reserved.

    3-3-5, Owa, Suwa-shi Nagano-ken, 392-8502 Japan
    Phone: +81-266-52-3131

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <cassert>
#include <cstring>
#include <deque>
#include "debug.h"
#include "domain.h"
#include "netutil.h"
#include "vncviewer.h"
#include "vnesproto.h"
#include "vnes.h"

#define POLL_TIMEOUT    3000    
#define WINDOW_SIZE     5       


typedef std::deque<unsigned int> RequestQueue;


static bool SendRequest(unsigned int itemID);
static void ParseConfigMessage(const vnes_msg_config_t& config);

static RequestQueue s_requests;
static Domain* s_servSocket;


int ConnectDomainSocket()
{
    DebugPrint("Entering: %s\n", __func__);

    if ( s_servSocket != 0 ) {
        DebugPrint("%s called on already connected socket.\n", __func__);
        delete s_servSocket;
        s_servSocket = 0;
    }

    s_servSocket = new Domain();
    if ( s_servSocket == 0 ) {
        return -1;
    }

    
    
    char path[UNIX_PATH_MAX];

    if ( ! s_servSocket->Create(Domain::DATAGRAM) ) {
        NetworkUtil::PrintSocketError(*s_servSocket);
        delete s_servSocket;
        s_servSocket = 0;
        return -1;
    }

    memset(path, 0, sizeof(path));
    strcpy(path+1, VNES_SOCKET_PATH_V);
    if ( ! s_servSocket->Bind(path) ) {
        NetworkUtil::PrintSocketError(*s_servSocket);
        delete s_servSocket;
        s_servSocket = 0;
        return -1;
    }

    memset(path, 0, sizeof(path));
    strcpy(path+1, VNES_SOCKET_PATH_E);
    if ( ! s_servSocket->Connect(path) ) {
        NetworkUtil::PrintSocketError(*s_servSocket);
        delete s_servSocket;
        s_servSocket = 0;
        return -1;
    }

    s_servSocket->SetTimeout(POLL_TIMEOUT);

    return 0;
}


void DisconnectDomainSocket()
{
    DebugPrint("Entering: %s\n", __func__);

    delete s_servSocket;
    s_servSocket = 0;
}


void PollDomainSocket()
{
    DebugPrint("Enering PollDomainSocket\n");

    if ( s_servSocket == 0 ) {
        DebugPrint("Server socket is closed\n");
        return;
    }

    vnes_msg_t msg;
    int result = s_servSocket->Recv((char*)&msg, sizeof(msg));
    if ( result != sizeof(msg) ) {
        DebugPrint("Error on domain receive: %d\n", result);
        return;
    }

    switch (msg.type) {
    case VNES_MSG_CONFIG:
        DebugPrint("Got the configuration message\n");
        ParseConfigMessage(msg.msgdata.config);
        break;

    default:
        DebugPrint("*** Got unexpected message\n");
        break;
    }
}



void StartFullscreenUpdate(unsigned int totalID)
{
    DebugPrint("Entering: %s\n", __func__);

    
    for ( unsigned int i = 1; i <= totalID; i++ ) {
        s_requests.push_back(i);
    }
}



void AckFullscreenUpdate(unsigned int ackID)
{
    unsigned int minACK;

    DebugPrint("Entering: %s\n", __func__);

    
    
    
    if ( (WINDOW_SIZE + 1) > ackID ) {
        minACK = 1;
    } else {
        minACK = ackID - WINDOW_SIZE + 1;
    }

    RequestQueue::iterator item = s_requests.begin();
    while ( item != s_requests.end() ) {
        if ((*item) > ackID) {
            
            
            
            break;

        } else if ((*item) < minACK) {
            
            
            
            SendRequest((*item));
            s_requests.erase(item);

            
            item = s_requests.begin();

        } else if ((*item) < ackID) {
            
            
            item++;

        } else if ((*item) == ackID) {
            
            
            s_requests.erase(item);
            break;

        } else {
            
            DebugPrint("AckFullscreenUpdate: Unhandled possibility occured\n");
            break;
        }
    }
}



void EndFullscreenUpdate()
{
    DebugPrint("Entering: %s\n", __func__);

    while( ! s_requests.empty() ) {
        SendRequest(s_requests.front());
        s_requests.pop_front();
    }
}


int SendDomainMessage(const vnes_msg_t* message)
{
    DebugPrint("Entering: %s\n", __FUNCTION__);
    assert( s_servSocket != 0 );
    assert( message != 0 );

    DebugPrint("SendMessage: Sending message ID %d\n", message->type);

    int ret;
    ret = s_servSocket->Send((char*)message, sizeof(vnes_msg_t));
    if ( ret != sizeof(vnes_msg_t) ) {
        if ( ret < 0 ) {
            NetworkUtil::PrintSocketError(*s_servSocket);
        }
        DebugPrint("SendRequest: Failed to send\n");
        return false;
    }

    return true;    
}

static bool
SendRequest(unsigned int itemID)
{
    DebugPrint("Entering: %s\n", __func__);
    assert( s_servSocket != 0 );

    DebugPrint("SendRequest: Sending ID %d\n", itemID);

    vnes_msg_t message;
    message.type = VNES_MSG_REQ_FULLSCREEN;
    message.msgdata.fullscrn.id = itemID;

    int ret;
    ret = s_servSocket->Send((char*)&message, sizeof(message));
    if ( ret != sizeof(message) ) {
        if ( ret < 0 ) {
            NetworkUtil::PrintSocketError(*s_servSocket);
        }
        DebugPrint("SendRequest: Failed to send\n");
        return false;
    }

    return true;
}

static void
ParseConfigMessage(const vnes_msg_config_t& config)
{
    memcpy(appData.confHostname, config.hostname, 32);
    memcpy(appData.confAddress, config.address, 32);
    memcpy(appData.confEssid, config.essid, 32);
    memcpy(appData.confKeyword, config.keyword, 8);
    appData.confNetworkMode = config.network;

    Uint32 red, green, blue;

    
    sscanf(config.windowMaxes, "%u,%u,%u", &red, &green, &blue);
    si.format.redMax = red;
    si.format.greenMax = green;
    si.format.blueMax = blue;

    
    sscanf(config.windowShifts, "%u,%u,%u", &red, &green, &blue);
    si.format.redShift = (Uint8)red;
    si.format.greenShift = (Uint8)green;
    si.format.blueShift = (Uint8)blue;

    
    si.format.depth = atoi(config.windowDepth);
    si.format.bitsPerPixel = si.format.depth;
    si.framebufferHeight = atoi(config.windowHeight);
    si.framebufferWidth = atoi(config.windowWidth);

    
    int a;
    int addr[4];
    sscanf(config.projServerIP, "%u.%u.%u.%u",
           &addr[0], &addr[1], &addr[2], &addr[3]);
    for ( a = 0; a < 4; a++ ) {
        appData.serverIP[a] = (char)(addr[a] & 0xFF);
    }

    
    if ( config.useEncryption ) {
        appData.encryptString = (char*)malloc(VNES_MAX_STRING);
        memcpy(appData.encryptString, config.projEncryptPassword, VNES_MAX_STRING);
		DebugPrint("Encryption string: %8.8s\n", config.projEncryptPassword);
    } else {
        appData.encryptString = 0;
    }
	appData.useEncryptionType = config.useEncryption;

	switch ( appData.useEncryptionType ) {
	case VNES_ENC_OFF:
		DebugPrint("Encryption type: off\n");
		appData.encBlockSize = 0;
		break;
	case VNES_ENC_DES:
		DebugPrint("Encryption type: DES\n");
		appData.encBlockSize = 8;
		break;
	case VNES_ENC_AES:
		DebugPrint("Encryption type: AES\n");
		appData.encBlockSize = 16;
		break;
	default:
		break;
	}
}
